# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

from typing import *
import contextlib
import datetime
import errno
import functools
import http.server
import inspect
import json
import os
import re
import shutil
import socketserver
import subprocess
import sys
import threading
import time
import traceback
import urllib.parse
import urllib.request

from .auth import AuthMiddleware
from .appconfig import AppConfig
from .browser import BrowserHook
from .core import Directories, Middleware, log
from .dialogcontroller import DialogController
from .guibase import AbstractAction, AbstractFunction, BrowserSideDatasource, Icon
from .gui import Dialog, View
from .datastore import DataStore
from .i18n import TrStrOrStrTyping
from .pyhtmlview import PyHtmlView
from .reqhandler import RequestHandler
from .request import Request
from .userfeedback import UserFeedbackController


# noinspection PyUnusedLocal,PyShadowingBuiltins,PyProtectedMember
class Application:
    """!
    Base class for a Yaji web application.
    """

    _idcnt = 0

    def __init__(self, parentid: Optional[str] = None, returntoparent: bool = True, *,
                 show_browser_closed_notification: bool = False, stop_implicitly_when_browser_closed: bool = False,
                 skip_shutdown_dialog: bool = False, browser_hook_heartbeat_threshold: int = 60*2,
                 head1: Optional[str] = None, head2: Optional[str] = None, icon: Optional[str] = None,
                 mainview_icon: Optional[Icon] = None):
        """!
        @param parentid: Optional id of a parent Application, which e.g. can be redirected back to after exit.
        @param returntoparent: If to return back to the `parentid` application after exit in the user's browser.
        @param show_browser_closed_notification: If to show a useful notification (including a way to close the
                                                 application) after the browser was closed by the user and restarted.
        @param stop_implicitly_when_browser_closed: If to consider the application as intendedly stopped when the user
                                                    has closed the browser instead of opening a new one.
        @param skip_shutdown_dialog: If not to show a ui-blocking 'application stopped' dialog on ui shutdown.
        @param browser_hook_heartbeat_threshold: The time window within the backend excepts a heartbeat from the browser
                                                 side before it tries to open a new browser window.
        @param head1: The 1st level header text.
        @param head2: The 2nd level header text.
        @param icon: The window icon.
        @param mainview_icon: The icon of the mainview header.
        """
        Application._idcnt += 1
        self.__id = f"{os.getpid()}n{Application._idcnt}"
        self.__parentid = parentid
        self.__returntoparent = returntoparent
        self.__nextbrowseropenusetakeover = bool(parentid)
        self.__url = None
        self.__takenover = False
        self.__wetookoverfromurl = None
        self.__wetookoverfromurl_last = None
        self.__httpd = None
        self.__stopped = False
        self.__browserhook = None
        self.__lastopenedbrowserpid = None
        self.__middlewares = []
        self.__appconfig = AppConfig(self)
        self.__events = []
        self.__event_lastid = -1
        self.__eventcondition = threading.Condition()
        self.__datastores = {}
        self.__datastores_lock = threading.RLock()
        self.__datastores_autoidx = 0
        self.__wasreopened = False
        self.__clientscript = b""
        self.__clientstyle = b""
        self.__urlmap: List[Tuple['Pattern', str, Callable]] = []
        self.__requestparam_type_converters: Dict[type, Callable[[str], Any]] = {}
        self.__userfeedback = UserFeedbackController(self)
        self.__dialogs = DialogController(self)
        self.__pyhtmlview_enabled = None
        self.__staticfilelocations = [Directories.staticfiles]
        self.__translations = {}
        self.show_browser_closed_notification = show_browser_closed_notification
        self.stop_implicitly_when_browser_closed = stop_implicitly_when_browser_closed
        self.skip_shutdown_dialog = skip_shutdown_dialog
        self.appconfig.setconfig('_yj_browserHookHeartbeatThreshold', browser_hook_heartbeat_threshold)
        self.appconfig.setconfig('_yj_abstractfilesystemdialog_showChangeDriveButton', sys.platform == "win32")
        if head1 is not None:
            self.head1 = head1
        if head2 is not None:
            self.head2 = head2
        if icon:
            self.icon = icon
        if mainview_icon:
            self.mainview_icon = mainview_icon
        for mn in sorted(dir(type(self))):
            mv = getattr(type(self), mn)
            if callable(mv):
                urlpats = getattr(mv, '_urlpats', None)
                if urlpats is not None:
                    def _getfct(f):
                        @functools.wraps(f)
                        def nf(*a, **b):
                            return f(self, *a, **b)
                        return nf
                    for paramname, paramtype in getattr(mv, '__annotations__', {}).items():
                        # don't do much more than RequestHandler does!
                        mv = RequestHandler.with_param_type(paramname, paramtype)(mv)
                    mvfct = _getfct(mv)
                    for repat, pat in urlpats:
                        self.__urlmap.append((repat, pat, mvfct))
                        log.debug("Registered request handler %s for url pattern '%s'.", mv.__name__, pat)
        doinspection = False
        for superclss in reversed(inspect.getmro(type(self))):
            if superclss is Application:
                doinspection = True
                continue
            if not doinspection:
                continue
            try:
                scfilename = inspect.getfile(superclss)
            except TypeError:
                continue
            if scfilename:
                self.try_addclientresources_from_default_location(scfilename)
        self.add_middleware(self.__PostParamsMiddleware())
        self.add_middleware(self.__ComputeContentMiddleware())
        for trk, trv in {
            'YajiIsShutDown': {
                'en': "This application was stopped.\n\nYou can close this browser tab now.",
                'de': "Diese Anwendung wurde beendet.\n\nSie können dieses Browser-Tab nun schließen."},
            'YajiClosedNotificationText': {
                'en': "It seems you have closed the browser tab which shows this application.\n\nIf you did that with"
                      " the intention to close the application, you can do that right now, or simply continue.",
                'de': "Scheinbar haben Sie das Browser-Tab geschlossen, das diese Anwendung angezeigt hat.\n\nWenn Sie"
                      " den Wunsch hatten, die Anwendung zu schließen, können Sie das nun direkt tun, oder einfach"
                      " fortfahren."},
            'YajiPyHtmlViewSupportConfirmClose': {
                'en': "Do you really want to close the application?",
                'de': "Möchten Sie die Anwendung wirklich schließen?"},
            'YajiClosedNotificationButtonLabel': {
                'en': "Close the application",
                'de': "Diese Anwendung schließen"},
            'YajiProgramErrorOccurred': {
                'en': "An internal program error occurred!\n\nTo be safe, you should stop and restart the application"
                      " now.",
                'de': "Es ist ein interner Programmfehler aufgetreten.\n\nUm sicherzugehen sollten Sie die Anwendung"
                      " beenden und neu starten."},
            'YajiShowDetails': {
                'en': "Show details",
                'de': "Zeige Details"},
            'YajiHideDetails': {
                'en': "Hide details",
                'de': "Verstecke Details"},
            'YajiOpenDir': {
                'en': "Open Directory",
                'de': "Verzeichnis öffnen"},
            'YajiOpenFile': {
                'en': "Open File",
                'de': "Datei öffnen"},
            'YajiGoto': {
                'en': "Go to",
                'de': "Gehe zu"},
            'YajiCancel': {
                'en': "Cancel",
                'de': "Abbrechen"},
            'YajiCurrent': {
                'en': "Current",
                'de': "Aktuell"},
            'YajiFilesystem': {
                'en': "Filesystem",
                'de': "Dateisystem"},
            'YajiDrives': {
                'en': "Drives",
                'de': "Laufw."},
        }.items():
            self.add_translations(trk, **trv)

    @property
    def head1(self) -> Optional[TrStrOrStrTyping]:
        """!
        The 1st level header text.
        """
        return self.appconfig.getconfig("_yj_head1")

    @head1.setter
    def head1(self, s: TrStrOrStrTyping):
        self.appconfig.setconfig("_yj_head1", s)

    @property
    def head2(self) -> Optional[TrStrOrStrTyping]:
        """!
        The 2nd level header text.
        """
        return self.appconfig.getconfig("_yj_head2")

    @head2.setter
    def head2(self, s: TrStrOrStrTyping):
        self.appconfig.setconfig("_yj_head2", s)

    @property
    def bodyleftviewactionlabel(self) -> Optional[TrStrOrStrTyping]:
        """!
        The menu action label for the left main body widget.
        """
        return self.appconfig.getconfig("_yj_bodyleftviewactionlabel")

    @bodyleftviewactionlabel.setter
    def bodyleftviewactionlabel(self, label: TrStrOrStrTyping):
        self.appconfig.setconfig("_yj_bodyleftviewactionlabel", label)

    @property
    def bodyrightviewactionlabel(self) -> Optional[TrStrOrStrTyping]:
        """!
        The menu action label for the right main body widget.
        """
        return self.appconfig.getconfig("_yj_bodyrightviewactionlabel")

    @bodyrightviewactionlabel.setter
    def bodyrightviewactionlabel(self, label: TrStrOrStrTyping):
        self.appconfig.setconfig("_yj_bodyrightviewactionlabel", label)

    @property
    def showonly(self) -> Optional[int]:
        """!
        Switches to show only the left or right main body widget.
        Is `0` for left only, `1` for right only, `None` for both.
        """
        return self.appconfig.getconfig("_yj_showonly")

    @showonly.setter
    def showonly(self, v: Optional[int]):
        self.appconfig.setconfig("_yj_showonly", v)

    @property
    def icon(self) -> Optional[Icon]:
        """!
        The application icon.

        You may set it to a path to a png file (relative to `/static` as the web browser sees it).
        See add_staticfile_location().
        """
        return self.appconfig.getconfig("_yj_icon")

    @icon.setter
    def icon(self, icon: Optional[Union[Icon, str]]):
        if self.__httpd:
            raise RuntimeError("It is not allowed to set the icon after starting.")
        if isinstance(icon, str):
            nicon = Icon(by_url=f"static/{icon}")
        else:
            if icon.srcfunc != 'byUrl':
                raise ValueError("The application icon must be a 'byUrl' one.")
            nicon = icon
        self.appconfig.setconfig("_yj_icon", nicon)

    @property
    def mainview_icon(self) -> Optional[Icon]:
        """!
        The mainview header icon.

        The application icon is used if no mainview icon is set.

        You may set it to a path to a png file (relative to `/static` as the web browser sees it).
        See add_staticfile_location().
        """
        return self.appconfig.getconfig("_yj_mainviewIcon")

    @mainview_icon.setter
    def mainview_icon(self, icon: Optional[Union[Icon, str]]):
        self.appconfig.setconfig("_yj_mainviewIcon", Icon(by_url=f"static/{icon}") if isinstance(icon, str) else icon)

    @property
    def show_browser_closed_notification(self) -> bool:
        """!
        If to show a useful notification (including a way to close the application) after the browser was closed by the
        user and restarted.
        """
        return self.appconfig.getconfig('_yj_showBrowserClosedNotification')

    @show_browser_closed_notification.setter
    def show_browser_closed_notification(self, value: bool):
        self.appconfig.setconfig('_yj_showBrowserClosedNotification', value)

    @property
    def stop_implicitly_when_browser_closed(self) -> bool:
        """!
        If to consider the application as intendedly stopped when the user has closed the browser instead of opening
        a new one.
        """
        return self.appconfig.getconfig('_yj_stopImplicitlyWhenBrowserClosed')

    @stop_implicitly_when_browser_closed.setter
    def stop_implicitly_when_browser_closed(self, value: bool):
        self.appconfig.setconfig('_yj_stopImplicitlyWhenBrowserClosed', value)

    @property
    def skip_shutdown_dialog(self) -> bool:
        """!
        If not to show a ui-blocking 'application stopped' dialog on ui shutdown.
        """
        return self.appconfig.getconfig('_yj_skipShutdownDialog')

    @skip_shutdown_dialog.setter
    def skip_shutdown_dialog(self, value: bool):
        self.appconfig.setconfig('_yj_skipShutdownDialog', value)

    @property
    def browser_hook_heartbeat_threshold(self) -> int:
        """!
        The time window within the backend excepts a heartbeat from the browser side before it tries to open a new
        browser window.
        """
        return self.appconfig.getconfig('_yj_browserHookHeartbeatThreshold')

    @property
    def url(self) -> str:
        """!
        The application root url for browser access.
        """
        return self.__url

    @property
    def appconfig(self) -> AppConfig:
        """!
        The application AppConfig.
        """
        return self.__appconfig

    @property
    def userfeedback(self) -> UserFeedbackController:
        """!
        User feedback handler (for message boxes, ...).

        Note: You must not use it in usual request handlers, but only in downstream ones!
        See reqhandler.RequestHandler.run_downstream().
        """
        return self.__userfeedback

    @property
    def urlmap(self) -> List[Tuple['Pattern', str, Callable]]:
        """!
        The application's url to request handler map.
        """
        return self.__urlmap

    @property
    def takenover(self) -> bool:
        """!
        If the current application is taken over by another one and paused at the moment.
        """
        return self.__takenover

    @property
    def id(self) -> str:
        """!
        The application id.
        """
        return self.__id

    @property
    def parentid(self) -> str:
        """!
        The id of the parent application (if any).
        """
        return self.__parentid

    @property
    def wasreopened(self) -> bool:
        """!
        If the browser needed to be opened again.
        """
        return self.__wasreopened

    @property
    def returntoparent(self) -> bool:
        """!
        If to return to the parent application after exit.
        """
        return self.__returntoparent

    @property
    def isrunning(self) -> bool:
        """!
        Checks if the application is currently running (started and no stop triggered yet).
        """
        return self.__httpd and (not self.__stopped)

    @property
    def isstaticapplication(self) -> bool:
        """!
        Returns if this application is currently running as static one. See storeasstaticapplication().
        """
        return self.appconfig.getconfig('_yj_isStaticApplication', False)

    @property
    def current_request(self) -> Request:
        """!
        Returns the Request associated to the current request (in a request handler).
        """
        return Request.current()

    def create_datastore(self, name: Optional[str] = None, *, curtain: Union[bool, Optional[str]] = None) -> DataStore:
        """!
        Creates a new (real serverside) DataStore and returns it.
        See also get_datastore(), bindserver().
        @param name: The datastore name (e.g. used in yaji.getDatastore() on browser side).
                     Default is a random name.
        @param curtain: A curtain expression (pointing to a yaji.Curtain in browser side javascript), or `True` for
                        the default curtain (which might not be a good solution).
        """
        result = self.__create_datastore_helper(name, lambda n: DataStore(n, self, curtain=curtain))
        log.debug(f"Datastore '{result.name}' created.")
        return result

    def get_datastore(self, name: str, create_if_not_exist: bool = True) -> Optional[DataStore]:
        """!
        Returns a (real serverside) DataStore by name. If one exists, it returns that, otherwise see parameters.
        See also create_datastore(), bindserver().
        @param name: The datastore name.
        @param create_if_not_exist: If to create a new one if no one exists with that name yet
                                    (otherwise returns `None`).
        """
        return self.__get_datastore_helper(name, create_if_not_exist, lambda: self.create_datastore(name))

    def create_clientlocal_datastore(self, name: Optional[str] = None) -> BrowserSideDatasource:
        """!
        Creates a new browserside data source and returns it.
        See also get_clientlocal_datastore(), bindlocal().
        @param name: The datastore name (e.g. used in yaji.getClientlocalDatastore() on browser side).
                     Default is a random name.
        """
        result = self.__create_datastore_helper(name, lambda n: BrowserSideDatasource(n))
        log.debug(f"Local browserside datastore '{result.name}' created.")
        return result

    def get_clientlocal_datastore(self, name: str, create_if_not_exist: bool = True) -> Optional[BrowserSideDatasource]:
        """!
        Returns a browserside data source by name. If one exists, it returns that, otherwise see parameters.
        See also create_clientlocal_datastore(), bindlocal().
        @param name: The datastore name.
        @param create_if_not_exist: If to create a new one if no one exists with that name yet
                                    (otherwise returns `None`).
        """
        return self.__get_datastore_helper(name, create_if_not_exist, lambda: self.create_clientlocal_datastore(name))

    def start(self) -> None:
        """!
        Starts the application.

        Note: If your instance was already started and stopped, you have to create a new instance for another run.
        """
        if self.__stopped:
            raise RuntimeError("You must not start an application which already ran.")
        if self.__httpd:
            return
        self.oninitialize()
        port = 7310
        connected = False
        while not connected:
            try:
                self.__httpd = self.__YajiTCPServer(("", port), Application.__YajiHTTPRequestHandler, app=self)
                connected = True
            except OSError as e:
                if e.errno == errno.EADDRINUSE:
                    port += 1
                else:
                    raise
        self.__url = f"http://localhost:{port}/"
        os.environ[f"YAJI_{self.id}_URL"] = self.__url
        threading.Thread(target=self.__httpd.serve_forever, daemon=False).start()
        self.__browserhook = BrowserHook(self.__url, self)
        self.__browserhook.start()
        log.debug(f"Application started with url '{self.url}'.")

    def stop(self) -> None:
        """!
        Stops the application.

        - It requests the parent application's returntakeover procedure in some situations (see later),
        - then invokes stopping the user interface (triggers `_yj_stopui`),
          - which on browser side leads to 'uiShutdown' and back redirection to parent application in some situations
        - then shuts down the http server.
          - so the _yj_stopui event might never actually arrive, depending on timings (not an actual problem).

        A uiShutdown disables the user interface by triggering the `OnUiShutdown` event (but only one time; further
        calls are ignored). Application code can register custom handlers to this event, but it does at least this:
        - visually disables the user interface and shows a 'shut down' text if `skip_shutdown_dialog==False`
          (the default)
        - if it happened due to the user tried to close a PyHtmlView and `show_browser_closed_notification==True` (not
          the default): makes a '_yj_application_tryclosebrowser' request, so PyHtmlView support can close the window
          or parent applications can take control back

        The browser side also runs `uiShutdown` e.g. when
        - `_yj_pullevent` requests fail (assumption: the application code side has stopped), or
        - after a `yaji.stop()` was called (see later).

        A uiShutdown does not stop stuff like event polling, and will leave `yaji.isrunning==True`!
        The browser side `yaji.stop()` does the following:
        - requests '_yj_application_stop', which calls `Application.stop()` on application code side
          - when backend answered: running uiShutdown
          - with disabled user interface during the request if
            - `skip_shutdown_dialog==False` (the default) or
            - it was explicitly called this way

        Browser side calls `yaji.stop()` e.g. in those situations:
        - custom application code on browser side triggered it for application exit
        - a '_yj_unhandled_client_error' occurred and the backend decided for shutdown (which is not the default)
        - by PyHtmlView support when the user tries to close the window

        There is a flag `yaji.isrunning` which is `True` if the backend application is not known to be stopped so
        far, but might even be `True` after a uiShutdown. It will become `False` when
        - `_yj_pullevent` requests fail (assumption: the application code side has stopped), or
        - the `_yj_stopui` (see above) event was received.
        """
        if not self.__httpd:
            return
        def _stop():
            log.debug("Stopping application.")
            self.__browserhook.stop()
            for i in range(10):
                if not self.takenover:
                    break
                time.sleep(1)
            if self.__wetookoverfromurl:
                try:
                    turl = f"{self.__wetookoverfromurl}_yj_returntakeover"
                    with urllib.request.urlopen(turl) as f:
                        f.read()
                except OSError:
                    pass
            self.triggerevent("_yj_stopui", targeturl=self.__wetookoverfromurl if self.returntoparent else None)
            self.__wetookoverfromurl = None
            norequestcnt = 0
            for i in range(120):
                time.sleep(0.25)
                with self.__YajiHTTPRequestHandler.openrequestcnt_lock:
                    reqs = self.__YajiHTTPRequestHandler.openrequestcnt
                if reqs == 0:
                    if norequestcnt == 12:
                        break
                    norequestcnt += 1
                else:
                    norequestcnt = 0
            self.__httpd.shutdown()
            self.__httpd.socket.close()
            log.debug("Application stopped.")
        self.__stopped = True
        threading.Thread(target=_stop).start()

    def waituntilstopped(self) -> None:
        """!
        Blocks execution while the application is running.
        """
        log.debug(f"Thread {threading.current_thread().name} waiting until app is stopped.")
        while self.isrunning:
            time.sleep(1)

    def storeasstaticapplication(self, targetdir: str) -> None:
        """!
        Stores the application as static files.
        Note: The application has to support that, since there are some pitfalls.
        @param targetdir: The target directory path.
        """
        log.debug(f"Storing as static application to '{targetdir}'.")
        self.oninitialize()
        self.__httpd = object()  # otherwise it is not considered as 'running' internally
        self.appconfig.setconfig("_yj_isStaticApplication", True)
        os.makedirs(f"{targetdir}/static")
        with open(f"{targetdir}/index.html", "wb") as f:
            f.write(self._get_rootpagecontent())
        with open(f"{targetdir}/_yj_clientscript", "wb") as f:
            f.write(self._do__yj_clientscript())
        with open(f"{targetdir}/_yj_initscript", "wb") as f:
            f.write(self._do__yj_initscript())
        for sf in os.listdir(Directories.staticfiles):
            fsf = f"{Directories.staticfiles}/{sf}"
            if os.path.isfile(fsf):
                shutil.copyfile(fsf, f"{targetdir}/static/{sf}")

    def setbodyleft(self, cfg: Optional[View], switchto: bool = True) -> None:
        """!
        Sets the left main body widget.
        @param cfg: A view specification.
        @param switchto: If to switch to this view (if in single-side mode).
        """
        self.appconfig.setconfig("_yj_bodyleft", cfg)
        if switchto and cfg:
            self.switch_to_bodyleft()

    def setbodyright(self, cfg: Optional[View], switchto: bool = True) -> None:
        """!
        Sets the right main body widget.
        @param cfg: A view specification.
        @param switchto: If to switch to this view (if in single-side mode).
        """
        self.appconfig.setconfig("_yj_bodyright", cfg)
        if switchto and cfg:
            self.switch_to_bodyright()

    def switch_to_bodyleft(self) -> None:
        """!
        Switches to the left body view (if in single-side mode).
        """
        self.appconfig.setconfig("_yj_switchtobody", 0)

    def switch_to_bodyright(self) -> None:
        """!
        Switches to the right body view (if in single-side mode).
        """
        self.appconfig.setconfig("_yj_switchtobody", 1)

    def setactions(self, actions: List[AbstractAction]) -> None:
        """!
        Sets the main menu actions.
        @param actions: List or datasource of actions.
        """
        self.appconfig.setconfig("_yj_actions", actions)

    def setheadcontrol(self, cfg: Optional[View]) -> None:
        """!
        Sets the head control widget.
        @param cfg: A view specification.
        """
        self.appconfig.setconfig("_yj_headcontrol", cfg)

    def setsidebar(self, cfg: Optional[View]) -> None:
        """!
        Sets the sidebar widget.
        @param cfg: A view specification.
        """
        self.appconfig.setconfig("_yj_sidebar", cfg)

    def setsplitterposition(self, v: float) -> None:
        """!
        Sets the main body widgets splitter position.
        @param v: The splitter position (from left `0.0` to right `1.0`).
        """
        self.appconfig.setconfig("_yj_splitterposition", v)

    def create_dialog(self, cfg: View, **showcfg: Any) -> 'Dialog':
        """!
        Creates a dialog.
        @param cfg: The view specification.
        @param showcfg: Additional configuration for dialog presentation.
        """
        return Dialog(self.__dialogs, cfg, **showcfg)

    def add_clientstyle(self, content: Optional[AnyStr] = None, *, frompath: Optional[str] = None) -> None:
        """!
        Adds additional css application styles.

        Call this from your constructor.
        @param content: The content to return (as string or bytes).
        @param frompath: The file path to fetch the content from.
        """
        self.__clientstyle += self.__get_content(content, frompath)

    def add_clientscript(self, content: Optional[AnyStr] = None, *, frompath: Optional[str] = None) -> None:
        """!
        Adds application client javascript.

        Call this from your constructor.
        @param content: The content to return (as string or bytes).
        @param frompath: The file path to fetch the content from.
        """
        self.__clientscript += self.__get_content(content, frompath)

    def add_middleware(self, middleware: 'Middleware') -> None:
        """!
        Adds a `Middleware`.
        @param middleware: The middleware to add.
        """
        if self.__httpd:
            raise RuntimeError("It is not allowed to add middleware after starting.")
        self.__middlewares.append(middleware)

    def add_translations(self, stringname: str, **texts: str) -> None:
        """!
        Adds translations in some languages for a string.
        @param stringname: The text stringname, as used in i18n.TrStr.
        @param texts: Different translations by two-letter language code.
        """
        if self.__httpd:
            raise RuntimeError("It is not allowed to add translations after starting.")
        self.__translations[stringname] = {**self.get_translations(stringname), **texts}

    def get_translations_stringnames(self) -> List[str]:
        """!
        Returns all available stringnames for translations. See also add_translations().
        """
        return list(self.__translations.keys())

    def get_translations(self, stringname: str, **texts: Any) -> Dict[str, str]:
        """!
        Returns translations in all available languages for a string. See also add_translations().
        @param stringname: The text stringname.
        """
        return self.__translations.get(stringname, None) or {}

    def try_addclientresources_from_default_location(self, fpath: str) -> None:
        """!
        Tries to add a style file and script file from a file naming convention, so you don't have to call
        add_clientstyle/add_clientscript().

        If you pass `"/some/path/foo.py"`, it tries to load `"/some/path/__foo.css"` (and `.js`).

        @param fpath: Full path of your module file. You could use `__file__` for that.
        """
        scdir, scbare = os.path.split(fpath)
        if not scdir:
            raise ValueError(f"Not an absolute path: {fpath}")
        if scbare.endswith(".py"):
            scbare = scbare[:-3]
        for fext, adfct in [("js", self.add_clientscript), ("css", self.add_clientstyle)]:
            faf = f"{scdir}/__{scbare}.{fext}"
            if os.path.exists(faf):
                adfct(frompath=faf)
            else:
                log.debug(f"Tried to add '{faf}', but does not exist.")

    def add_staticfile_location(self, path: str) -> None:
        """!
        Adds a local file path as an additional location for reading static files from.
        Whenever static files are requested (e.g. by the browser requesting `/static/...`), Yaji tries to find it in one
        of those locations (in insertion order).
        @param path: The root path to add as static files source.
        """
        self.__staticfilelocations.append(os.path.abspath(path))

    def triggerevent(self, name: str, **eventdata: Any) -> None:
        """!
        Triggers an event on browser side.
        @param name: The event name (e.g. used in yaji.addEventHandler() on browser side).
        @param eventdata: Additional event data to pass.
        """
        log.debug("Triggering event '%s', %s.", name, eventdata)
        with self.__eventcondition:
            eventid = self.__event_lastid + 1
            self.__event_lastid = eventid
            eventdata.update({'name': name})
            self.__events.append((eventid, datetime.datetime.now(), eventdata))

    def do_stop_implicitly_when_browser_closed(self, value: bool = True) -> ContextManager:
        """!
        Use such instances via `with` for temporarily enabling implicit application stop when browser closed.
        @param value: If `False`, does the reverse this, same as `dont_stop_implicitly_when_browser_closed()`.
        """
        return Application.__SetStopImplicitlyWhenBrowserClosed(self, value)

    def dont_stop_implicitly_when_browser_closed(self) -> ContextManager:
        """!
        Use such instances via `with` for temporarily disabling implicit application stop when browser closed.
        """
        return self.do_stop_implicitly_when_browser_closed(False)

    def add_requestparam_type_converter(self, paramtype: type, converter: Callable[[str], Any]) -> None:
        """!
        Adds a custom type converter for request params.
        @param paramtype: The python type.
        @param converter: A function which converts a string coming from the web client to a `paramtype`.
        """
        self.__requestparam_type_converters[paramtype] = converter

    def bindserver(self, datasource: Union[str, DataStore], data_direction: Optional[str] = None, *,
                   convertername: Optional[str] = None):
        """!
        Declares binding a property of a `View` to a application code side data source.
        Note: You typically should use binddata() instead.
        @param datasource: The data source name or DataStore.
        @param data_direction: The data flow direction. See `BindDirection`.
        @param convertername: The converter name.
        """
        datasource = self.get_datastore(getattr(datasource, "name", datasource))
        dscfg = {'islocal': False, **datasource._to_simple_repr_()[1]}
        return View._DataBinding(dscfg, data_direction, convertername)

    def bindlocal(self, datasource: Union[str, BrowserSideDatasource], data_direction: Optional[str] = None, *,
                  convertername: Optional[str] = None):
        """!
        Declares binding a property of a `View` to a local browser side data source.
        Note: You should use binddata() instead, unless you want to use it with a data source name which might not exist
        yet.
        @param datasource: The data source name or BrowserSideDatasource.
        @param data_direction: The data flow direction. See `BindDirection`.
        @param convertername: The converter name.
        """
        datasource = self.get_clientlocal_datastore(getattr(datasource, "name", datasource))
        dscfg = {'islocal': True, **datasource._to_simple_repr_()[1]}
        return View._DataBinding(dscfg, data_direction, convertername)

    def binddata(self, datasource: Union[str, DataStore, BrowserSideDatasource], data_direction: Optional[str] = None,
                 *, convertername: Optional[str] = None):
        """!
        Declares binding a property of a `View` to an application code side or local browser side data source.
        This is similar to bindserver() and bindlocal(), but detects automatically if this is a server or local binding.
        See also create_datastore(), create_clientlocal_datastore().
        @param datasource: The data source name or DataStore/BrowserSideDatasource. This data source must already exist.
        @param data_direction: The data flow direction. See `BindDirection`.
        @param convertername: The converter name.
        """
        datasource = self.get_datastore(getattr(datasource, "name", datasource), create_if_not_exist=False)
        if isinstance(datasource, DataStore):
            return self.bindserver(datasource, data_direction, convertername=convertername)
        elif isinstance(datasource, BrowserSideDatasource):
            return self.bindlocal(datasource, data_direction, convertername=convertername)

    def bindevent(self, function: AbstractFunction, *, curtain: Union[bool, Optional[str]] = None):
        """!
        Declares binding an event property of a `View` to a application code side or local browser side action.
        This triggers the execution of that action whenever that event is triggered.
        @param function: The function to bind to this event.
        @param curtain: A curtain expression (pointing to a yaji.Curtain in browser side javascript), or `True` for
                        the default curtain (which might not be a good solution).
        """
        return View._EventBinding(function, curtain)

    def enable_authentication(self, *, authfct: Callable[[str, str], bool], realm: Optional[str] = None) -> None:
        """!
        Enables authentication by a custom authenticator function.
        @param authfct: The authenticator function: f(username, password) -> `True` iff authenticated.
        @param realm: The http auth realm name.
        """
        class _CustomAuthMiddleware(AuthMiddleware):
            def _authenticate_by_password(self, username, password):
                return authfct(username, password)
        self.add_middleware(_CustomAuthMiddleware(realm=realm or type(self).__name__))

    @property
    def _middlewares(self) -> List['Middleware']:
        """!
        Returns the sorted list of middlewares.
        @a_infrastructure
        """
        return self.__middlewares

    def _openbrowser_pyhtmlviewargs(self) -> Dict[str, Any]:
        """!
        Returns additional arguments for pyhtmlview.
        @a_dooverride_optional
        """
        return dict(wndclassname=type(self).__name__, scriptedclosing="yaji.pyhtmlview._onwindowclose()",
                    icon=self._staticfile_path_to_abspath(self.icon.src[len("static/"):] if self.icon else "yaji.png"))

    def _openbrowser(self, url: str) -> None:
        """!
        Opens a browser view for ui rendering.
        @a_infrastructure
        @param url: The url to open.
        """
        self.__pyhtmlview_enabled = False
        self.__lastopenedbrowserpid = 0
        self.onopenbrowserinformationoutput(0, f"Your application runs at url {url}.")
        log.debug(f"Trying to open a browser view for '{url}'.")
        if self.__nextbrowseropenusetakeover:
            self.__nextbrowseropenusetakeover = False
            try:
                self.__wetookoverfromurl = self.__wetookoverfromurl_last = os.environ[f"YAJI_{self.parentid}_URL"]
                turl = f"{self.__wetookoverfromurl}_yj_takeover?url={urllib.parse.quote_plus(self.url)}"
                with urllib.request.urlopen(turl) as f:
                    f.read()
                log.debug(f"Taken over from application {self.parentid} by url '{turl}'.")
                return
            except OSError:
                pass
        cmdline = os.getenv("YAJI_CALLCMD")
        if not cmdline:
            if PyHtmlView.is_system_compatible():
                log.debug("Enabling PyHtmlView.")
                PyHtmlView.openview(show_url=url, **self._openbrowser_pyhtmlviewargs())
                self.__pyhtmlview_enabled = True
                return
        if not cmdline:
            if sys.platform == "win32":
                # noinspection PyUnresolvedReferences
                threading.Thread(target=lambda: os.startfile(url)).start()
            else:
                try:
                    subprocess.check_output(["xdg-open --version"], shell=True)
                    cmdline = "nohup xdg-open %u"
                except subprocess.CalledProcessError:
                    self.onopenbrowserinformationoutput(1, "Please start a browser and navigate to the application.")
        if cmdline:
            def _open():
                nonlocal cmdline
                cmdline_shell = False
                if isinstance(cmdline, str):
                    cmdline = [cmdline]
                    cmdline_shell = True
                cmdline = [c.replace("%u", url) for c in cmdline]
                with subprocess.Popen(cmdline, shell=cmdline_shell, stdout=subprocess.DEVNULL,
                                      stderr=subprocess.DEVNULL) as bp:
                    self.__lastopenedbrowserpid = bp.pid
                    bp.wait()
            threading.Thread(target=_open, daemon=True).start()

    def _tryclosebrowser(self, withchecks: bool = True) -> bool:
        """!
        Tries to close the browser window.

        Note: This will not work in all situations, e.g. typically not if the app is opened in the system browser.
        Your application should handle that situation gracefully (e.g. by the default app shutdown notification).
        @a_infrastructure
        @param withchecks: If set, it makes some checks before, e.g. if implicit stopping is enabled.
        """
        if withchecks:
            if not (self.stop_implicitly_when_browser_closed or self.show_browser_closed_notification):
                return False
        log.debug("Trying to close the browser window.")
        success = False
        if self.__pyhtmlview_enabled and self.__lastopenedbrowserpid:
            try:
                os.kill(self.__lastopenedbrowserpid, 15)
            except OSError:
                pass
            success = True
        elif self.__wetookoverfromurl_last:
            try:
                with urllib.request.urlopen(f"{self.__wetookoverfromurl_last}_yj_application_tryclosebrowser") as f:
                    tosucc = f.read()
            except Exception:
                tosucc = ""
            return tosucc == b"true"
        return success

    def _browser_was_reopened(self) -> None:
        """!
        Notifies that the browser needed to be reopened.
        @a_infrastructure
        """
        self.__wasreopened = True
        self.onbrowserreopened()

    def _get_requestparam_type_converter(self, paramtype):
        """!
        Returns a type converter for a request param.
        @a_infrastructure
        @param paramtype: The python type of the parameter to convert.
        """
        return self.__requestparam_type_converters.get(paramtype, None)

    def _json_make_serializable(self, o: Any) -> Any:
        """!
        Makes some special objects json serializable.
        @a_infrastructure
        @a_dooverride_optional
        @param o: The object to serialize.
        """
        if isinstance(o, datetime.date) and not isinstance(o, datetime.datetime):
            o = datetime.datetime.combine(o, datetime.time())
        elif isinstance(o, datetime.time) and not isinstance(o, datetime.datetime):
            o = datetime.datetime.combine(datetime.date.fromtimestamp(0), o)
        if isinstance(o, datetime.datetime):
            return {'_yj_ser_obj_': '__Date', 'value': o.timestamp()}
        simplereprfct = getattr(o, '_to_simple_repr_', None)
        if simplereprfct:
            tn, td = simplereprfct()
            if tn is None:
                return None
            return {'_yj_ser_obj_': tn, **td}  # browserside has to provide an interpreter for it
        else:
            raise TypeError(type(o).__name__)

    def _staticfile_path_to_abspath(self, relpath: str) -> str:
        """!
        Searches a file in the pool of static files locations by relative path and returns the absolute path if such a
        file exists. See add_staticfile_location().
        @param relpath: The relative static file path.
        """
        for bfile in [f"{bdir}/{relpath}" for bdir in self.__staticfilelocations]:
            if os.path.isfile(bfile):
                return os.path.abspath(bfile)

    def _findhandler(self, request: Request) -> Optional[Callable]:
        """!
        Returns the handler function for a request.
        @a_dooverride_optional
        @param request: The request to serve.
        """
        fct, urlparams = self.__get_requesthandler_for_url(request.urlbarepath, self)
        for k, v in urlparams.items():
            request._set_lparam(k, [v])
        return fct

    def _get_rootpagecontent(self) -> bytes:
        """!
        Returns the main page content.
        @a_infrastructure
        """
        with open(f"{Directories.staticfiles}/yaji.html", "rb") as f:
            cnt = f.read()
        cnt = cnt.replace(b"{{ appiconsrc }}", f"static/{self.icon or 'yaji.png'}".encode())
        cnt = cnt.replace(b"{{ additionalstyle }}", self.__clientstyle)
        return cnt

    def _callhandler(self, handler: Callable,
                     request: Request) -> Tuple[Optional[Any], Optional[Exception], Optional[str]]:
        """!
        Calls a request handler for a request and returns its result, but includes some error handling.
        @a_infrastructure
        @param handler: The request handler.
        @param request: The request.
        """
        responsebody = exception = errortext = None
        try:
            responsebody = handler(**{k: v for k, v in request.params.items() if not k.startswith("__")})
        except Exception as e:
            exception, errortext = e, (traceback.format_exc() or "error")
            self.onprocessrequesterror(request.urlpath, e)
        return responsebody, exception, errortext

    def oninitialize(self) -> None:
        """!
        Initializes the application.
        @a_dooverride_optional
        This is called during application startup, so later than `__init__` (or never if the app doesn't start).
        Note: You should just override `__init__` instead if possible!
        """
        pass

    def onprocessrequesterror(self, request: str, error: Exception) -> None:
        """!
        Reacts on process request errors, e.g. by logging.
        @a_dooverride_optional
        @param request: The path part of the request url.
        @param error: The Exception.
        """
        log.error(f"Program error in processing request '{request}': {traceback.format_exc()}")

    def onopenbrowsererror(self) -> None:
        """!
        Reacts on errors when opening the browser.
        @a_dooverride_optional
        """
        log.error("It was impossible to open a browser for showing the application.\n"
                  "Please open a browser manually and go to the application address (see above).")

    def onunhandledclienterror(self, error: str, src: str, line: str) -> bool:
        """!
        Reacts on unhandled client errors. Returning `True` makes the application shut down.
        @a_dooverride_optional
        @param error: An error description (might be empty).
        @param src: The source file where the error occurred (might be empty).
        @param line: The line in the source file where the error occurred (might be empty).
        """
        errortext = str(error)
        src = src or ""
        if line:
            src += f" :{line}"
        if src:
            errortext += f"\n({src})"
        self.triggerevent("_yj_showunhandlederror", errortext=errortext)
        return False

    def onopenbrowserinformationoutput(self, kind: int, message: str) -> None:
        """!
        Reacts on information output (e.g. by printing it to stdout) while opening the browser.
        @a_dooverride_optional
        @param kind: The internal code of the message.
        @param message: The message text.
        """
        print(message)

    def onbrowserreopened(self) -> None:
        """!
        Reacts on the fact that the browser was opened again.
        @a_dooverride_optional
        """
        pass

    @RequestHandler.for_urls(auto=True)
    def _do__yj_clientscript(self):
        return self.__clientscript

    @RequestHandler.for_urls(auto=True)
    def _do__yj_initscript(self):
        with self.__eventcondition:
            appconfigs = self.appconfig.getconfigs(_markinitialized=True)
            srunning = json.dumps(self.isrunning)
            sreopened = json.dumps(self.__wasreopened)
            seventlastid = json.dumps(self.__event_lastid)
            sappconfigs = json.dumps(appconfigs, default=self._json_make_serializable)
            si18n = json.dumps(self.__translations)
            self.__wasreopened = False
            return f"yaji._initialize({srunning}, {sreopened}, {seventlastid}, {sappconfigs}, {si18n});".encode()

    @RequestHandler.for_urls(auto=True)
    def _do__yj_pullevent(self, lastid: int):
        self.__browserhook.heartbeat()
        begintime = datetime.datetime.now()
        while (datetime.datetime.now() - begintime).total_seconds() < 3:
            result = []
            for eid, etime, eobject in self.__events:
                if eid > lastid:
                    result.append({'id': eid, 'data': eobject})
            if len(result) > 0:
                return result
            with self.__eventcondition:
                self.__eventcondition.wait(1)
        return {}

    @RequestHandler.for_urls(auto=True)
    def _do__yj_lasteventid(self):
        with self.__eventcondition:
            return self.__event_lastid

    @RequestHandler.for_urls(auto=True)
    def _do__yj_datastore_pull(self, name, id: Optional[int] = None):
        with self.__get_datastore_cell(name, id) as (datastore, cell, ptr):
            if cell:
                rows = []
                for irow in range(datastore.rowcount(ptr)):
                    cols = []
                    for icol in range(datastore.columncount(ptr)):
                        icell = datastore.ptr_to_node(datastore.valuepointer(irow, icol, ptr))
                        cols.append((icell.ajaxid, icell.value))
                    rows.append(cols)
                return {"value": cell.value, "childs": rows, "id": cell.ajaxid}

    @RequestHandler.for_urls(auto=True)
    def _do__yj_datastore_push(self, name, value, id: Optional[int] = None):
        with self.__get_datastore_cell(name, id) as (datastore, cell, ptr):
            if cell:
                datastore.setvalue(value, ptr=ptr)

    @RequestHandler.for_urls(auto=True)
    def _do__yj_datastore_info(self, name):
        return self.get_datastore(name, create_if_not_exist=False)

    @RequestHandler.for_urls(auto=True)
    def _do__yj_application_tryclosebrowser(self, checks: bool = True):
        return self._tryclosebrowser(checks)

    @RequestHandler.for_urls(auto=True)
    def _do__yj_application_stop(self):
        self.stop()

    @RequestHandler.for_urls(auto=True)
    def _do__yj_unhandled_client_error(self, error="", src="", line=""):
        shutdown = self.onunhandledclienterror(error, src, line)
        return {'shutdown': bool(shutdown)}

    @RequestHandler.for_urls(auto=True)
    def _do__yj_answer_userfeedback(self, **_):
        self.userfeedback._interpret_response(self.current_request)

    @RequestHandler.for_urls(auto=True)
    def _do__yj_dialogs_list(self):
        return self.__dialogs.get_dialogs()

    @RequestHandler.for_urls(auto=True)
    def _do__yj_dialogs_close(self, dialogid: int):
        dlg = self.__dialogs.get_dialog_by_id(dialogid)
        if dlg:
            dlg.close()

    @RequestHandler.for_urls(auto=True)
    def _do__yj_listfs(self, path, only_dirs, show_hidden):
        dirs = []
        if sys.platform == "win32" and path == "":
            import string
            from ctypes import windll
            drivesmask = windll.kernel32.GetLogicalDrives()
            for letter in string.ascii_uppercase:
                if drivesmask & 1:
                    dirs.append((False, letter + ":"))
                drivesmask >>= 1
        else:
            if sys.platform == "win32" and re.match("/([A-Z]):", path):
                path = path[1:]
            path = os.path.abspath(path)
            try:
                for x in os.listdir(path):
                    isfile = os.path.isfile(f"{path}/{x}")
                    if not (only_dirs and isfile):
                        if show_hidden or (not x.startswith(".")):
                            dirs.append((isfile, x))
            except IOError:
                pass
            dirs = sorted(dirs)
            if path != os.path.dirname(path):
                dirs.insert(0, (False, ".."))
        return {'path': path, 'dirs': dirs}

    @RequestHandler.for_urls(auto=True)
    def _do__yj_setappconfigvalue(self, key, value):
        self.appconfig.setconfig(key, json.loads(value))

    @RequestHandler.for_urls(auto=True)
    def _do__yj_takeover(self, url):
        self.__takenover = True
        self.appconfig.setconfig("_yj_takeoverurl", url)

    @RequestHandler.for_urls(auto=True)
    def _do__yj_returntakeover(self):
        self.__takenover = False
        self.appconfig.setconfig("_yj_takeoverurl", None)

    def __get_requesthandler_for_url(self, urlpath: str,
                                     app: 'Application') -> Tuple[Optional[Callable], Dict[str, str]]:
        """!
        Returns the registered request handler function for a request url path.
        @param urlpath: The path segment of the request url.
        """
        for repat, pat, fct in self.__urlmap:
            m = repat.fullmatch(urlpath)
            if m:
                return fct, m.groupdict()
        return None, {}

    @contextlib.contextmanager
    def __get_datastore_cell(self, datastorename: str, id: Optional[int]):
        """!
        Helper for datastore accesses.
        """
        datastore = self.__datastores[datastorename]
        with datastore:
            ptr = None
            if id is not None:
                cell = DataStore.getnodebyid(id)
                if cell is not None:
                    ptr = datastore.node_to_ptr(cell)
            else:
                cell = datastore.rootnode
            yield datastore, cell, ptr

    def __create_datastore_helper(self, name: Optional[str], createfct: Callable[[str], Any]) -> Any:
        """!
        Creates and adds a new (serverside or browserside) datastore.
        @param name: The datastore name (or `None`).
        @param createfct: Callback function for actually creating a new datastore.
        """
        with self.__datastores_lock:
            if not name:
                while (not name) or (name in self.__datastores):
                    name = f"_yj_{self.__datastores_autoidx}"
                    self.__datastores_autoidx += 1
            elif name in self.__datastores:
                raise RuntimeError(f"Datastore '{name}' already exists.")
            self.__datastores[name] = result = createfct(name)
        return result

    def __get_datastore_helper(self, name: str, create_if_not_exist: bool,
                               creatfct: Callable[[], Any]) -> Optional[Any]:
        """!
        Returns a (serverside or browserside) datastore by name if one exists, otherwise see parameters.
        @param name: The datastore name.
        @param create_if_not_exist: If to create a fresh one (otherwise returns `None`) if no one exists with that name.
        @param creatfct: Callback function for actually creating a new datastore.
        """
        with self.__datastores_lock:
            r = self.__datastores.get(name, None)
            if (not r) and create_if_not_exist:
                r = creatfct()
        return r

    @staticmethod
    def __get_content(content: Optional[AnyStr], frompath: Optional[str]) -> bytes:
        """!
        Returns a bytes content with an additional ending newline.
        @param content: The content to return (as string or bytes).
        @param frompath: The file path to fetch the content from.
        """
        result = b""
        if isinstance(content, str):
            result += content.encode() + b"\n"
        elif isinstance(content, bytes):
            result += content + b"\n"
        elif content is not None:
            result += str(content).encode() + b"\n"
        if frompath is not None:
            with open(frompath, "rb") as f:
                result += f.read() + b"\n"
        return result

    class __SetStopImplicitlyWhenBrowserClosed:
        """!
        @a_infrastructure
        See do_stop_implicitly_when_browser_closed() and dont_stop_implicitly_when_browser_closed().
        """

        def __init__(self, app: 'Application', val: bool):
            self.app = app
            self.orig = False
            self.val = val

        def __enter__(self):
            self.orig = self.app.stop_implicitly_when_browser_closed
            self.app.stop_implicitly_when_browser_closed = self.val
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.app.stop_implicitly_when_browser_closed = self.orig

    class __PostParamsMiddleware(Middleware):
        """!
        Middleware for parsing http POST parameters.
        """

        @Middleware.request_processor(index=-200_000)
        def _readpostparams(self, request):
            nativerequesthandler = request.scratchpad.get('__nativerequesthandler', None)
            if ('content-length' in request.header) and nativerequesthandler:  # parse POST parameters
                data_string = nativerequesthandler.rfile.read(int(request.header['content-length']))
                conttype = request.header.get('content-type',
                                              "application/x-www-form-urlencoded").split(";")[0].strip()
                if conttype == "application/json":
                    pparams = {k: [v] for k, v in json.loads(data_string.decode()).items()}
                elif conttype == "application/x-www-form-urlencoded":
                    pparams = urllib.parse.parse_qs(data_string.decode(), keep_blank_values=True)
                else:
                    pparams = {}
                for k, v in pparams.items():
                    # noinspection PyProtectedMember
                    request._set_lparam(k, v)

    class __ComputeContentMiddleware(Middleware):
        """!
        Middleware which computes the response content, e.g. by fetching files or executing a request handler.
        """

        def __callhandler(self, handler: Callable,
                          request: Request) -> Tuple[Optional[Any], Optional[Exception], Optional[str]]:
            """!
            Calls a request handler for a request and returns its result.
            @a_infrastructure
            @param handler: The request handler.
            @param request: The request.
            """
            log.debug("Calling handler '%s'.", handler.__name__)
            return request.application._callhandler(handler, request)

        # noinspection PyProtectedMember
        @Middleware.request_processor(index=-100_000)
        def _computecontent(self, request):
            if request.urlbarepath == "/":
                request.set_response_header("Content-type", "text/html;charset=utf-8")
                request.response_body = request.application._get_rootpagecontent()
            elif request.urlbarepath.startswith("/static/"):
                bfile = request.application._staticfile_path_to_abspath(request.urlbarepath[7:])
                if bfile:
                    with open(bfile, "rb") as fff:
                        cnt = fff.read()
                    request.response_body = cnt
            else:
                request.set_response_header("Content-type", "text/javascript;charset=utf-8")
                handler = request.application._findhandler(request)
                if handler:
                    result, _, errortxt = self.__callhandler(handler, request)
                    if isinstance(result, bytes):
                        request.response_body = result
                    else:
                        jresult = json.dumps(result, default=request.application._json_make_serializable)
                        request.response_body = jresult.encode()
                    request.response_errortext = errortxt

    class __YajiTCPServer(socketserver.ThreadingTCPServer):
        """!
        The tcp server.
        @a_infrastructure
        """

        def __init__(self, *a, app: 'Application', **b):
            super().__init__(*a, **b)
            self.app = app

    # noinspection PyProtectedMember
    class __YajiHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
        """!
        The http request handler.
        @a_infrastructure
        """

        openrequestcnt_lock = threading.Lock()
        openrequestcnt = 0

        def log_message(self, sformat, *args):
            pass  # skip logging from http.server

        def do_HEAD(self):
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()

        # noinspection PyPep8Naming
        def do_POST(self):
            return self.do_GET()

        def do_GET(self):
            with self.openrequestcnt_lock:
                self.openrequestcnt += 1
            try:
                log.debug(f"Requested '%s'.", self.path)
                request = Request(self.path, {k.lower(): v for k, v in self.headers.items()}, self.server.app)
                request.scratchpad['__nativerequesthandler'] = self  # you should not use that
                # generate answer
                response = request
                with request._in_request():
                    for requestprocessor in self.__requestprocessors():
                        if response.skip_processing:
                            break
                        response = requestprocessor(response) or response
                # send answer
                if response.response_httpcode is None:
                    if response.response_errortext:
                        response.response_httpcode = 500
                    elif response.response_body is None:
                        response.response_httpcode = 404
                    else:
                        response.response_httpcode = 200
                self.send_response(response.response_httpcode)
                for headerkey in response.get_response_header_keys():
                    self.send_header(headerkey, response.get_response_header(headerkey))
                self.end_headers()
                try:
                    if 400 <= response.response_httpcode <= 599:
                        self.wfile.write((response.response_errortext or "").encode())
                    else:
                        self.wfile.write(response.response_body)
                except BrokenPipeError:
                    log.debug("Connection lost while responding to '%s'.", self.path)
            finally:
                with self.openrequestcnt_lock:
                    self.openrequestcnt -= 1

        @functools.lru_cache()
        def __requestprocessors(self) -> List[Callable]:
            requestprocessors = []
            for mw in self.server.app._middlewares:
                for mwk in dir(mw):
                    mwrp = getattr(mw, mwk)
                    mwrpt = getattr(mwrp, '_yj_middleware', None)
                    if mwrpt:
                        requestprocessors.append((mwrpt[0], mwrp))
            requestprocessors.sort(key=lambda x: x[0])
            return [x[1] for x in requestprocessors]
