# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os


class Directories:
    """!
    Some important paths.
    """

    base = os.path.dirname(os.path.abspath(__file__))

    staticfiles = f"{base}/static"


class Middleware:
    """!
    A middleware controls some internal stuff like request processing in a custom way.

    See `Application.add_middleware()`.
    """

    @staticmethod
    def request_processor(index: int = 0):
        """!
        Marks a function as a request processor. It takes a `Request` parameter and reads and/or modifies parts of it.
        @param index: The execution order index (low value: early execution).
        """
        def _f(f):
            f._yj_middleware = (index, )
            return f
        return _f


class Log:
    """!
    Logging.
    """

    def __init__(self):
        self.__logger = logging.getLogger("yaji")
        if os.getenv("YAJI_LOGDEBUG") == "1":
            sh = logging.StreamHandler()
            sh.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
            sh.setLevel(logging.DEBUG)
            self.__logger.setLevel(logging.DEBUG)
            self.__logger.addHandler(sh)

    def debug(self, s: str, *a) -> None:
        """!
        Logs a debug message.
        """
        self.__logger.debug(s, *a)

    def info(self, s: str, *a) -> None:
        """!
        Logs an info message.
        """
        self.__logger.log(s, *a)

    def warning(self, s: str, *a) -> None:
        """!
        Logs a warning message.
        """
        self.__logger.warning(s, *a)

    def error(self, s: str, *a) -> None:
        """!
        Logs an error message.
        """
        self.__logger.error(s, *a)


log = Log()
