# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

from typing import *
import datetime
import functools
import re
import threading
import typing


class RequestHandler:
    """!
    Decorators for request handler functions, i.e. methods of an Application which handle http requests.
    """

    @classmethod
    def __compile_re_for_urlpattern(cls, urlpattern: str) -> 'Pattern':
        """!
        Returns a compiled `re` regexp for a url pattern string.
        @param urlpattern: The url pattern string to convert.
        """
        srepat = ""
        inparam = False
        remainingurlpattern = urlpattern
        while remainingurlpattern:
            if inparam:
                ine = remainingurlpattern.find(">")
                if ine == -1:
                    raise ValueError(f"bad url pattern; '<' and '>' imbalanced: {urlpattern}")
                paramname = remainingurlpattern[:ine]
                srepat += f"(?P<{paramname}>[^?/]*)"
                inparam = False
                remainingurlpattern = remainingurlpattern[ine+1:]
            else:
                inb = remainingurlpattern.find("<")
                if inb > -1:
                    inparam = True
                else:
                    inb = len(remainingurlpattern)
                srepat += re.escape(remainingurlpattern[:inb])
                remainingurlpattern = remainingurlpattern[inb+1:]
        return re.compile(srepat)

    # noinspection PyShadowingNames,PyUnresolvedReferences
    @classmethod
    def for_urls(cls, *urlpatterns: Union[str, 're.Pattern'], auto: bool = False):
        """!
        Makes a method a request handler for a particular url pattern.
        A url pattern is a string like `"foos/<fooid>/bars/"`, with `"<fooid>"` binding a parameter that you have to add
        to your handler function signature.
        @param urlpatterns: The url patterns to bind the request handler to (strings or compiled `re` regexps).
        @param auto: If to construct the url pattern from the method name.
        """
        def _fct(fct):
            pats = list(urlpatterns)
            if auto:
                u = fct.__name__
                for prefix in ["_do_", "do_"]:
                    if u.startswith(prefix):
                        u = u[len(prefix):]
                        break
                pats.append(f"/{u}")
            fct._urlpats = getattr(fct, '_urlpats', [])
            for pat in pats:
                if isinstance(pat, str):
                    fct._urlpats.append((cls.__compile_re_for_urlpattern(pat), pat))
                else:
                    fct._urlpats.append((pat, pat.pattern))
            return fct
        return _fct

    @classmethod
    def with_param_type(cls, paramname: str, paramtype: type):
        """!
        Defines a parameter of a request handler to be of a particular data type.
        For custom types, either see add_do_param_type_converter() or make your type constructor accept a (str) call.
        Note: You usually should use type annotations on request handlers instead of this function!
        @param paramname: The parameter name. This is the name in the `...&name=...&...` part of the url and of the
                          argument of the request handler.
        @param paramtype: The python type of this parameter. See also Application.add_request_param_type_converter().
        """
        def _fct(fct):
            @functools.wraps(fct)
            def _f_fct(self, **a):
                nonlocal paramtype
                vi = a.get(paramname, None)
                if vi is not None:
                    # noinspection PyUnresolvedReferences,PyProtectedMember
                    if isinstance(paramtype, typing._GenericAlias):  # TODO# nicer with typing.get_args in python 3.8
                        paramtype = paramtype.__args__[0]
                    if (paramtype is datetime.datetime) or (paramtype is datetime.date) or (paramtype is datetime.time):
                        months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
                        if vi:
                            try:
                                vi = datetime.datetime.fromtimestamp(float(vi) / 1000)
                            except ValueError:
                                vi = datetime.datetime(year=int(vi[11:15]), month=months.index(vi[4:7]),
                                                       day=int(vi[8:10]), hour=int(vi[16:18]), minute=int(vi[19:21]),
                                                       second=int(vi[22:24]))
                        else:
                            vi = None
                    elif paramtype is bool:
                        if vi in [True, "1", "true", "True", "yes", "y"]:
                            vi = True
                        elif vi in [False, "0", "false", "False", "no", "n", ""]:
                            vi = False
                        else:
                            raise ValueError(f"not a valid bool: {vi}")
                    else:
                        cnvrtr = self._get_requestparam_type_converter(paramtype)
                        vi = cnvrtr(vi) if cnvrtr else paramtype(vi)
                    a[paramname] = vi
                return fct(self, **a)
            return _f_fct
        return _fct

    @classmethod
    def run_downstream(cls):
        """!
        Defines this request handler to run downstream. In this mode, the browser instantly gets an empty response and
        does not wait longer. This is recommended (and sometimes required; e.g. for user feedback) for operations which
        can take longer.

        Note: This forbids to return any data from the handler, but it does not forbid to modify data sources, to
        set appconfig, and some other ways to transfer results.
        """
        def _fct(fct):
            @functools.wraps(fct)
            def _f_fct(self, **_):
                currentrequest = self.current_request
                # noinspection PyProtectedMember
                currentrequest._set_runs_downstream()
                def nfct(*a, **b):
                    # noinspection PyProtectedMember
                    with currentrequest._in_request():
                        if fct(self, *a, **b) is not None:
                            raise RuntimeError("Downstream request handlers must not return a value.")
                threading.Thread(target=self._callhandler, args=(nfct, currentrequest)).start()
            return _f_fct
        return _fct
