# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

from typing import *
import contextvars
import json
import urllib.parse


class Request:
    """!
    A request from the client, including ways to give the response.
    """

    _current_clientrequest_ctxvar_ = contextvars.ContextVar('_current_clientrequest_ctxvar_', default=None)

    def __init__(self, urlpath: str, header: Dict[str, str], application: 'Application'):
        """!
        @a_infrastructure
        @param urlpath: The complete request url path, including query string.
        @param header: The http request headers.
        """
        self.__urlpath = urlpath
        self.__url = urllib.parse.urlparse(urlpath)
        self.__lparams = {}
        for k, v in urllib.parse.parse_qs(self.__url.query, keep_blank_values=True).items():
            self._set_lparam(k, v)
        self.__header = header
        self.__application = application
        self.__response_headers = {}
        self.__response_body = None
        self.__response_errortext = None
        self.__response_httpcode = None
        self.__scratchpad = {}
        self.__skipprocessing = False
        self.__runs_downstream = False

    # noinspection PyProtectedMember
    @staticmethod
    def current() -> 'Request':
        """!
        Returns the Request associated to the current request (in a request handler).
        """
        return Request._current_clientrequest_ctxvar_.get()

    @property
    def urlpath(self) -> str:
        """!
        The request url's complete path, like `"/foo/bar?id=13&p=a"`.
        """
        return self.__urlpath

    @property
    def urlbarepath(self) -> str:
        """!
        The request url's path without query string, like `"/foo/bar"`.
        """
        return self.__url.path

    @property
    def urlpathquery(self) -> str:
        """!
        The request url's querystring, like `"id=13&p=a"`.
        """
        return self.__url.query

    @property
    def lparams(self) -> Dict[str, List[str]]:
        """!
        The request parameters as list (for having no or more than one value by parameter key).
        """
        return dict(self.__lparams)

    @property
    def params(self) -> Dict[str, str]:
        """!
        The request parameters in a flat structure (not as list, contrary to `oparams`).
        Note: This drops values beyond the first one, so don't use it if you _have_ lists.
        """
        return {k: v[0] for k, v in self.__lparams.items() if len(v) > 0}

    @property
    def header(self) -> Dict[str, str]:
        """!
        The request headers.
        """
        return dict(self.__header)

    @property
    def runs_downstream(self) -> bool:
        """!
        If this request is handled downstream. See reqhandler.RequestHandler.run_downstream().
        """
        return self.__runs_downstream

    @property
    def uieventdata(self) -> Dict[str, Optional[Any]]:
        """!
        The event data dictionary which contains details about an ui event coming from the browser.

        This only makes sense in a request handler for an event binding (see `Application.bindevent()`) and is `None`
        otherwise.
        """
        juieventdata = self.get_param('__yj_uieventdata')
        if juieventdata:
            return json.loads(juieventdata)

    def get_param(self, key: str, defaultval: Optional[str] = None) -> Optional[str]:
        """!
        Returns the value of a request parameter.
        Note: This drops values beyond the first one, so don't use it if you _have_ lists.
        @param key: The parameter key.
        @param defaultval: The default value, if no such parameter exists.
        """
        return self.params.get(key, defaultval)

    def get_param_as_list(self, key: str) -> List[str]:
        """!
        Returns the value of a request parameter as list (for having no or more than one value by parameter key).
        @param key: The parameter key.
        """
        return self.__lparams.get(key, [])

    def _set_lparam(self, key: str, value: List[str]) -> None:
        """!
        Sets a request parameter value. Only infrastructure and middleware would use that.
        @a_infrastructure
        @param key: The parameter key.
        @param value: The parameter values (as list).
        """
        self.__lparams[key] = value

    def _set_runs_downstream(self) -> None:
        """!
        Sets this request to be handled downstream.
        @a_infrastructure
        """
        self.__runs_downstream = True

    @property
    def response_body(self) -> Optional[bytes]:
        """!
        The response body.
        """
        return self.__response_body

    @response_body.setter
    def response_body(self, value: Optional[bytes]):
        self.__response_body = value

    @property
    def response_errortext(self) -> Optional[str]:
        """!
        The error text (if an error occurred).
        """
        return self.__response_errortext

    @response_errortext.setter
    def response_errortext(self, value: Optional[str]):
        self.__response_errortext = value

    @property
    def response_httpcode(self) -> Optional[int]:
        """!
        The http response code.
        """
        return self.__response_httpcode

    @response_httpcode.setter
    def response_httpcode(self, value: Optional[int]):
        self.__response_httpcode = value

    def set_response_header(self, key: str, value: str) -> None:
        """!
        Sets an http response header entry.
        @param key: The header key.
        @param value: The header value.
        """
        self.__response_headers[key] = value

    def get_response_header(self, key: str) -> str:
        """!
        Returns the http response header entry as stored before with set_response_header().
        @param key: The header key.
        """
        return self.__response_headers.get(key, "")

    def get_response_header_keys(self) -> List[str]:
        return [k for k, v in self.__response_headers.items() if v]

    @property
    def scratchpad(self) -> Dict[str, Any]:
        """!
        A dictionary for storing some custom things, typically by middleware.
        """
        return self.__scratchpad

    @property
    def application(self) -> 'Application':
        """!
        The Application with got this request.
        """
        return self.__application

    @property
    def skip_processing(self) -> bool:
        """!
        If to skip further request processing. Used by middleware.
        @a_infrastructure
        """
        return self.__skipprocessing

    @skip_processing.setter
    def skip_processing(self, value: bool):
        self.__skipprocessing = value

    def _in_request(self) -> ContextManager:
        """!
        Returns a context manager for setting the current request.
        @a_infrastructure
        """
        return Request.__InRequest(self)

    class __InRequest:
        """!
        Context manager for setting the current request.
        @a_infrastructure
        """

        def __init__(self, request: 'Request'):
            self.__request = request
            self.__curreq_token = None

        def __enter__(self):
            self.__curreq_token = Request._current_clientrequest_ctxvar_.set(self.__request)

        def __exit__(self, exc_type, exc_val, exc_tb):
            Request._current_clientrequest_ctxvar_.reset(self.__curreq_token)


if TYPE_CHECKING:  # ugly hack for type annotations
    from .app import *
