# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for building with gnu autotools.
"""

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

buildmake = anise.framework.features.loadfeature("build.make").featuremodule


class internals:
    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION, requires=["milieus"])
    def _initproject():
        universe.milieus.register("autotools", AutotoolsMilieuComposite)


class AutotoolsMilieuComposite(buildmake.MakeMilieuComposite):
    """!
    Build environment composite for autotools builds.

    See milieus.MilieuComposite for general infos.
    """

    def __init__(self, autoreconfpath="autoreconf", makepath="make", processcount=1):
        """!
        @param autoreconfpath: The path to the `autoreconf` tool.
        @param makepath: The path to the `make` tool.
        @param processcount: The maximum number of concurrent processes to execute.
        """
        buildmake.MakeMilieuComposite.__init__(self, makepath, processcount)
        self.autoreconfpath = autoreconfpath
        self.autoreconfcmd = ["{autoreconfpath}", "[[args]]"]

    def call_autoreconf(self, *args, raise_on_errors="unable to run autoreconf", **kwargs):
        cmd = [x.format(**self.__dict__) for x in self.autoreconfcmd]
        i = cmd.index("[[args]]")
        cmd = cmd[:i] + list(args) + cmd[i + 1:]
        return self.call(*cmd, raise_on_errors=raise_on_errors, **kwargs)


def build(source, milieu=None, autoreconf_args=(), configure_args=(), make_args=(), makeinstall_args=()):
    """!
    Builds with autotools (and autoreconf).
    @param source: The anise.framework.files.Filestructure to build.
    @param milieu: The milieu to use. See milieus.Milieu for details.
    @param autoreconf_args: Additional arguments for the 'autoreconf' call (as list of strings).
    @param configure_args: Additional arguments for the './configure' call (as list of strings).
    @param make_args: Additional arguments for the 'make' call (as list of strings).
    @param makeinstall_args: Additional arguments for the 'make install' call (as list of strings).
    @return: A anise.framework.files.Filestructure.
    """
    with universe.executionscope.newsubscope("Building via autotools"):
        milieu = universe.milieus.get(milieu, defaultmilieu="autotools")
        sourcepath = milieu.createtempdir().path()
        source.dl(to=sourcepath)
        result = milieu.createtempdir()
        with anise.utils.basic.ChDir(sourcepath):
            milieu.call_autoreconf("--install", *autoreconf_args)
            milieu.call("./configure", *configure_args, raise_on_errors="unable to run ./configure")
            milieu.call_make(*make_args, raise_on_errors="unable to run make")
            milieu.call_make("DESTDIR="+result.path(), *makeinstall_args, "install",
                             raise_on_errors="unable to run 'make install'")
        return result
