# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for compiling binaries with qmake.
"""

import os
import re
import shutil

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

make = anise.framework.features.loadfeature("build.make").featuremodule


class internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION, requires=["milieus"])
    def _initproject():
        universe.milieus.register("qt", QMakeMilieuComposite)


class QMakeMilieuComposite(make.MakeMilieuComposite):
    """!
    Build environment composite for Qt and qmake builds.
    
    See milieus.MilieuComposite for general infos.
    """

    def __init__(self, qtenv=None, qmakepath=None, bindir=None, pluginsdir=None, datadir=None, makepath="make"):
        """!
        @param qtenv: Optionally choose an qt environment by name (use `qtchooser -l` for a list).
        @param qmakepath: The path to the Qt qmake tool.
        @param bindir: The Qt bin directory.
        @param pluginsdir: The Qt plugins directory.
        @param datadir: The Qt data directory.
        @param makepath: The path to the make tool.
        """
        make.MakeMilieuComposite.__init__(self, makepath=makepath)
        self.qtenv = qtenv
        self.qmakepath = qmakepath
        self.bindir = bindir
        self.pluginsdir = pluginsdir
        self.datadir = datadir
        self.qmakecmd = ["{qmakepath}", "[[args]]"]

    def initialize(self):
        queryoutput = None
        if None in [self.bindir, self.pluginsdir, self.datadir, self.qmakepath]:
            if self.qtenv:
                if self.call("qtchooser", "-l")[0] != 0:
                    raise anise.framework.exceptions.RequirementsMissingError("'qtchooser' seems to be unavailable")
                r2, queryoutput = self.call("qtchooser", "-qt=" + self.qtenv, "-run-tool=qmake", "-query")
                if r2 != 0:
                    raise anise.framework.exceptions.RequirementsMissingError("unable to find qt environment " + self.qtenv)
            if not queryoutput:
                self.qmakepath = self.qmakepath or "qmake"
                r2, queryoutput = self.call(self.qmakepath, "-query")
                if r2 != 0:
                    raise anise.framework.exceptions.RequirementsMissingError("unable to query " + self.qmakepath)
        def getfromqueryoutput(k):
            m = re.search("^{}:(.*)$".format(k), queryoutput, flags=re.MULTILINE)
            if m:
                return m.group(1)
        self.bindir = self.bindir or getfromqueryoutput("QT_INSTALL_BINS")
        self.pluginsdir = self.pluginsdir or getfromqueryoutput("QT_INSTALL_PLUGINS")
        self.datadir = self.datadir or getfromqueryoutput("QT_INSTALL_DATA")
        self.qmakepath = self.qmakepath or (self.bindir + "/qmake")

    def call_qmake(self, *args, **kwargs):
        """!
        Calls the qmake tool.
        """
        cmd = [x.format(**self.__dict__) for x in self.qmakecmd]
        i = cmd.index("[[args]]")
        cmd = cmd[:i] + list(args) + cmd[i+1:]
        return self.call(*cmd, **kwargs)


def basedependencies():
    """!
    Returns a list of dependencies, which are typical for Qt projects.
    @return: List of dependencies.Dependency.
    """
    dependencies = anise.framework.features.loadfeature("dependencies").featuremodule
    return [dependencies.Dependency(dependencies.Type.Required, debian=["libqt5concurrent5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5core5a"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5gui5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5help5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5network5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5printsupport5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5qml5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5quick5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5script5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5sql5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5sql5-sqlite"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5svg5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5webkit5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5widgets5"]),
            dependencies.Dependency(dependencies.Type.Required, debian=["libqt5xml5"])]


def build(project, source, milieu=None, targetname=None, targetfile=None, qmakeoptions=None):
    """!
    Builds with qmake.
    @param project: The qmake project file path relative to `source`s root.
    @param source: The anise.framework.files.Filestructure to build.
    @param milieu: The milieu to use. See milieus.Milieu for details.
    @param targetname: The name of the make target for creating.
    @param targetfile: The file name of the target file to create.
    @param qmakeoptions: Additional options for running 'qmake'.
    @return: A anise.framework.files.Filestructure.
    """
    with universe.executionscope.newsubscope("Building '{}' via qmake".format(targetname) if targetname
                                             else "Building via qmake"):
        milieu = universe.milieus.get(milieu, defaultmilieu="qt")
        if milieu.call_qmake("-v")[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'qmake' seems to be unavailable")
        if qmakeoptions is None:
            qmakeoptions = []
        with milieu.createtempdir() as tempdir:
            with anise.utils.basic.ChDir(tempdir.path()):
                sourcepath = milieu.createtempdir().path()
                source.dl(to=sourcepath)
                milieu.call_qmake(*(["-makefile"] + qmakeoptions + [sourcepath + "/" + project]),
                                    raise_on_errors="unable to run qmake")
                milieu.make(targetname)
                if targetfile is None:
                    targetfile = targetname
                targetpath = milieu.createtempdir().path()
                result = anise.framework.files.Filestructure(targetpath + "/" + os.path.basename(targetfile))
                if os.path.isdir(targetfile):
                    anise.framework.files.Filestructure(targetfile).dl(to=result.path())
                else:
                    shutil.copy2(targetfile, targetpath)
        return result
