# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for injecting some kinds of data at anise runtime into some file structures, in order to make it available
to other build processes (Makefiles, ...) or to the project's runtime.
"""

import datetime

import anise.framework.features
import anise.framework.files
import anise.utils
import anise.framework.projectinformations

from anise.framework.projects import universe


class internals:

    @staticmethod
    def _sync():
        """!
        Sync if not marked as `omitsyncvcs`.
        """
        if not getattr(universe, "omitsyncvcs", False):
            universe.versioncontrol.current.syncversioncontrolsystem(
                commitmessage="automated commit for data injections")
        return True

    class Pool:
        """
        Storage of all data injections.
        """

        def add(self, method, outfile, propnames=None, do_in_source=False, **params):
            """!
            Add a data injection.
            @param method: The implementation for the injection. This is the name of a function with the parameters
                           'outfile' and 'data', writing the data (a dict with key/value pairs) to a given file in some
                           format (as absolute path or relative to the project root directory).
            @param outfile: Path to the target file for the injections (relative to anise project root path).
            @param propnames: List of entry names you want to inject (beyond the few default ones).
            @param do_in_source: If the source itself should be updated as well (instead of only the on-the-fly copies 
                                 for packages).
            @param params: Additional parameters for calling the injection implementation.
            """
            while outfile.startswith("/"):
                outfile = outfile[1:]
            f = lambda x: method(outfile=x+outfile, data=getessentials(propnames), **params)
            if do_in_source:
                def _release():
                    with universe.executionscope.newsubscope("Injecting data to '{}'".format(outfile)):
                        internals._sync()
                        f("")
                universe.addhook(anise.framework.features.loadfeature("releasing").featuremodule.HOOK_RELEASE,
                                 _release, provides=["PRE", "datainjections"])
            else:
                def _enrich(destdir):
                    with universe.executionscope.newsubscope("Injecting data to '{}'".format(outfile)):
                        f(destdir+"/")
                universe.addhook(anise.framework.features.loadfeature("packages").featuremodule.HOOK_ENRICHRAWPACKAGE,
                                 _enrich)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.datainjections.pool = internals.Pool()


def getessentials(propnames=None):
    """!
    Returns essential project data for injecting it into some file.
    @param propnames: Additional properties to inject on top of the very fundamental ones.
    @return: a list of name/value tuples containing the data for injections.
    """
    data = []
    data.append(("version", universe.packages.getversion()))
    data.append(("buildtimeutc", datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")))
    if universe.base.websiteurl:
        data.append(("homepage", universe.base.websiteurl))
    for propname in propnames or []:
        propvalue = getattr(universe, propname, None)
        data.append((propname, propvalue))
    return data


def inject_python(outfile=None, data=None, asclass=None):
    """!
    Injects project data into a Python file, so it is available at program's runtime.
    @param outfile: The target file for injection.
    @param data: The data to inject.
    @param asclass: Optional class name. If set, the data is written as members into a class with this name.
    """
    result = """
# This file is automatically created by 'anise' in order to make some
# information available at runtime. Do not change stuff here.
#   Information about anise:
#   {anisehomepage}
""".format(anisehomepage=anise.framework.projectinformations.homepage)[1:]
    if asclass:
        result += "class " + asclass + ":\n"
        indent = "    "
    else:
        result += ""
        indent = ""
    # determine destination
    if outfile:
        if outfile.startswith("/"):
            fulloutfile = outfile
        else:
            fulloutfile = universe.specialpaths.projectdir + "/" + outfile
    else:
        tempdir = anise.framework.files.TempDir()
        fulloutfile = tempdir.path() + "/out.py"
    # write them to outfile
    for propname, propvalue in data or []:
        propvalue = repr(propvalue)
        result += "{indent}{propname} = {propvalue}\n".format(**locals())
    anise.utils.basic.writetofile(fulloutfile, result)
    return anise.framework.files.Filestructure(fulloutfile)


def inject_c(outfile=None, data=None):
    """!
    Injects project data into a C++ header file, so it is available at program's runtime.
    @param outfile: The target file for injection.
    @param data: The data to inject.
    """
    result = """
// This file is automatically created by 'anise' in order to make some
// information available at runtime. Do not change stuff here.
//   Information about anise:
//   {anisehomepage}
#ifndef ANISE_PROJECT_INFORMATION_H
#define ANISE_PROJECT_INFORMATION_H
""".format(anisehomepage=anise.framework.projectinformations.homepage)[1:]
    # determine destination
    if outfile:
        if outfile.startswith("/"):
            fulloutfile = outfile
        else:
            fulloutfile = universe.specialpaths.projectdir + "/" + outfile
    else:
        tempdir = anise.framework.files.TempDir()
        fulloutfile = tempdir.path() + "/out.h"
    # write them to outfile
    for propname, propvalue in data or []:
        propvalue = str(propvalue)
        result += "const char* {propname} = \"{propvalue}\";\n".format(**locals())
    result += "#endif\n"
    anise.utils.basic.writetofile(fulloutfile, result)
    return anise.framework.files.Filestructure(fulloutfile)
