# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for defining and processing distributables (installation packages for target systems, ...).
"""

import datetime
import hashlib
import os

import anise.framework.features
import anise.framework.report

from anise.framework.projects import universe


HOOK_PREPARE_DISTRIBUTABLES_POOL = anise.framework.features.Hook(
    doc="This hook is called just before the distributables pool is prepared for further processing.")


class internals:

    class Pool:
        """!
        Storage of all dependencies.
        """

        def __init__(self):
            self.list = []
            self.initialized = False

        def addgroup(self, sources, **params):
            """!
            Adds a group of distributables.

            Each group is stored as FileGroup instance.
            @param sources: The generator tasks, which returns the distributables.
            @param params: Additional parameters for the generator tasks.
            """
            self.list.append(FileGroup(sources, **params))

        def find_by_name(self, name):
            """!
            Searches a FileGroup by name and returns it.
             
            @param name: The name of the group to find.
            """
            for x in self.list:
                if x.name == name:
                    return x

        def get(self):
            """!
            Prepares the pool for further processing (i.e. firing HOOK_PREPARE_DISTRIBUTABLES_POOL) and returns the 
            pool.
            """
            if not self.initialized:
                for fct in universe.gethooks(HOOK_PREPARE_DISTRIBUTABLES_POOL):
                    fct()
                self.initialized = True
            return self.list

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.distributables.pool = internals.Pool()
        universe.distributables.currentgroup = None
        universe.distributables.homepagesectionbegin = \
            "The following packages are available. Choose the one which matches your operating system or environment. " \
            "Please also check for :ref:`requirements<dldependencies>`.\n\n"

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _add_homepage_section():

        def my_homepage_section(homepage, sourcedir, backgroundimage, *a):
            distributablespool = universe.distributables.pool.get()
            if len(distributablespool) > 0:
                downloads = ""
                _downloads = universe.distributables.homepagesectionbegin
                for dlgroup in distributablespool:
                    universe.distributables.currentgroup = dlgroup
                    __downloads = ""
                    __downloads += f"**{dlgroup.name}**\n\n"
                    grouphasvisibleitems = False
                    dlgdesc = dlgroup.description or ""
                    for fn in dlgroup.sources:
                        if fn.getdata("invisible_on_homepage", False):
                            d = fn.getdata("description", None)
                            if d and (d not in dlgdesc):
                                dlgdesc += "\n" + d
                    if dlgdesc:
                        grouphasvisibleitems = True
                        __downloads += dlgdesc.encode('ascii', 'xmlcharrefreplace').decode() + "\n\n"
                    for ifn, fn in enumerate(dlgroup.sources):
                        fn.dl(to=homepage.path())
                        if not fn.getdata("invisible_on_homepage", False):
                            grouphasvisibleitems = True
                            bn = os.path.basename(fn.path())
                            hasher = hashlib.sha256()
                            fd = open(fn.path(), "rb")
                            isize = 0.0
                            for line in fd.readlines():
                                hasher.update(line)
                                isize += len(line)
                            fd.close()
                            adddesc = fn.getdata("description", None)
                            if adddesc:
                                __downloads += adddesc.encode('ascii', 'xmlcharrefreplace').decode() + "\n\n"
                            ctime = fn.getdata("ctime", None)
                            if ctime:
                                ctime = datetime.datetime.fromtimestamp(float(ctime))
                            else:
                                ctime = datetime.datetime.now()
                            ssize = "B"
                            if isize > 1024:
                                isize = isize / 1024.0
                                ssize = "KB"
                            if isize > 1024:
                                isize = isize / 1024.0
                                ssize = "MB"
                            if isize > 1024:
                                isize = isize / 1024.0
                                ssize = "GB"
                            if isize > 1024:
                                isize = isize / 1024.0
                                ssize = "TB"
                            ssize = "{0:.1f}".format(isize) + " " + ssize
                            __downloads += (f":File: `{bn} <{bn}>`_\n"
                                            f":Created: {ctime.strftime('%Y-%m-%d %H:%M')}\n"
                                            f":Sha256sum: :samp:`{hasher.hexdigest()}`\n"
                                            f":Size: {ssize}\n\n.. rst-class:: dummy__\n\n")
                        if dlgroup.linkto is not None and ifn == 0:  # only do that for the first file
                            universe.addhook(
                                anise.framework.features.loadfeature("homepage").featuremodule.
                                HOOK_UPLOAD_HOMEPAGE_COLLECTLINKS,
                                (lambda x: (lambda: x))([(bn, dlgroup.linkto)]))
                    if grouphasvisibleitems:
                        downloads += _downloads
                        downloads += __downloads
                        _downloads = ""
                universe.distributables.currentgroup = None
                if downloads:
                    deptxtfct = universe.homepage.sections.gettextfct("dependencies")
                    if deptxtfct:  # if there is a 'dependencies' section, we try to include it here
                        rst_heading = universe.docrender.internals.rst_heading
                        downloads += (f"{rst_heading('Dependencies', variant=':', anchor='dldependencies')}\n"
                                      ) + deptxtfct(homepage, sourcedir, backgroundimage, *a)
                        universe.homepage.sections.replacetextfct("dependencies", lambda *a, **b: None)
                    return downloads

        universe.homepage.sections.add(4000, "download", "Download", my_homepage_section, important=True)


class FileGroup:
    """!
    A file group.
    """

    def __init__(self, sources, name, description=None, linkto=None):
        """!
        @param sources: List of task implementations which generates the download group content.
                        For a single entry, you may also pass it directly without a list around it.
        @param name: The file group name.
        @param description: The file group description text.
        @param linkto: An optional link path, which is always kept up to date (pointing to the latest version). This
                       is only applied to the first file in the group.
        """
        self.sources = sources if hasattr(sources, "__iter__") else [sources]
        self.name = name
        self.description = description
        self.linkto = linkto
