# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for generating all kinds of documentation like output.
"""

from typing import *
import base64
import colorsys
import datetime
import json
import os
import subprocess

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.framework.report
import anise.utils

from anise.framework.projects import universe


# TODO docu


# noinspection PyPep8Naming
class internals:

    @staticmethod
    def rst_heading(s: str, *, sub: bool = False, variant: str = '=', anchor: Optional[str] = None) -> str:
        ll = (variant * len(s))
        return "\n\n" + (f".. _{anchor}:\n\n" if anchor else '') + ('' if sub else f'{ll}\n') + f'{s}\n{ll}\n\n'

    class ShortSnippets:#TODO
        """!
        Storage of text snippets for use in the documentation texts like `@a_foo`.
        """

        def __init__(self):
            self.storage = {}
            self.add("dooverride",
                     "Override this method in custom subclasses.")
            self.add("dooverride_optional",
                     "Override this method in custom subclasses or leave the default implementation.")
            self.add("infrastructure",
                     "This is part of a particular piece of internal infrastructure and is typically not required "
                     "to be used.")

        def add(self, name, text):
            """!
            Adds a text snippet to the storage, so it can be used.
            @param name: The snippet name. This is the `foo` in `@a_foo`.
            @param text: The snippet text.
            """
            self.storage[name] = text

        def toaliases(self):
            ast = "@par\n<div class='anisedochint'>{}</div>\n\n"
            return {"a_"+k: ast.format(self.storage[k].replace("&", "&amp;").replace("<", "&lt;").
                                       replace(">", "&gt;")) for k in self.storage}

    @staticmethod
    def scalecolor(c, brightness, saturation=None):
        """!
        Internally used for scaling a color tuple to a new one with the same hue but a given brightness.
        """
        h, l, s = colorsys.rgb_to_hls(*c)
        if saturation is not None:
            s *= saturation
            if s > 1:
                s = 1.0
        return colorsys.hls_to_rgb(h, brightness, s)


class AbstractApiReferenceLanguage:

    def _sphinx_extensions(self) -> List[str]:
        return []

    def _sphinx_config(self, path: str) -> str:
        return ""

    def _sphinx_transfer(self, srcpath: str, outpath: str, heading: str) -> None:
        anise.framework.files.Filestructure(srcpath).dl(to=outpath)


class Python3ApiReferenceLanguage(AbstractApiReferenceLanguage):

    def _sphinx_extensions(self):
        return ['sphinx.ext.autodoc']

    def _sphinx_transfer(self, srcpath, outpath, heading):
        subprocess.check_call(['sphinx-apidoc', '-o', outpath, srcpath])
        with open(f'{outpath}.rst', 'w') as f:
            f.write(f"""{internals.rst_heading(heading)}
.. toctree::
   :glob:

   {os.path.basename(outpath)}/*
""")


class CppApiReferenceLanguage(AbstractApiReferenceLanguage):

    def _sphinx_extensions(self):
        return ['breathe']

    def _sphinx_config(self, path):
        return f"breathe_projects = {{ {repr(os.path.basename(path))}: {repr(path)}, }}"

    def _sphinx_transfer(self, srcpath, outpath, heading):
        super()._sphinx_transfer(srcpath, outpath, heading)
        with open(f'{outpath}.rst', 'w') as f:
            f.write(f"""{internals.rst_heading(heading)}
.. doxygenclass:: sh::filesystem::Eurl
   :project: {repr(os.path.basename(srcpath))}
   :members:
""")


class JavaScriptApiReferenceLanguage(AbstractApiReferenceLanguage):

    def _sphinx_extensions(self):
        return ['sphinx_js']

    def _sphinx_config(self, path):
        return f"js_source_path = {repr(path)}"

    def _sphinx_transfer(self, srcpath, outpath, heading):
        super()._sphinx_transfer(srcpath, outpath, heading)
        with open(f'{outpath}.rst', 'w') as f:
            f.write(f"""{internals.rst_heading(heading)}
.. js:autoclass:: YajiClient
   :members:   
""")


class AbstractPiece:

    def __init__(self, name: str, rootpath: str):
        self.__name = name
        self.__rootpath = rootpath

    @property
    def name(self):
        return self.__name

    @property
    def rootpath(self):
        return self.__rootpath

    def copyto(self, path: str) -> None:
        anise.framework.files.Filestructure(path=self.rootpath).dl(to=f'{path}/{self.name}')

    def _sphinx_extensions(self) -> List[str]:
        return []

    def _sphinx_config(self, path: str) -> str:
        return ""


class RstTextsPiece(AbstractPiece):

    def __init__(self, name: str, rootpath: str):
        super().__init__(name, rootpath)


class ApiReferencePiece(AbstractPiece):

    def __init__(self, name: str, rootpath: str, language: AbstractApiReferenceLanguage, *,
                 heading: Optional[str] = None):
        super().__init__(name, rootpath)
        self.__language = language
        self.__heading = heading or "API reference"

    def copyto(self, path):
        # noinspection PyProtectedMember
        self.__language._sphinx_transfer(self.rootpath, f'{path}/{self.name}', self.__heading)

    def _sphinx_extensions(self) -> List[str]:
        # noinspection PyProtectedMember
        return self.__language._sphinx_extensions()

    def _sphinx_config(self, path):
        # noinspection PyProtectedMember
        return self.__language._sphinx_config(f'{path}/{self.name}')


class AbstractOutputSpec:

    def __init__(self, path: str, formatname: str):
        self.path = path
        self.formatname = formatname

    def _sphinx_config(self, path: str) -> str:
        return ""

    def resultpath(self, projectname: str) -> str:
        raise NotImplementedError()


class HtmlOutputSpec(AbstractOutputSpec):

    def __init__(self, path: str, *, theme: Optional[str] = None,
                 title: Optional[str] = None, short_title: Optional[str] = None):
        super().__init__(path, 'html')
        self.__theme = theme or "alabaster"
        self.__title = title
        self.__short_title = short_title

    def _sphinx_config(self, path: str) -> str:
        res = f"html_theme = {repr(self.__theme)}\n"
        if self.__short_title or self.__title:
            res += f"html_short_title = {repr(self.__short_title or self.__title)}\n"
        if self.__title:
            res += f"html_title = {repr(self.__title)}\n"
        return res

    def resultpath(self, projectname):
        return "html"


class PdfOutputSpec(AbstractOutputSpec):

    def __init__(self, path: str):
        super().__init__(path, 'latexpdf')

    def resultpath(self, projectname):
        return f"latex/{projectname}.pdf"


class ManpageOutputSpec(AbstractOutputSpec):

    def __init__(self, path: str):
        super().__init__(path, 'man')

    def resultpath(self, projectname):
        return "man"


class PlaintextOutputSpec(AbstractOutputSpec):

    def __init__(self, path: str):
        super().__init__(path, 'text')

    def resultpath(self, projectname):
        return "text"


def render(pieces: List[AbstractPiece], outputspec: AbstractOutputSpec,
           projectname: str, *, author: str, heading: Optional[str] = None, mainbody: Optional[str] = None,
           basecolor: Optional[Tuple[float, float, float]] = None,
           exclude_in_toc: Optional[List[str]] = None, include_in_toc: Optional[List[str]] = None) -> None:
    with anise.framework.files.TempDir() as tdir:
        tf = tdir.path()
        extensions = ['sphinx.ext.autosummary', 'sphinx.ext.inheritance_diagram']
        sconfig = ""
        sls = ""
        for piece in pieces:
            piece.copyto(tf)
            # noinspection PyProtectedMember
            extensions += piece._sphinx_extensions()
            # noinspection PyProtectedMember
            sconfig += piece._sphinx_config(tf) + "\n"
            if piece.name not in (exclude_in_toc or []) and (include_in_toc is None or (piece.name in include_in_toc)):
                sls += f"   {piece.name}\n"
        # noinspection PyProtectedMember
        sconfig += outputspec._sphinx_config(tf) + "\n"
        html_theme_options = {}
        basecolor = basecolor or (.1, .2, .3)
        for bcaa in range(10):
            for bcab in range(10):
                for bcac in range(10):
                    sbcac = "abcdefghijk"[bcac]
                    r, g, b = internals.scalecolor(basecolor, (bcab+1)/10, (bcac+1)/10)
                    html_theme_options[f'brandingcolor_{sbcac}{bcaa}{bcab}'] = (f"rgb({r*255},{g*255},{b*255},"
                                                                                f"{(bcaa+1)/10})")
        with open(f'{tf}/conf.py', 'w') as f:
            f.write(f"""
import base64, json, sys
tf = {repr(tf)}
sys.path.insert(0, tf)
project = {repr(projectname)}
author = {repr(author)}
#TODO version, release
#TODO rst_epilog, rst_prolog
#TODO nitpicky
extensions = {repr(extensions)}
copyright = "{datetime.datetime.now().year}, " + author
exclude_patterns = ['.directory', 'Thumbs.db', '.DS_Store', '.git', '.svn', '.piget', '.idea']
html_theme_path = [{repr(f'{os.path.dirname(__file__)}/docrender/__static')}]
html_theme_options = json.loads(base64.b64decode({base64.b64encode(json.dumps(html_theme_options).encode())}))
html_sidebars = {{ '**': ['globaltoc.html', 'searchbox.html'] }}
template_ = ''
{sconfig}
""")
        with open(f'{tf}/index.rst', 'w') as f:
            f.write(f"""{internals.rst_heading(heading, variant='_') if heading else ''}
.. toctree::
   :maxdepth: 3
   :glob:

{sls}

{mainbody}
""")
        with anise.framework.files.TempDir() as tbdir:
            anise.framework.files.Filestructure(f"{universe.specialpaths.anisedatadir}/icons").dl(to=f"{tf}/icons")
            subprocess.check_call(["sphinx-build", '-M', outputspec.formatname, tf, tbdir.path()])
            tbdir.dl(subpath=outputspec.resultpath(projectname), to=outputspec.path)
            """TODO
        r, o = anise.utils.basic.call("doxygen", tpath + "/cfg",
                                          raise_on_errors="unable to generate content via doxygen")
        anise.utils.logging.logdebug(o)
"""