# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for specifying and processing project documentation.
"""

import colorsys
import math
import os
import re
import shlex
import shutil

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.framework.report
import anise.utils

from anise.framework.projects import universe


class internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.documentation.exports = internals.Exports()
        universe.documentation.pool = internals.Pool()
        universe.documentation.snippets = internals.ShortSnippets()
        universe.documentation.readme_include_developer_doc = False
        universe.documentation.htmltheme = None
        universe.documentation.generate_readme = True

        def _writedocumentations(destdir):
            with universe.executionscope.newsubscope("Adding documentation"):
                internals.writedocumentations(destdir=destdir, throw_on_existing=True, fortype=PackageExport)
        universe.addhook(anise.framework.features.loadfeature("packages").featuremodule.HOOK_ENRICHRAWPACKAGE,
                         _writedocumentations)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION,
                                          provides=["_compose_readmes"], requires=["base"])
    def _compose_readmes():

        if not universe.documentation.generate_readme:
            return

        def token_in_map(t, m):
            for mk, ms in m:
                if mk == t:
                    return ms
            return None

        maturity = universe.base.getprojectstatusdescription()
        smaturity = (universe.docrender.internals.rst_heading('Maturity') + maturity) if maturity else ""
        dincludedeveloperdoc = universe.documentation.readme_include_developer_doc
        dkwargs = {}
        if isinstance(dincludedeveloperdoc, dict):
            dkwargs = dincludedeveloperdoc
            dincludedeveloperdoc = True
        readmedefaultloc = universe.specialpaths.projectdir + "/_meta/readme.src"
        if os.path.isfile(readmedefaultloc):
            readmetxt = anise.utils.basic.readfromfile(readmedefaultloc, onestring=True)
            universe.documentation.exports.add(HomepageExport(key="FULLREADME", heading="Manual",
                                                              includedeveloperdoc=dincludedeveloperdoc, **dkwargs))
        else:
            readmetxt = ""
        if not token_in_map("FULLREADME", universe.documentation.pool.list):
            _lauthors = getattr(universe.base, "authors", [])
            _authors = " and ".join(([", ".join(_lauthors[0:-1])] if len(_lauthors) > 1 else []) + _lauthors[-1:])
            if getattr(universe.base, "longdescription", ""):
                universe.documentation.pool.add("_LONGDESCRIPTION", universe.base.longdescription)
                _slongdesc = (universe.docrender.internals.rst_heading('About') + "TODO copydoc _LONGDESCRIPTION")
            else:
                _slongdesc = ""

            def fulltxt():
                _sversion = getattr(universe.packages, "getversion", lambda: None)()
                if _sversion:
                    _sversion = "You are currently reading the manual for version {}.".format(_sversion)
                else:
                    _sversion = ""
                if universe.base.license:
                    _slicense = " under the terms of the {}".format(universe.base.license.name)
                else:
                    _slicense = ""
                return f"""
{universe.docrender.internals.rst_heading('License')}

{universe.name} is written by {_authors}{_slicense}.

Please read the `LICENSE` file from the package and the :ref:`Dependencies<hhdependencies>` section for included
third-party stuff.

{_slongdesc}

{universe.docrender.internals.rst_heading('Up-to-date?')}

Are you currently reading from another source than the homepage? Are you in doubt if that place
is up-to-date? If yes, you should visit {universe.base.websiteurl or 'the homepage'} and check that. {_sversion}

{smaturity}

.. _hhdependencies:

{universe.docrender.internals.rst_heading('Dependencies')}

{universe.dependencies.getdependencydocument()}

{readmetxt}
                    """
            universe.documentation.pool.add("FULLREADME", fulltxt)
        heading = universe.name[0].upper() + universe.name[1:] + " Manual"
        if not os.path.exists(universe.specialpaths.projectdir+"/README"):
            universe.documentation.exports.add(PackageExport(key="FULLREADME", outputfile="/README",
                                                             format=ExportFormat.Plaintext, heading=heading,
                                                             includedeveloperdoc=dincludedeveloperdoc, **dkwargs))
        if not os.path.exists(universe.specialpaths.projectdir+"/README.pdf"):
            universe.documentation.exports.add(PackageExport(key="FULLREADME", outputfile="/README.pdf",
                                                             format=ExportFormat.PDF, heading=heading,
                                                             includedeveloperdoc=dincludedeveloperdoc, **dkwargs))

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION, requires=["_compose_readmes"])
    def _add_homepage_section():

        def my_homepage_section(homepage, sourcedir, backgroundimage, *a):
            documentation = "The following documentation is available:\n\n"
            hpexports = [x for x in universe.documentation.exports.list if type(x) is HomepageExport]
            hpexports = sorted(hpexports, key=lambda x: x.index)
            for echan in hpexports:
                os.mkdir(homepage.path() + "/" + echan.key)
                elinktarget = echan.linktarget if echan.linktarget else (echan.key + ".html")
                echan.source = anise.framework.files.Filestructure(sourcedir)
                echan.generate(homepage.path() + "/" + echan.key)
                documentation += f"`{echan.heading} <{echan.key}/{elinktarget}>`_\n\n"
            if len(hpexports) > 0:
                return documentation

        universe.homepage.sections.add(3000, "documentation", "Documentation", my_homepage_section)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _add_release_task_for_export_to_source():
        if len([x for x in universe.documentation.exports.list if isinstance(x, SourceExport)]) > 0:
            def create_documentations_in_source():
                internals.writedocumentations(universe.specialpaths.projectdir, throw_on_existing=False,
                                              fortype=SourceExport)
            universe.releasing.tasks.add(create_documentations_in_source)

    class Pool:
        """!
        Storage of all documentation text generators.
        """

        def __init__(self):
            self.list = []

        def add(self, key, src):
            """!
            Adds a documentation.
            @param key: The documentation key for storage (later used for referencing this text).
            @param src: The documentation source text (or a function which returns one).
            """
            self.list.append([key, src])

        def gettexts(self):
            """!
            Returns all texts.
            @return: A list of key/value pairs. Both are always strings.
            """
            return [[x[0], (x[1]() if callable(x[1]) else x[1])] for x in self.list]

        def load_anisedoc_source(self, src):
            """!
            Loads the content of anisedoc sources (which contain ANISEDOC tags) into the documentation store.
            @param src: File or directory with anisedoc sources.
            """
            if os.path.isdir(src):
                for s in os.listdir(src):
                    self.load_anisedoc_source(src + "/" + s)
            elif os.path.isfile(src):
                currentprefix = None
                currentdoc = None
                currentcontent = ""

                def end_of_doc():
                    nonlocal currentdoc, currentcontent
                    exists = False
                    for entry in self.list:
                        if entry[0] == currentdoc:
                            entry[1] += currentcontent
                            exists = True
                            break
                    if not exists:
                        self.add(currentdoc, currentcontent)
                    currentdoc = None
                    currentcontent = ""

                for line in anise.utils.basic.readfromfile(src, onestring=False):
                    if currentdoc:
                        if line.startswith(currentprefix):
                            currentcontent += line[len(currentprefix):] + "\n"
                        else:
                            end_of_doc()
                    else:
                        i = line.find("ANISEDOC ")
                        if i == 0:
                            anise.utils.logging.logwarning(
                                "skipped an ANISEDOC tag because it has a null-length prefix")
                        elif i > -1:
                            currentprefix = line[0:i]
                            if not re.search("[A-Za-z0-9]", currentprefix):
                                currentdoc = line[i + 9:].strip()
                if currentdoc:
                    end_of_doc()

    class Exports:
        """!
        Storage of all documentation export definitions (i.e. where and how documentation products are to be created).
        """

        def __init__(self):
            self.list = []

        def getbyname(self, name):
            """!
            Returns all export definitions with a given name.
            @param name: A export definition name.
            @return: List of export definitions.
            """
            return [x for x in self.list if x.name == name]

        def add(self, export):
            """!
            Exports a documentation to some targets.
            @param export: A export definition which specify an documentation export target (see Export).
            """
            self.list.append(export)

    class ShortSnippets:#TODO
        """!
        Storage of text snippets for use in the documentation texts like `@a_foo`.
        """

        def __init__(self):
            self.storage = {}
            self.add("dooverride",
                     "Override this method in custom subclasses.")
            self.add("dooverride_optional",
                     "Override this method in custom subclasses or leave the default implementation.")
            self.add("infrastructure",
                     "This is part of a particular piece of internal infrastructure and is typically not required "
                     "to be used.")

        def add(self, name, text):
            """!
            Adds a text snippet to the storage, so it can be used.
            @param name: The snippet name. This is the `foo` in `@a_foo`.
            @param text: The snippet text.
            """
            self.storage[name] = text

        def toaliases(self):
            ast = "@par\n<div class='anisedochint'>{}</div>\n\n"
            return {"a_"+k: ast.format(self.storage[k].replace("&", "&amp;").replace("<", "&lt;").
                                       replace(">", "&gt;")) for k in self.storage}

    @staticmethod
    def writedocumentations(destdir, throw_on_existing, fortype):
        """!
        Immediately exports the documentation as the export definitions currently specify.
        """
        doclist = [x for x in universe.documentation.exports.list if isinstance(x, fortype)]
        scopelist = [universe.executionscope.newsubscope(
            "Generating {} documentation ".format(y.format) + ("'{}'".format(y.heading)
                                                               if y.heading else y.key)) for y in doclist]
        for i, echan in enumerate(doclist):
            anise.framework.report.topscope().setprogress(
                anise.framework.report.ExecutionScope.ProgressDeterminate(float(i)/len(doclist)))
            with scopelist[i]:
                fullfn = destdir + "/" + echan.outputfile
                if throw_on_existing and os.path.exists(fullfn):
                    raise anise.framework.exceptions.BadFormatError(
                        "there already is a '{fullfn}', which is forbidden.".format(fullfn=fullfn))
                anise.utils.basic.mkdirs(os.path.dirname(fullfn))
                echan.generate(fullfn)

    class Export:
        """!
        Defines a documentation export.
        """

        def __init__(self, key, format, name=None, heading=None, includedeveloperdoc=False, imagepaths=None, additionaltext="",
                     excludepatterns=None, source=None):
            """!
            @param key: One of the keys stored in pool. Defines which documentation to export.
            @param format: The output format (one of documentation.ExportFormat).
            @param name: The export definition name. Finding definitions by name is possible with `exports.getbyname`.
            @param heading: The title.
            @param includedeveloperdoc: If the output shall include the developer documentation as well (does not work 
                                        for all output formats).
            @param imagepaths: Directories from where images are used in the documentation.
            @param additionaltext: Additional text to append to the selected documentation.
            @param excludepatterns: File patterns to exclude from reading.
            @param source: The documentation source root.
            """
            imagepaths = imagepaths or []
            excludepatterns = excludepatterns or []
            self.key = key
            self.format = format
            self.name = name
            self.heading = heading
            self.includedeveloperdoc = includedeveloperdoc
            self.imagepaths = imagepaths
            self.additionaltext = additionaltext
            self.excludepatterns = excludepatterns
            self.source = source or anise.framework.files.TaskExecution(universe.packages.makerawpackage)

        def generate(self, targetpath):
            """!
            Generates the content.
            @a_dooverride_optional
            """
            if self.format == ExportFormat.PDF:
                internals.generate_userdocumentation_aspdf(
                    head1=universe.name, additionaltext=self.additionaltext, head2=self.heading,
                    sourcedir=self.source.path(), targetfile=targetpath,
                    documentations=universe.documentation.pool.gettexts(), targetdocumentation=self.key,
                    includedeveloperdoc=self.includedeveloperdoc, excludepatterns=self.excludepatterns,
                    imagepaths=self.imagepaths,
                    aliases=universe.documentation.snippets.toaliases())
            elif self.format == ExportFormat.Plaintext:
                internals.generate_userdocumentation_asplaintext(
                    projectname=universe.name, additionaltext=self.additionaltext, head=self.heading,
                    sourcedir=self.source.path(), targetfile=targetpath,
                    documentations=universe.documentation.pool.gettexts(), targetdocumentation=self.key,
                    excludepatterns=self.excludepatterns,
                    aliases=universe.documentation.snippets.toaliases())
            elif self.format == ExportFormat.HTML:
                backgroundimage = universe.base.getbackgroundimage()
                internals.generate_userdocumentation_ashtml(
                    head1=universe.name, head2=self.heading,
                    additionaltext=self.additionaltext, sourcedir=self.source.path(), targetdir=targetpath,
                    basecolor=universe.base.basecolor, backgroundimage=backgroundimage,
                    documentations=universe.documentation.pool.gettexts(),
                    excludepatterns=self.excludepatterns,
                    aliases=universe.documentation.snippets.toaliases())
            else:
                raise anise.framework.exceptions.BadInputError(
                    "unknown document output format '{format}'.".format(format=self.format))

    class AbstractFileExport(Export):
        """!
        Abstract base class for exports, which write documentation content to files.
        """

        def __init__(self, key, outputfile, format, name=None, heading=None, includedeveloperdoc=False,
                     imagepaths=None, additionaltext="",
                     excludepatterns=None, source=None):
            internals.Export.__init__(self, key, format, name, heading, includedeveloperdoc, imagepaths, additionaltext, excludepatterns,
                                      source)
            self.outputfile = outputfile

    @staticmethod
    def generate_userdocumentation_ashtml(
            *, head1, head2, additionaltext="", sourcedir, targetdir, basecolor,
            backgroundimage=None, logoimage=None,
            documentations, excludepatterns=None,
            aliases=None):
        """!
        Writes all user documentations as html (powered by Doxygen). You find it afterwards in the `(documentkey).html`
        files (and tons of other files).
        This will contain the developer documentation as well, so references into it will work.
        @param head1: The first heading (typically the project name).
        @param head2: The second heading.
        @param additionaltext: Auxiliary source code passed to Doxygen for injecting additional content.
        @param sourcedir: The directory where the sourcefiles are.
        @param targetdir: The directory for the html output.
        @param basecolor: The basecolor for colouring the output.
        @param backgroundimage: Absolute path to the background image (will be copied).
        @param logoimage: Absolute path to the logo image (will be copied).
        @param documentations: An anise documentations tuple for further pages, which do not exist in the sourcecode
                               (e.g. the readme sources) as added by anise.features.documentation.pool.add.
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param aliases: Custom command aliases (as string/string-dict).
        """
        internals.generate_raw(
            head1, head2, sourcedir, targetdir, "HTML", additionaltext,
            basecolor, backgroundimage, logoimage,
            documentations, htmlheadertranslation=lambda x: x.replace("$title", head2 or ""),
            excludepatterns=excludepatterns, aliases=aliases)

    @staticmethod
    def generate_userdocumentation_asmanpage(#TODO section
            *, projectname, entityname, section, additionaltext="", sourcedir, targetdir, documentations,
            targetdocumentation, excludepatterns=None, aliases=None):
        """!
        Writes one user documentation page as manpage (powered by Doxygen) and returns the path to the result file.
        @param projectname: The project name.
        @param entityname: The `NAME` for the manpage (typically the name of the command, function or whatever).
        @param section: The section number (see man manpage).
        @param additionaltext: Auxiliary source code passed to Doxygen for injecting additional content.
        @param sourcedir: The directory where the sourcefiles are.
        @param targetdir: The directory for the manpage output.
        @param documentations: An anise documentations tuple for further pages, which do not exist in the sourcecode
                               (e.g. the readme sources) as added by anise.features.documentation.pool.add.
        @param targetdocumentation: The key for the documentation which must be written.
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param aliases: Custom command aliases (as string/string-dict).
        """
        entityname = entityname.replace(" ", "-")
        _itext = """/*!
\\page {entityname}
\\copydoc {targetdocumentation}
        */
        """.format(**locals())
        internals.generate_raw(
            projectname, None, sourcedir, targetdir, "MAN", additionaltext + "\n" + _itext,
            (0, 0, 7), None, None, [], documentations, excludepatterns=excludepatterns,
            aliases=aliases)
        return "{targetdir}/man/man{section}/{entityname}.{section}".format(**locals())

    @staticmethod
    def generate_userdocumentation_asplaintext(
            *, projectname, head, additionaltext="", sourcedir, targetfile, documentations, targetdocumentation,
            excludepatterns=None, aliases=None):
        """!
        Writes one user documentation page as plain text (powered by Doxygen).
        @param projectname: The project name.
        @param head: The heading.
        @param additionaltext: Auxiliary source code passed to Doxygen for injecting additional content.
        @param sourcedir: The directory where the sourcefiles are.
        @param targetfile: The target file for the plaintext output.
        @param documentations: An anise documentations tuple for further pages, which do not exist in the sourcecode
                               (e.g. the readme sources) as added by anise.features.documentation.pool.add.
        @param targetdocumentation: The key for the documentation which must be written.
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param aliases: Custom command aliases (as string/string-dict).
        """
        with anise.framework.files.TempDir() as tempdir:
            f = internals.generate_userdocumentation_asmanpage(
                projectname=projectname, entityname=head, section=7, additionaltext=additionaltext, sourcedir=sourcedir,
                targetdir=tempdir.path(), documentations=documentations, targetdocumentation=targetdocumentation,
                excludepatterns=excludepatterns,
                aliases=aliases)
            r, plaintext = anise.utils.basic.call(
                "man -l {0} | col -b".format(shlex.quote(f)), shell=True,
                raise_on_errors="Problem while unformatting man page")
            res = ["  " + projectname + ": " + head]
            res.append("-" * len(res[0]) + "\n")
            began = None
            for l in plaintext.split("\n"):
                if began is not None and began >= 0:
                    res.append(l)
                elif began is None and l.startswith("NAME"):
                    began = -1
                elif began is not None:
                    began += 1
        anise.utils.basic.writetofile(targetfile, "\n".join(res[:-1]))

    @staticmethod
    def generate_userdocumentation_aspdf(
            *, head1, head2, additionaltext="", sourcedir, targetfile, documentations, targetdocumentation,
            includedeveloperdoc=False, excludepatterns=None, imagepaths=None, aliases=None):
        """!
        Writes one user documentation page as pdf (powered by Doxygen).
        @param head1: The first heading (typically the project name).
        @param head2: The second heading.
        @param additionaltext: Auxiliary source code passed to Doxygen for injecting additional content.
        @param sourcedir: The directory where the sourcefiles are.
        @param targetfile: The target file for the pdf output.
        @param documentations: An anise documentations tuple for further pages, which do not exist in the sourcecode
                               (e.g. the readme sources) as added by anise.features.documentation.pool.add.
        @param targetdocumentation: The key for the documentation which must be written.
        @param includedeveloperdoc: If the pdf shall include the developer documentation.
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param imagepaths: Directories from where images are used in the documentation.
        @param aliases: Custom command aliases (as string/string-dict).
        """
        with anise.framework.files.TempDir() as tempdir:
            internals.generate_raw(
                head1, head2, sourcedir, tempdir.path()+"/TODO.pdf", "LATEX",
                additionaltext, (0, 0, 7), None, None, [], documentations, excludepatterns=excludepatterns,
                imagepaths=imagepaths, aliases=aliases)
            tempdir.dl(subpath="TODO.pdf", to=targetfile)

    @staticmethod
    def generate_homepage(
            *, head1, head2, sourcedir, targetdir, basecolor, backgroundimage, logoimage, sections,
            excludepatterns=None, aliases=None):
        """!
        Writes the project homepage (powered by Doxygen). You find it afterwards in `index.html`
        (and tons of other files).
        @param head1: The first heading (typically the project name).
        @param head2: The second heading (typically a one-liner description).
        @param sourcedir: The directory where the sourcefiles are.
        @param targetdir: The directory for the html output.
        @param basecolor: The basecolor for colouring the output.
        @param backgroundimage: Absolute path to the background image (will be copied).
        @param logoimage: Absolute path to the logo image (will be copied).
        @param sections: A list of (sectiontitle,sectionsource,additionalparams) tuples for the actual content.
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param aliases: Custom command aliases (as string/string-dict).
        """
        page = ""
        for sectionkey, sectiontext, additionalparams, sectionhead in sections:
            page += (f"{universe.docrender.internals.rst_heading(sectionhead, variant='^', anchor=sectionkey)}\n"
                     f"{sectiontext}\n\n")
        internals.generate_raw(
            head1, head2, sourcedir, targetdir, "HTML", page,
            basecolor, backgroundimage, logoimage, [],
            excludepatterns=excludepatterns, aliases=aliases)
        return targetdir

#TODO signature formattings

    @staticmethod
    def generate_raw(#TODO
            head1, head2, sourcedir, targetdir, outputformat, additionalsources,
            basecolor, backgroundimage, logoimage, documentations,
            htmlheadertranslation=None, excludepatterns=None, imagepaths=None, aliases=None):
        """!
        Writes some output with Doxygen. Mostly used by some higher-level functions in this module.
        @param head1: The first heading.
        @param head2: The second heading.
        @param sourcedir: The source root directory.
        @param targetdir: The directory for the output.
        @param outputformat: The desired output format (as in doxygen config's GENERATE_xxx keys).
        @param additionalsources: Auxiliary rst source code for additional content.
        @param basecolor: The basecolor for colouring the output.
        @param backgroundimage: Absolute path to the background image (will be copied).
        @param logoimage: Absolute path to the logo image (will be copied).
        @param documentations: An anise documentations tuple for further pages, which do not exist in the sourcecode
                               (e.g. the readme sources) as added by anise.features.documentation.pool.add.
        @param htmlheadertranslation: A function which can replace stuff in the default html header (html output only).
        @param excludepatterns: File path patterns to exclude for composing the documentation.
        @param imagepaths: Directories from where images are used in the documentation.
        @param aliases: Custom command aliases (as string/string-dict).
        """
        pieces = []
        with anise.framework.files.TempDir() as tempd:
            for dk, dt in documentations:
                ff = f'{tempd.path()}/{dk}.rst'
                with open(ff, 'w') as f:
                    f.write(dt)
                pieces.append(universe.docrender.RstTextsPiece(f'{dk}.rst', ff))
            outputspec = universe.docrender.HtmlOutputSpec(targetdir, title=head1,
                                                           theme=universe.documentation.htmltheme)
            author = ", ".join(universe.base.authors)
            if backgroundimage:
                shutil.copyfile(backgroundimage, f"{targetdir}/_background.png")
                #TODO theme.conf
            if logoimage:
                shutil.copyfile(logoimage, f"{targetdir}/_logo.png")
            if callable(additionalsources):
                additionalsources = additionalsources()
            universe.docrender.render(pieces, outputspec, head1, author=author, mainbody=additionalsources,
                                      basecolor=basecolor)


class ExportFormat:
    """!
    Enumeration of output formats.
    """
    Plaintext = "Plaintext"
    PDF = "PDF"
    HTML = "HTML"


class HomepageExport(internals.Export):
    """!
    Describes an export of an anise documentation directly into the homepage.
    """

    def __init__(self, key, name=None, heading=None, includedeveloperdoc=False, imagepaths=None, additionaltext="", excludepatterns=None,
                 source=None, linktarget=None,
                 index=100000):
        """!
        @param key: One of the keys stored in pool. Defines which documentation to export.
        @param name: The export definition name. Finding definitions by name is possible with `exports.getbyname`.
        @param heading: The title.
        @param includedeveloperdoc: If the output shall include the developer documentation as well (does not work for
                                    all output formats).
        @param imagepaths: Directories from where images are used in the documentation.
        @param additionaltext: Additional text to append to the selected documentation.
        @param excludepatterns: File patterns to exclude from reading.
        @param source: The documentation source root.
        @param linktarget: Link target.
        @param index: A position number. Used for ordering.
        """
        internals.Export.__init__(self, key, ExportFormat.HTML, name, heading, includedeveloperdoc, imagepaths, additionaltext,
                                  excludepatterns, source)
        self.linktarget = linktarget
        self.index = index


class PackageExport(internals.AbstractFileExport):
    """!
    Describes an export of an anise documentation into the distribution package(s).
    """

    def __init__(self, key, outputfile, format, name=None, heading=None, includedeveloperdoc=False, imagepaths=None, additionaltext="",
                 excludepatterns=None, source=None):
        """!
        @param key: One of the keys stored in pool. Defines which documentation to export.
        @param outputfile: Target file path relative to the 'raw package's root.
        @param format: The output format (one of documentation.ExportFormat).
        @param name: The export definition name. Finding definitions by name is possible with `exports.getbyname`.
        @param heading: The title.
        @param includedeveloperdoc: If the output shall include the developer documentation as well (does not work for
                                    all output formats).
        @param imagepaths: Directories from where images are used in the documentation.
        @param additionaltext: Additional text to append to the selected documentation.
        @param excludepatterns: File patterns to exclude from reading.
        @param source: The documentation source root.
        """
        internals.AbstractFileExport.__init__(self, key, outputfile, format, name, heading, includedeveloperdoc,
                                              imagepaths, additionaltext,
                                              excludepatterns, source)


class SourceExport(internals.AbstractFileExport):
    """!
    Describes an export of an anise documentation back into the source tree.
    """

    def __init__(self, key, outputfile, format, name=None, heading=None, includedeveloperdoc=False, imagepaths=None, additionaltext="",
                 excludepatterns=None, source=None):
        """!
        @param key: One of the keys stored in pool. Defines which documentation to export.
        @param outputfile: Target file path relative to the 'raw package's root.
        @param format: The output format (one of documentation.ExportFormat).
        @param name: The export definition name. Finding definitions by name is possible with `exports.getbyname`.
        @param heading: The title.
        @param includedeveloperdoc: If the output shall include the developer documentation as well (does not work for
                                    all output formats).
        @param imagepaths: Directories from where images are used in the documentation.
        @param additionaltext: Additional text to append to the selected documentation.
        @param excludepatterns: File patterns to exclude from reading.
        @param source: The documentation source root.
        """
        internals.AbstractFileExport.__init__(self, key, outputfile, format, name, heading, includedeveloperdoc,
                                              imagepaths, additionaltext, excludepatterns, source)


def has_developerdocumentation(also_include_in_readme=False, **kwargs):#TODO
    """!
    Configures a project for creating a Doxygen developer documentation.
    @param also_include_in_readme: If a developer documentation shall be included in the README files.
    @param kwargs: Additional arguments for the export definition.
    """
    def _add_homepage_export():
        additionaltext = lambda: """
    /**
=========
Main page
========= 

This is a documentation for developers of {projectname} as well as interested hackers, who want
to work with the source code. It is automatically created from the source code comments. It is
not guaranteed that the information is correct or up-to-date (even if it is based on the latest version).
However, it should give a good overview.

You are current reading the developer documentation based on version {version}.

{maturity}
    */""".format(projectname=universe.name, version=universe.packages.getversion(),
                 maturity=universe.base.getprojectstatusdescription())
        universe.documentation.exports.add(
            HomepageExport(key="developerdoc", heading="Developer Documentation", includedeveloperdoc=True,
                           additionaltext=additionaltext, linktarget="index.html", **kwargs))
    universe.addhook(anise.framework.features.HOOK_BEFORE_EXECUTION, _add_homepage_export)
    universe.documentation.readme_include_developer_doc = kwargs if also_include_in_readme else False
