# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for transferring file structures to a destination place, e.g. on a web or file server.
"""

import os
import random
import shlex
import shutil
import stat
import sys

import anise.framework.features
import anise.framework.files
import anise.framework.report
import anise.utils

from anise.framework.projects import universe

homepage = anise.framework.features.loadfeature("homepage").featuremodule


class internals:

    @staticmethod
    def generatenewindex(appidx, indexpattern, projectname):
        if not "/{}/".format(projectname) in appidx:
            appidx += indexpattern.format(name=projectname)
        return appidx

    @staticmethod
    def updateindexfile(appidxpath, indexpattern, projectname):
        try:
            appidx = anise.utils.basic.readfromfile(appidxpath)
        except Exception:
            appidx = ""
        appidx = internals.generatenewindex(appidx, indexpattern, projectname)
        anise.utils.basic.writetofile(appidxpath, appidx)

    @staticmethod
    def createprogresscallback(execscope):
        def _x(gross, done):
            execscope.setprogress(anise.framework.report.ExecutionScope.ProgressDeterminate(0.95*float(done)/gross))
        return _x


def sshupload(*, host, username, destination, source, identityfile=None,
              indexpattern="{name}\n", port=22, transferstrategy=None,
              contentdescription="content",
              collectlinkshook=homepage.HOOK_UPLOAD_HOMEPAGE_COLLECTLINKS):
    """!
    Deploys content to a server (e.g. a web server) via ssh.
    @param host: The server host name or address.
    @param username: The login username.
    @param destination: The destination file path.
    @param source: An anise.framework.files.Filestructure.
    @param identityfile: Path to a ssh identity file for login via public-key.
    @param indexpattern: Pattern for index entries.
    @param port: Port number.
    @param transferstrategy: File transfer strategy.
    @param contentdescription: Textual description of the kind of content.
    @param collectlinkshook: Which hook to trigger for collecting a list of symlinks to create/refresh after upload.
    """
    hp = source.with_modified_rootname(universe.name)
    umask = os.umask(0)
    os.umask(umask)

    def _chmod(_p):
        for x in os.listdir(_p):
            px = _p+"/"+x
            os.chmod(px, (stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO) & ~umask)
            if os.path.isdir(px):
                _chmod(px)
    _chmod(hp.path())
    with universe.executionscope.newsubscope("Deploying {} to ssh server".format(contentdescription)):
        hpdir = hp.path()
        sshtgt = "{}@{}:/{}".format(username, host, destination)
        if transferstrategy == "tar":
            with anise.framework.files.TempDir() as tartmpdir:
                tarfilename = "hp.{0}.{1}.tgz".format(os.getpid(), random.randrange(sys.maxsize))
                anise.utils.basic.maketarball(hpdir, tartmpdir.path()+"/"+tarfilename)
                with anise.framework.files.TempDir() as tmpdir:
                    serverroot = tmpdir.path() + "/m"
                    os.mkdir(serverroot)
                    with anise.utils.ssh.Mount(sshtgt, serverroot, port=port, identityfile=identityfile):
                        tartmpdir.dl(
                            subpath=tarfilename, to=serverroot,
                            progresscallback=internals.createprogresscallback(anise.framework.report.topscope()))
                        p = serverroot + "/" + universe.name
                        internals.updateindexfile(serverroot + "/_apps.txt", indexpattern, universe.name)
                        with universe.executionscope.newsubscope("Expanding file structure on server"):
                            pdst = shlex.quote(destination+"/"+universe.name)
                            dst = shlex.quote(destination)
                            tarf = shlex.quote(tarfilename)
                            options = (["-i", identityfile] if identityfile else []) + \
                                [username+"@"+host,
                                 "rm -rf {pdst} ; cd {dst} && tar xfz {tarf} && rm {tarf}".format(**locals())]
                            anise.utils.basic.call("ssh", "-p", str(port), *options,
                                                   raise_on_errors="unable to call ssh")
        else:
            with anise.framework.files.TempDir() as tmpdir:
                serverroot = tmpdir.path() + "/m"
                os.mkdir(serverroot)
                with anise.utils.ssh.Mount(sshtgt, serverroot, port=port, identityfile=identityfile):
                    p = serverroot + "/" + universe.name
                    if os.path.exists(p):
                        shutil.rmtree(p)
                    hp.dl(to=p, progresscallback=internals.createprogresscallback(anise.framework.report.topscope()))
                    internals.updateindexfile(serverroot + "/_apps.txt", indexpattern, universe.name)
        sshcmd = ""
        for fct in universe.gethooks(collectlinkshook):
            for src, tgt in fct():
                ssrc = shlex.quote(destination + "/" + universe.name + "/" + src)
                stgt = shlex.quote(destination + "/" + universe.name + "/" + tgt)
                sshcmd += "rm {stgt};ln -s {ssrc} {stgt};".format(**locals())
        if len(sshcmd) > 0:
            sshcmd = sshcmd[:-1]
            options = (["-i", identityfile] if identityfile else []) + [username+"@"+host, sshcmd]
            anise.utils.basic.call("ssh", "-p", str(port), *options, raise_on_errors="unable to call ssh")


def localupload(*, destination, source, indexpattern="{name}\n",
                contentdescription="content",
                collectlinkshook=homepage.HOOK_UPLOAD_HOMEPAGE_COLLECTLINKS):
    """!
    Deploys content to a local place (e.g. for testing).
    @param destination: The destination file path.
    @param source: An anise.framework.files.Filestructure.
    @param indexpattern: Pattern for index entries.
    @param contentdescription: Textual description of the kind of content.
    @param collectlinkshook: Which hook to trigger for collecting a list of symlinks to create/refresh after upload.
    """
    hp = source.with_modified_rootname(universe.name)
    with universe.executionscope.newsubscope("Deploying {} locally".format(contentdescription)):
        pdestination = destination + "/" + universe.name
        anise.utils.basic.mkdirs(pdestination)
        hp.dl(to=pdestination, progresscallback=internals.createprogresscallback(anise.framework.report.topscope()))
        internals.updateindexfile(destination + "/_apps.txt", indexpattern, universe.name)
        for fct in universe.gethooks(collectlinkshook):
            for src, tgt in fct():
                ssrc = pdestination + "/" + src
                stgt = pdestination + "/" + tgt
                if os.path.exists(stgt):
                    os.unlink(stgt)
                os.symlink(ssrc, stgt)
