# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for generating a project homepage.
"""

import anise.framework.features
import anise.framework.files
import anise.framework.report
import anise.utils

from anise.framework.projects import universe


HOOK_UPLOAD_HOMEPAGE_COLLECTLINKS = anise.framework.features.Hook(
    wizard_functionbody="return [('sourcepath', 'destinationpath')]",
    doc="This hook is called for collecting symlinks to deploy together with the homepage downloads.")


class internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.homepage.sections = internals.Sections()
        universe.homepage.icon = None

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _basic_homepage_sections():

        # About
        def my_homepage_section_about(homepage, sourcedir, backgroundimage, *a):
            return universe.base.longdescription

        universe.homepage.sections.add(1000, "about", "About", my_homepage_section_about)

        # Imprint
        def my_homepage_section_imprint(homepage, sourcedir, backgroundimage, *a):
            if not hasattr(universe.homepage, "imprint"):
                anise.utils.logging.logwarning("The universe object does not define an imprint!")
                return None
            else:
                result = universe.homepage.imprint.replace("[MAIL]", universe.base.email)
                feedbacksect = universe.homepage.sections.getsectionbykey("feedback")
                if feedbacksect:  # if there is a 'feedback' section, we try to include it here
                    result += universe.docrender.internals.rst_heading(feedbacksect[4], variant=':') + \
                              feedbacksect[2](homepage, sourcedir, backgroundimage, *a)
                    universe.homepage.sections.replacetextfct("feedback", lambda *a, **b: None)

                return result

        universe.homepage.sections.add(8000, "imprint", "Imprint", my_homepage_section_imprint)

        # Feedback
        def my_homepage_section_feedback(homepage, sourcedir, backgroundimage, *a):
            return (f"If you have funny rants about {universe.name} itself or about some techniques it uses, some"
                    f" constructive feedback, a cool patch or a mysterious problem, feel free to mail it"
                    f" to {universe.base.email}.")

        universe.homepage.sections.add(9000, "feedback", "Feedback", my_homepage_section_feedback)

    class Sections:
        """!
        Storage of all homepage sections.
        """

        def __init__(self):
            self.list = []

        def add(self, index, key, header, textfct, **additionalparams):
            """!
            Adds a homepage section.
            @param index: A position index number. This controls the ordering of sections.
            @param key: The section name. Also used as heading.
            @param header: The section header text.
            @param textfct: A function returning the section Markdown source as string.
            @param additionalparams: Additional section parameters (advanced).
            """
            self.list.append([index, key, textfct, additionalparams, header])

        def getsectionbykey(self, key):
            """!
            Gets the section record for a key.
            @a_infrastructure
            @param key: The section name.
            """
            for l in self.list:
                if l[1] == key:
                    return l

        def gettextfct(self, key):
            """!
            Gets the generator function for a key, which returns the section Markdown source.
            @param key: The section name.
            @return: The generator function.
            """
            r = self.getsectionbykey(key)
            if r:
                return r[2]

        def replacetextfct(self, key, textfct):
            """!
            Replaces a generator function for a key.
            @param key: The section name.
            @param textfct: The new generator function.
            """
            r = self.getsectionbykey(key)
            if r:
                r[2] = textfct


def makehomepage():
    """!
    Creates the project homepage from the registered sections.
    @return: The homepage as framework.files.Filestructure.
    """
    with universe.executionscope.newsubscope("Creating the project homepage"):
        homepage = anise.framework.files.TempDir(dirname=universe.name + "-homepage")
        if not getattr(universe, "omitsyncvcs", False):
            universe.versioncontrol.current.syncversioncontrolsystem(
                commitmessage="automated commit for make homepage")
        rawpkg = universe.packages.makerawpackage().mv(universe.name)
        sourcedir = rawpkg.path()
        _bkgimg = universe.base.getbackgroundimage()
        backgroundimage = (universe.specialpaths.projectdir + "/" + _bkgimg) if _bkgimg else None
        anise.utils.basic.writetofile(homepage.path() + "/license", universe.base.license.getlicensetext())
        maturity = universe.base.getprojectstatusdescription(True)
        if len(maturity) > 0:
            maturity = "state: " + maturity
        sections = []
        rawsectionlist = sorted(universe.homepage.sections.list)
        rawsection2scope = []
        for rs in rawsectionlist:
            rawsection2scope.append(universe.executionscope.newsubscope("Generating section '{}'".format(rs[1])))
        topscope = anise.framework.report.topscope()
        for i, section in enumerate(rawsectionlist):
            topscope.setprogress(
                anise.framework.report.ExecutionScope.ProgressDeterminate(float(i)/len(rawsectionlist)))
            with rawsection2scope[i]:
                key = section[1]
                content = section[2](homepage, sourcedir, backgroundimage)
                additionalparams = section[3]
                header = section[4]
                if content is not None:
                    sections.append((key, content, additionalparams, header))
        logoimage = f"{universe.specialpaths.projectdir}/{universe.homepage.icon}" if universe.homepage.icon else None
        universe.documentation.internals.generate_homepage(
            head1=universe.name, head2=universe.base.summary,
            sourcedir=sourcedir, targetdir=homepage.path(), basecolor=universe.base.basecolor,
            backgroundimage=backgroundimage, logoimage=logoimage, sections=sections)
        return homepage
