# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for specifying, generating and processing image galleries. It is typically used as part of the project
homepage.
"""

import os
import shutil

import anise.framework.features

from anise.framework.projects import universe


class internals:

    class Pool:
        """!
        Storage of all image galleries.
        """

        def __init__(self):
            self.list = []

        def add(self, gallery, name=""):
            """!
            Adds an image gallery.
            @param gallery: An instance of MediaGallery specifying the gallery details.
            @param name: Gallery name.
            """
            self.list.append((name, gallery))

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.mediagalleries.pool = internals.Pool()

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _add_homepage_section():

        def my_homepage_section(homepage, sourcedir, backgroundimage, *a):#TODO
            if len(universe.mediagalleries.pool.list) > 0:
                galleries = ""
                i = 0
                for name, gallery in universe.mediagalleries.pool.list:
                    galleries += "\n#### {galleryname}\n<div class='anisedoc_imagecarousel'>\n".format(galleryname=name)
                    ngallerydir = "gallery{0}/".format(i)
                    cgallery = gallery.cloneto(homepage.path() + "/" + ngallerydir)
                    for img in cgallery.getmedialist():
                        mtitle = img.description.replace("\n", " ").replace('"', "''")
                        mpath = ngallerydir + img.path
                        if img.mediatype == "image":
                            galleries += "<div class='anisedoc_image'>\n" \
                                         "\image html \"{mpath}\" \"{mtitle}\"</div>".format(**locals())
                        elif img.mediatype == "video":
                            galleries += "<div class='anisedoc_video'>{mpath}|||{mtitle}</div>\n".format(**locals())
                    i += 1
                    galleries += "</div>\n"
                return galleries

        universe.homepage.sections.add(6500, "gallery", "Gallery", my_homepage_section)


class Media:
    """!
    One image in a gallery.
    """

    def __init__(self, path, description, mediatype):
        """!
        @param path: The image file path.
        @param description: Description text.
        """
        self.path = path
        self.description = description
        self.mediatype = mediatype


class MediaGallery:
    """!
    An image gallery.
    """

    def __init__(self, path=None, paths=None):
        """!
        @param path: The directory path where the image gallery is stored.
        @param paths: The directory paths where the image gallery is stored.
        """
        paths = paths or []
        if path is not None:
            paths.append(path)
        self.paths = []
        for path in paths:
            while path.endswith("/"):
                path = path[:-1]
            self.paths.append(path)

    def _matchfile(self, name):
        for pat in [".png", ".jpg", ".jpeg"]:
            if name.endswith(pat):
                return "image"
        for pat in [".mp4", ".webm"]:
            if name.endswith(pat):
                return "video"
        return False

    def getmedialist(self):
        """!
        Gets the list of media in the gallery.
        @return: List of Media instances.
        """
        result = []
        for path in self.paths:
            isabsolute = path.startswith("/")
            if isabsolute:
                fullpath = path
            else:
                fullpath = universe.specialpaths.projectdir + "/" + path
            for img in os.listdir(fullpath):
                mediatype = self._matchfile(img)
                if mediatype:
                    fimgdesc = path + "/" + img + ".txt"
                    desc = ""
                    if os.path.isfile(fimgdesc):
                        with open(fimgdesc, "r") as f:
                            desc = f.read()
                    result.append(Media(img, desc, mediatype))
        result.sort(key=lambda x: x.path)
        return result

    def cloneto(self, destpath):
        """!
        Clones the media gallery directory to a different place.
        @param destpath: The destination where to copy the image gallery data. Absolute path.
        @return: The new MediaGallery instance pointing to the cloned directory.
        """
        os.makedirs(destpath, exist_ok=True)
        for path in self.paths:
            isabsolute = path.startswith("/")
            if isabsolute:
                fullpath = path
            else:
                fullpath = universe.specialpaths.projectdir + "/" + path
            for img in os.listdir(fullpath):
                if self._matchfile(img) or img.endswith(".txt"):
                    shutil.copyfile(path + "/" + img, destpath + "/" + img)
        return MediaGallery(destpath)
