# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Features which provides tools to run a procedure in different environments in a transparent way.
"""

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


class internals:

    @staticmethod
    def get(x=None, defaultmilieu=None):
        """!
        Returns a Milieu (or MilieuComposite).

        Typically this milieu is created by the given specification. The specification is a list of tuples.
        Each tuple has at least a `'class'` key pointing to a Milieu(Composite) subclass and optional 
        parameters. The milieux will be created stacked together.
        @param x: The milieu specification (or a direct milieu instance).
        @param defaultmilieu: The default milieu to create if no other one is specified.
        """
        defaultmilieu = defaultmilieu or universe.milieus.defaultmilieu
        if not isinstance(x, list):
            x = [x]
        result = None
        for part in x:
            if part is None or isinstance(part, str) or isinstance(part, type):
                part = {'class': part}
            if isinstance(part, dict):
                part = dict(part)
                clss = part.pop("class", None)
                if not clss:
                    if result:
                        raise anise.framework.exceptions.BadInputError("invalid to omit 'class' in env configuration")
                    result = internals.get(defaultmilieu)
                    continue
                if isinstance(clss, str):
                    _clss = clss
                    clss = universe.milieus.milieustore.get(_clss, None)
                    if clss is None:
                        clss = eval(_clss, universe().__dict__)
                oresult = result
                result = clss(**part)
                if 'setinner' in dir(result):  # a.k.a. if isinstance(result, MilieuComposite):
                    if oresult is None:
                        oresult = internals.get(defaultmilieu)
                    result.setinner(oresult)
                if isinstance(result, Milieu):
                    if oresult is not None:
                        anise.utils.logging.logdebug("Skipped Milieu because inside chain: " + str(result))
                        result = oresult
            else:
                result = part
            result._initialize()
        return result

    @staticmethod
    def registermilieu(name, createcallable):
        """!
        Registers a milieu by a name.

        This allows shorter names for large class names, and also to register easier factories.
        @param name: The name to register the new milieu with.
        @param createcallable: The class or callable which creates the instances.
        """
        universe.milieus.milieustore[name] = createcallable

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.milieus.milieustore = {}
        universe.milieus.get = internals.get
        universe.milieus.register = internals.registermilieu
        universe.milieus.defaultmilieu = "root"
        universe.milieus.register("root", DefaultMilieu)


class Milieu:
    """!
    Milieux are a functionality which allows a build task to work in different environments.

    This is advanced functionality and requires the build task to be coded in a special way. Instead of directly
    executing actions, e.g. executing external commands, it must use the call() method of a Milieu instance 
    for it.

    Tasks can get a Milieu by calling milieus.get (internals.get). Making milieux 
    available on the other hand is possible with `defaultmilieu` and milieus.register 
    (internals.registermilieu).

    See subclasses or implement an own one. See also MilieuComposite.

    Subclasses often not only implement methods from this class, but also add new tools to it.
    """

    def __init__(self):
        self._initialized = False

    def _initialize(self):
        """!
        Called from outside for initialization.
        @a_infrastructure
        """
        if not self._initialized:
            self._initialized = True
            self.initialize()

    def initialize(self):
        """!
        Initializes the milieu (composite).
        @a_dooverride_optional
        """
        pass

    def call(self, *cmdline, shell=False, decode=True, raise_on_errors=None, split_output_channels=False):
        """!
        Executes a command line.
        @param cmdline: The command line as list of strings.
        @param shell: If the command shall be interpreted by a shell (only the first part of cmdline will apply!).
        @param decode: If the result is to be decoded to a string.
        @param raise_on_errors: If errors shall raise Exceptions (also specifies the exception text in this case).
        @param split_output_channels: If to separately record output and error stream.
        @return: Tuple of returncode, program output.
        """
        res = anise.utils.basic.call(*cmdline, shell=shell, decode=decode, raise_on_errors=raise_on_errors,
                                     split_output_channels=split_output_channels)
        anise.utils.logging.logdebug("Called {}:\n{}".format(cmdline, res[1]))
        return res

    def createtempdir(self, dirname=None):
        """!
        Returns a fresh anise.framework.files.TempDir.
        @param dirname: An optional name which the root directory should get; if not given, it gets an random name.
        """
        return anise.framework.files.TempDir(dirname=dirname)


class MilieuComposite:
    """!
    A milieu composite brings parts of functionality which can be stacked on top of a Milieu. 

    They behave largely equally to `Milieu`s.
    """

    def __init__(self):
        self.inner = None
        self._initialized = False

    def _initialize(self):
        """!
        Called from outside for initialization.
        @a_infrastructure
        """
        if not self._initialized:
            self._initialized = True
            self.initialize()

    def initialize(self):
        """!
        Initializes the milieu (composite).
        @a_dooverride_optional
        """
        pass

    def setinner(self, inner):
        """!
        Sets the inner milieu.
        @a_infrastructure
        """
        self.inner = inner

    def __getattr__(self, item):
        return getattr(self.inner, item)


class DefaultMilieu(Milieu):
    """!
    The default milieu.
    """
    pass
