# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Features for creating distribution packages.

Those are typically packages for download by end users.
"""

import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

ui = anise.framework.features.loadfeature("ui").featuremodule


HOOK_ENRICHRAWPACKAGE = anise.framework.features.Hook(
    wizard_functionparams=["destdir"], doc="This hook is called for inserting additional files to the raw package.")


class internals:

    @staticmethod
    def enrichrawpackage(packagedir):
        """!
        Adds files to a raw packages (as created by makerawpackage), like the license file.
        @param packagedir: The raw package root directory path.
        """
        with universe.executionscope.newsubscope("Enriching the rawpackage"):
            for hook in universe.gethooks(HOOK_ENRICHRAWPACKAGE):
                hook(destdir=packagedir)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.addtask(internals.getversion, "packages.", important=False)
        universe.addtask(internals.makerawpackage, "packages.", invisible=True)

    _makerawpackage_cache = None

    @staticmethod
    def getversion():
        """!
        Determines the version string of the project.
        """
        if not getattr(universe, "omitsyncvcs", False):
            universe.versioncontrol.current.syncversioncontrolsystem(commitmessage="automated anise commit",
                                                                     forceskipped=False)
        if hasattr(universe, "_fullversion"):
            return universe._fullversion
        elif hasattr(universe, "versioncontrol"):
            r = universe.versioncontrol.current.getfullversion()
            universe._fullversion = r
            return r
        else:
            anise.utils.logging.logwarning("No version control system found. Only using base version.")
            r = universe.base.version
            universe._fullversion = r
            return r

    @staticmethod
    def makerawpackage():
        """!
        Builds a raw package in some flavour. This is used by other components (e.g. packaging) for creating
        distributable binary packages.
        """
        targetdir = anise.framework.files.TempDir().path() + "/" + universe.name
        if internals._makerawpackage_cache is None:
            with universe.executionscope.newsubscope("Creating a raw package"):
                internals._makerawpackage_cache = universe.versioncontrol.current.createrawpackage()
                internals.enrichrawpackage(packagedir=internals._makerawpackage_cache.path())
        internals._makerawpackage_cache.dl(to=targetdir)
        return anise.framework.files.Filestructure(targetdir)


def tarpackage(source, namepostfix=None, pkgdescription=""):
    """!
    Builds a tarball which can be distributed. This is mostly a source package and is possibly not the most convenient
    package type for an end-user.
    @param source: An anise.framework.files.Filestructure.
    @param namepostfix: Postfix for the package root directory name.
    @param pkgdescription: The package description text.
    @return: A list containing new tarball as framework.files.Filestructure.
    """
    with universe.executionscope.newsubscope("Creating the tar package"):
        if not getattr(universe, "omitsyncvcs", False):
            universe.versioncontrol.current.syncversioncontrolsystem(commitmessage="automated commit from anise")
        pkgroot = universe.name + "-" + universe.packages.getversion() + (("-"+namepostfix) if namepostfix else "")
        tarname = anise.framework.files.TempDir().path() + "/" + pkgroot + ".tgz"
        srcfs = source.with_modified_rootname(pkgroot)
        anise.utils.basic.maketarball(rootdir=srcfs.path(), tarname=tarname)
        result = anise.framework.files.Filestructure(tarname)
        result.description = pkgdescription
        return result


# this class is a convenience, which avoids some boilerplate code in _projectdesc
class RawPackage(anise.framework.files.TaskExecution):
    """!
    A anise.framework.files.Filestructure subclass which contains the raw package, which is the typical starting point
    for any further packaging steps or for fetching parts of the project file structure for some processing.

    This is a convenience shortcut, which is equivalent to a anise.framework.files.TaskExecution with `makerawpackage`
    as content generator.
    """

    def __init__(self):
        anise.framework.files.TaskExecution.__init__(self, lambda: universe.packages.makerawpackage())
