# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for creation of Android distribution packages.
"""

import os
import xml.dom.minidom

import anise.framework.exceptions
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


class internals:

    @staticmethod
    def buildandroidpackage_with_ant(*, projectroot, name, version, targetname, androidpath):
        """!
        Builds an Android package with Ant.
        @param projectroot: The path to the Android package.
        @param name: The project name.
        @param version: The project version.
        @param targetname: The Ant target name which shall be built.
        @param androidpath: Path to the Android sdk's `android` tool.
        """
        with anise.utils.basic.ChDir(projectroot):
            r, o = anise.utils.basic.call(androidpath, "update", "project", "-p", ".",
                                          raise_on_errors="unable to call android sdk")
            r, o = anise.utils.basic.call("ant", targetname, raise_on_errors="errors while creating android package")
        resfile = None
        for f in os.listdir(projectroot + "/bin"):
            if f.endswith(".apk") and not f.endswith("-unaligned.apk"):
                if resfile:
                    raise anise.framework.exceptions.UnexpectedSituationError(
                        "More than one .apk file found in " + projectroot + "/bin")
                else:
                    resfile = f
        if resfile is None:
            raise anise.framework.exceptions.UnexpectedSituationError("No .apk file found in " + projectroot + "/bin")
        destresfile = "{name}-{version}.apk".format(**locals())
        os.rename(projectroot + "/bin/" + resfile, projectroot + "/" + destresfile)
        return anise.framework.files.Filestructure(projectroot + "/" + destresfile)


def package_with_ant(androidpath, source, relpath="", targetname="debug"):
    """!
    Builds an Android package with Ant.
    @param androidpath: Full path to the `android` tool (something like `.../android-sdk/tools/android`).
    @param source: An anise.framework.files.Filestructure.
    @param relpath: The root path of the android project, relative to `source`s root.
    @param targetname: Name of the Ant target to build.
    @return: An anise.framework.files.Filestructure.
    """
    with universe.executionscope.newsubscope("Creating Android installation package"):
        return internals.buildandroidpackage_with_ant(
            name=universe.name, version=universe.packages.getversion(), projectroot=source.path()+"/"+relpath,
            targetname=targetname, androidpath=androidpath)


def inject_projectdata_to_androidmanifest(outfile, data):
    """!
    Updates some project data (like version number) in the `AndroidManifest.xml`.
    @param outfile: The folder which holds the `AndroidManifest.xml`. Can be absolute or relative to the anise project
                    directory.
    @param data: Additional data to inject.
    """
    if outfile.startswith("/"):
        manifest = outfile + "/AndroidManifest.xml"
    else:
        manifest = universe.specialpaths.projectdir + "/" + outfile + "/AndroidManifest.xml"
    version = universe.packages.getversion()
    _version = version.split(".")
    _version2 = []
    if len(_version) > 1:
        _version2.append(_version[0])
    if len(_version) > 2:
        _version2.append(_version[1])
    _version2.append(_version[-1])
    _version2 = ["00000"+x for x in _version2]
    versioncode = int(_version2[0][-2:] + _version2[1][-2:] + _version2[2][-5:])
    doc = xml.dom.minidom.parse(manifest)
    doc.documentElement.attributes["android:versionName"] = version
    doc.documentElement.attributes["android:versionCode"] = str(versioncode)
    with open(manifest, "w") as f:
        doc.writexml(f, encoding="utf-8")
