# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for creation of appc container images.
"""

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


class internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.packages.appc.fullnamepattern = getattr(universe.packages.appc, "fullnamepattern", "{name}")

    @staticmethod
    def call(cmd, raise_on_errors="unable to run acbuild"):
        r, o = anise.utils.basic.call(cmd, raise_on_errors=raise_on_errors)
        anise.utils.logging.logdebug(o)


class PortDescription:
    """!
    Description for Appc network ports to be added to the container image.
    """

    def __init__(self, name, protocol, port):
        """!
        @param name: The port name.
        @param protocol: The protocol (typically "tcp" or "udp").
        @param port: The port number as available inside the container.
        """
        self.name = name
        self.protocol = protocol
        self.port = int(port)


class EnvironmentVariableDescription:
    """!
    Description for Appc environment variables to be added to the container image.
    """

    def __init__(self, name, value):
        """!
        @param name: The variable name.
        @param value: The value.
        """
        self.name = name
        self.value = value


def appcdebianpackage(debianpackage, fullname=None, ports=None, environment_variables=None, execcommand=None,
                      run=None, run_late=None):
    """!
    Builds an appc container image based on Debian.
    
    You need an existing Debian package source.
    @param debianpackage: An anise.framework.files.Filestructure containing the debian package.
    @param fullname: The full container name.
    @param ports: A list of PortDescription port descriptions to add to the container image.
    @param environment_variables: A list of EnvironmentVariableDescription descriptions to add to the container image.
    @param execcommand: A list of strings to specify the command to execute inside the container after boot.
    @param run: A list of command-line lists to execute for preparation.
    @param run_late: A list of command-line lists to execute for preparation after installing the package.
    @return: An anise.framework.files.Filestructure.
    """
    with universe.executionscope.newsubscope("Creating appc container image"):
        if anise.utils.basic.call("acbuild version", shell=True)[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'acbuild' seems to be unavailable")
        projectname = universe.name
        projectversion = universe.packages.getversion()
        containerdir = anise.framework.files.TempDir().path()
        sudocmd = getattr(universe, "sudocmd", ["sudo"])
        fullname = fullname or universe.packages.appc.fullnamepattern
        ports = ports or []
        environment_variables = environment_variables or []
        run = run or []
        run_late = run_late or []
        with anise.utils.basic.ChDir(containerdir):
            internals.call(["acbuild", "begin", "docker://debian"])
            internals.call(["acbuild", "set-name", fullname.format(name=projectname)])
            aciname = "{name}-{version}.aci".format(name=projectname, version=projectversion)
            internals.call(["acbuild", "copy", debianpackage.path(), "/root/pkg.deb"])
            internals.call([*sudocmd, "acbuild", "run", "--", "apt-get", "update"])
            for l in run:
                internals.call([*sudocmd, "acbuild", "run", "--", *l], raise_on_errors="unable to run 'run' stuff")
            for e in environment_variables:
                internals.call(["acbuild", "environment", "add", e.name, e.value])
            internals.call([*sudocmd, "acbuild", "run", "--", "/usr/bin/dpkg", "-i", "/root/pkg.deb"],
                           raise_on_errors=None)
            internals.call([*sudocmd, "acbuild", "run", "--", "apt-get", "-f", "-y", "install"])
            for p in ports:
                internals.call(["acbuild", "port", "add", p.name, p.protocol, str(p.port)])
            internals.call(["acbuild", "set-name", fullname.format(name=projectname)])
            for l in run_late:
                internals.call([*sudocmd, "acbuild", "run", "--", *l], raise_on_errors="unable to run 'run_late' stuff")
            if execcommand:
                internals.call([*sudocmd, "acbuild", "set-exec", "--", *execcommand])
            internals.call([*sudocmd, "acbuild", "write", aciname])
            internals.call([*sudocmd, "acbuild", "end"])
            return anise.framework.files.Filestructure(containerdir + "/" + aciname)
