# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for creation of Debian distribution packages.
"""

import datetime
import gzip
import os
import math
import shutil

import anise.framework.exceptions
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


class internals:

    @staticmethod
    def builddebpackage(*, name, version, image, licensename, authors, email, summary, description, section,
                        architecture, websiteurl, menuentries, executablelinks, requireddependencies,
                        stronglyrecommendeddependencies, suggesteddependencies, services, rawpkgpath, projectfsroot,
                        prerm="", postinst=""):
        """!
        Builds a Debian package (.deb).
        @param name: The package name.
        @param version: The package version string.
        @param image: A project screenshot
        @param licensename: The name of the project license.
        @param authors: Project authors as list of strings.
        @param email: Project email adress.
        @param summary: Short project description.
        @param description: Longer project description.
        @param section: The Debian section name.
        @param architecture: The Debian architecture name.
        @param websiteurl: The project homepage url.
        @param menuentries: Menu entries to register (list of MenuEntry).
        @param executablelinks: Symlinks to executables to be created (string/string dictionary).
        @param requireddependencies: List of required dependencies.
        @param stronglyrecommendeddependencies: List of dependencies, which are not strongly required but very typical.
        @param suggesteddependencies: List of optional dependencies.
        @param services: List of services to be registered.
        @param rawpkgpath: The raw source directory; can be damaged.
        @param projectfsroot: The project source directory path.
        @param prerm: Bash code snippet executed before the package is removed on a target machine.
        @param postinst: Bash code snippted executed after the package is installed on a target machine.
        """
        if anise.utils.basic.call("fakeroot -v", shell=True)[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'fakeroot' seems to be unavailable")
        if anise.utils.basic.call("dpkg --version", shell=True)[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'dpkg' seems to be unavailable")
        sdescription = "\n".join([" " + x for x in description.split("\n") if x.strip() != ""])
        conffiles = []
        sdepends = ("\nDepends: " + (", ".join(requireddependencies))) if len(requireddependencies) > 0 else ""
        srecommends = ("\nRecommends: " + (", ".join(stronglyrecommendeddependencies))) \
            if len(stronglyrecommendeddependencies) > 0 else ""
        ssuggests = ("\nSuggests: " + (", ".join(suggesteddependencies))) if len(suggesteddependencies) > 0 else ""
        workdir = anise.framework.files.TempDir().path()
        iworkdir = workdir + "/pkgtemp"
        packagerootname = "{name}-{version}-{architecture}".format(**locals())
        packageroot = "{iworkdir}/fs/{packagerootname}".format(**locals())
        placedoc = packageroot + "/usr/share/doc/" + name
        placepixmaps = packageroot + "/usr/share/pixmaps"
        placedebian = packageroot + "/DEBIAN"
        os.makedirs(iworkdir)
        os.makedirs(placedoc)
        if not rawpkgpath.endswith("/"):
            rawpkgpath += "/"
        _files = [rawpkgpath]
        while len(_files) > 0:
            f = _files.pop()
            if os.path.isdir(f):
                for x in os.listdir(f):
                    _files.append(f + "/" + x)
            else:
                ff = f[len(rawpkgpath) + 1:]
                gg = packageroot + "/" + ff
                dd = os.path.dirname(gg)
                if not os.path.exists(dd):
                    os.makedirs(dd)
                shutil.copy2(f, gg)
        if not os.path.isdir(packageroot + "/usr/bin"):
            os.makedirs(packageroot + "/usr/bin")
        for executablelink in executablelinks.keys():
            os.symlink(executablelinks[executablelink], packageroot + "/usr/bin/" + executablelink)
        licensetext = """
Format: http://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: {name}
Source: {websiteurl}
Upstream-Contact: The {name} Team <{email}>

Files: *
Copyright: {year} {authors}
License: {licensename}
"""[1:].format(name=name, websiteurl=websiteurl, email=email,
                   licensename=licensename, year=datetime.datetime.now().strftime("%Y"), authors=authors)
        with open(placedoc + "/copyright", "w") as f:
            f.write(licensetext)
        changelog = """
{name} {version} unstable; urgency=low

* New upstream release.
* See website for details.

-- The {name} Team <{email}>  Mon, 14 Jan 2013 13:37:00 +0000
"""[1:].format(**locals())
        cchangelog = gzip.compress(changelog.encode("utf-8"))
        with open(placedoc + "/changelog.gz", "wb") as f:
            f.write(cchangelog)
        os.makedirs(placepixmaps)
        os.makedirs(packageroot + "/usr/share/menu")
        os.makedirs(packageroot + "/usr/share/applications")
        for menuentry in menuentries:
            sdebiancategory = menuentry.category[0]
            sfreedesktopcategory = menuentry.category[1]
            ename = menuentry.name
            etitle = menuentry.title
            cmd = menuentry.command
            cmd2 = cmd.replace('"', '\\"')
            iconsrc = projectfsroot + "/" + menuentry.icon
            iconfname = "{name}.{ename}.png".format(**locals())
            shutil.copyfile(iconsrc, placepixmaps + "/" + iconfname)
            sisterminal = "false" if menuentry.gui else "true"
            sneeds = "X11" if menuentry.gui else "text"
            with open(packageroot + "/usr/share/menu/{ename}".format(ename=ename), "w") as f:
                f.write("""
?package({ename}): \\
  command="{cmd2}" \\
  needs="{sneeds}" \\
  section="{sdebiancategory}" \\
  title="{etitle}" \\
  icon="/usr/share/pixmaps/{iconfname}"
"""[1:].format(**locals()))
            with open(packageroot + "/usr/share/applications/{ename}.desktop".format(ename=ename), "w") as f:
                f.write("""
[Desktop Entry]
Name={etitle}
Exec={cmd}
Terminal={sisterminal}
Type=Application
Categories={sfreedesktopcategory};
Icon=/usr/share/pixmaps/{iconfname}
"""[1:].format(**locals()))
        os.makedirs(packageroot + "/etc/init")
        startservicescall = ""
        stopservicescall = ""
        for service in services:
            servicename = service.name
            servicecommand = service.command
            with open(packageroot + "/etc/init/{servicename}.conf".format(name=name, servicename=servicename),
                      "w") as f:
                f.write("""
# {name} - {name} job file

description "{name} service '{servicename}'"
author "The {name} Team <{email}>"

start on runlevel [2345]

stop on runlevel [016]

exec {servicecommand}
"""[1:].format(**locals()))
            startservicescall += "service {servicename} start\n"
            stopservicescall += "service {servicename} stop &>/dev/null\n"
            conffiles.append("/etc/init/{servicename}.conf".format(**locals()))
        size = 0
        for dirpath, dirnames, filenames in os.walk(packageroot):
            for f in filenames:
                ff = dirpath + "/" + f
                if not os.path.islink(ff):
                    size += os.path.getsize(ff)
        size = int(math.ceil(size / 1024.))
        os.makedirs(placedebian)
        # license
        with open(placedebian + "/copyright", "w") as f:
            f.write(licensetext)
        # prerm
        with open(placedebian + "/prerm", "w") as f:
            f.write("""
{stopservicescall}
set -e
{prerm}
"""[1:].format(**locals()))
        # postinst
        with open(placedebian + "/postinst", "w") as f:
            f.write("""
#!/bin/bash
set -e
{postinst}
{startservicescall}
if test -x /usr/bin/update-menus; then update-menus; fi
"""[1:].format(**locals()))
        # control
        with open(placedebian + "/control", "w") as f:
            f.write("""
Package: {name}
Version: {version}
Section: {section}
Priority: optional
Architecture: {architecture}{sdepends}{srecommends}{ssuggests}
Installed-Size: {size}
Maintainer: The {name} Team <{email}>
Provides: {name}
Homepage: {websiteurl}
Description: {summary}
{sdescription}
"""[1:].format(**locals()))
        # conffiles
        with open(placedebian + "/conffiles", "w") as f:
            f.write("\n".join(conffiles))
        anise.utils.basic.call("chmod", "-R", "u+rw,g+r,g-w,o+r,o-w", packageroot)
        anise.utils.basic.call("chmod", "-R", "0755", placedebian)
        anise.utils.basic.call("chmod", "0644", placedebian + "/conffiles")
        debfilename = "{name}_{version}_{architecture}.deb".format(**locals())
        r, o = anise.utils.basic.call(["fakeroot", "dpkg", "-b", packageroot, workdir + "/" + debfilename],
                                      raise_on_errors="unable to create debian package")
        anise.utils.logging.logdebug(o)
        return anise.framework.files.Filestructure(workdir + "/" + debfilename)


class Category:
    """!
    Enumeration for Debian's categories for software packages.
    """
    ApplicationsAccessibility = ("Applications/Accessibility", "System")
    ApplicationsAmateurradio = ("Applications/Amateur Radio", "Utility")
    ApplicationsDatamanagement = ("Applications/Data Management", "System")
    ApplicationsEditors = ("Applications/Editors", "Utility")
    ApplicationsEducation = ("Applications/Education", "Education")
    ApplicationsEmulators = ("Applications/Emulators", "System")
    ApplicationsFilemanagement = ("Applications/File Management", "System")
    ApplicationsGraphics = ("Applications/Graphics", "Graphics")
    ApplicationsMobiledevices = ("Applications/Mobile Devices", "Utility")
    ApplicationsNetwork = ("Applications/Network", "Network")
    ApplicationsNetworkCommunication = ("Applications/Network/Communication", "Network")
    ApplicationsNetworkFiletransfer = ("Applications/Network/File Transfer", "Network")
    ApplicationsNetworkMonitoring = ("Applications/Network/Monitoring", "Network")
    ApplicationsNetworkWebbrowsing = ("Applications/Network/Web Browsing", "Network")
    ApplicationsNetworkWebnews = ("Applications/Network/Web News", "Network")
    ApplicationsOffice = ("Applications/Office", "Office")
    ApplicationsProgramming = ("Applications/Programming", "Development")
    ApplicationsProjectmanagement = ("Applications/Project Management", "Development")
    ApplicationsScience = ("Applications/Science", "Education")
    ApplicationsScienceAstronomy = ("Applications/Science/Astronomy", "Education")
    ApplicationsScienceBiology = ("Applications/Science/Biology", "Education")
    ApplicationsScienceChemistry = ("Applications/Science/Chemistry", "Education")
    ApplicationsScienceDataanalysis = ("Applications/Science/Data Analysis", "Education")
    ApplicationsScienceElectronics = ("Applications/Science/Electronics", "Education")
    ApplicationsScienceEngineering = ("Applications/Science/Engineering", "Education")
    ApplicationsScienceGeoscience = ("Applications/Science/Geoscience", "Education")
    ApplicationsScienceMathematics = ("Applications/Science/Mathematics", "Education")
    ApplicationsScienceMedicine = ("Applications/Science/Medicine", "Education")
    ApplicationsSciencePhysics = ("Applications/Science/Physics", "Education")
    ApplicationsScienceSocial = ("Applications/Science/Social", "Education")
    ApplicationsShells = ("Applications/Shells", "System")
    ApplicationsSounds = ("Applications/Sound", "AudioVideo")
    ApplicationsSystem = ("Applications/System", "System")
    ApplicationsSystemAdministration = ("Applications/System/Administration", "System")
    ApplicationsSystemHardware = ("Applications/System/Hardware", "System")
    ApplicationsSystemLanguageenvironment = ("Applications/System/Language Environment", "System")
    ApplicationsSystemMonitoring = ("Applications/System/Monitoring", "System")
    ApplicationsSystemPackagemanagement = ("Applications/System/Package Management", "System")
    ApplicationsSystemSecurity = ("Applications/System/Security", "System")
    ApplicationsTerminalemulators = ("Applications/Terminal Emulators", "System")
    ApplicationsText = ("Applications/Text", "Utility")
    ApplicationsTvandradio = ("Applications/TV and Radio", "AudioVideo")
    ApplicationsViewers = ("Applications/Viewers", "Utility")
    ApplicationsVideo = ("Applications/Video", "AudioVideo")
    ApplicationsWebdevelopment = ("Applications/Web Development", "Development")
    GamesAction = ("Games/Action", "Game")
    GamesAdventure = ("Games/Adventure", "Game")
    GamesBlocks = ("Games/Blocks", "Game")
    GamesBoard = ("Games/Board", "Game")
    GamesCard = ("Games/Card", "Game")
    GamesPuzzles = ("Games/Puzzles", "Game")
    GamesSimulation = ("Games/Simulation", "Game")
    GamesStrategy = ("Games/Strategy", "Game")
    GamesTools = ("Games/Tools", "Game")
    GamesToys = ("Games/Toys", "Game")
    Help = ("Help", "System")
    ScreenSaving = ("Screen/Saving", "System")
    ScreenLocking = ("Screen/Locking", "System")


class ServiceDescription:
    """!
    Description for Debian services to be included in a package.
    """

    def __init__(self, name, command):
        """!
        @param name: The display name.
        @param command: The command to be executed.
        """
        self.name = name
        self.command = command


class MenuEntry:
    """!
    Specification for one menu entry added by a Debian installation package.
    """

    def __init__(self, *, name, title, category, command, gui, icon):
        """!
        @param name: Internal menu entry name.
        @param title: Menu entry Title.
        @param category: Menu entry category.
        @param command: Menu entry command.
        @param gui: If the command opens a gui (instead of a terminal application).
        @param icon: The path to a menu entry icon.
        """
        self.name = name
        self.title = title
        self.category = category
        self.command = command
        self.gui = gui
        self.icon = icon


def debpackage(source, dependencies=None, executablelinks=None, menuentries=None, services=None,
               prerm="", postinst="", architecture="all"):
    """!
    Builds a Debian package.
    @param source: An anise.framework.files.Filestructure.
    @param dependencies: A list of Debian dependencies for installation on target machines.
    @param executablelinks: A dict for links to executables for deployment on target machines.
    @param menuentries: A list of menu entries for creation on target machines (list of MenuEntry).
    @param services: A list of services for deployment on target machines.
    @param prerm: Piece of bash script executed before removal of the package on target machines.
    @param postinst: Piece of bash script executed after installation of the package on the target machines.
    @param architecture: The target architecture name.
    @return: An anise.framework.files.Filestructure.
    """
    with universe.executionscope.newsubscope("Creating debian package"):
        srcsrcfs = universe.packages.makerawpackage().mv(universe.name)
        for src, dst in [("README", "/usr/share/doc/{name}/README"),
                         ("README.pdf", "/usr/share/doc/{name}/README.pdf")]:
            s = srcsrcfs.path() + "/" + src
            d = os.path.abspath(source.path() + "/" + dst.format(name=universe.name))
            if os.path.isfile(s):
                anise.utils.basic.mkdirs(os.path.dirname(d))
                srcsrcfs.dl(subpath=src, to=d)
        executablelinks = executablelinks or {}
        menuentries = menuentries or []
        services = services or []
        dependencies = list(dependencies or []) + universe.dependencies.pool.list
        dependencies = [x for x in dependencies if hasattr(x, "debian")]
        reqdependencies = [x for x in dependencies if (x.deptype in [universe.dependencies.Type.Required,
                                                                     universe.dependencies.Type.Included])]
        recdependencies = [x for x in dependencies
                           if (x.deptype == universe.dependencies.Type.Required_HasAlternatives)]
        sugdependencies = [x for x in dependencies
                           if (x.deptype == universe.dependencies.Type.Recommended_HasAlternatives)]
        reqdependencies = [dependency for dependencyitem in reqdependencies for dependency in dependencyitem.debian]
        recdependencies = [dependency for dependencyitem in recdependencies for dependency in dependencyitem.debian]
        sugdependencies = [dependency for dependencyitem in sugdependencies for dependency in dependencyitem.debian]
        return internals.builddebpackage(
            name=universe.name, version=universe.packages.getversion(), image="image",
            licensename=universe.base.license.namedebian, authors=universe.base.authors[0],
            email=universe.base.email, summary=universe.base.summary, description=universe.base.longdescription,
            section="web", menuentries=menuentries, executablelinks=executablelinks,
            requireddependencies=reqdependencies, stronglyrecommendeddependencies=recdependencies,
            suggesteddependencies=sugdependencies, services=services, architecture=architecture,
            websiteurl=universe.base.websiteurl, rawpkgpath=source.path(),
            projectfsroot=universe.specialpaths.projectdir, prerm=prerm, postinst=postinst)


def only_programfiles(source):
    """!
    Returns a filtered set of files for a Debian program directory (no documentation, license, tests, ...).
    @param source: An anise.framework.files.Filestructure.
    @return: A filtered anise.framework.files.Filestructure.
    """
    srcfs = source.with_modified_rootname("x")
    for remv in ["README", "README.pdf", "LICENSE", "_meta", "_test"]:
        x = srcfs.path() + "/" + remv
        if os.path.isfile(x):
            os.unlink(x)
        elif os.path.isdir(x):
            shutil.rmtree(x)
    return srcfs
