# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for creation of oci container images.
"""

import shlex

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


class internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.packages.oci.fullnamepattern = getattr(universe.packages.oci, "fullnamepattern", "{name}")

    @staticmethod
    def call(cmd, raise_on_errors="unable to run buildah"):
        r, o = anise.utils.basic.call(cmd, raise_on_errors=raise_on_errors)
        anise.utils.logging.logdebug(o)
        return o


class PortDescription:
    """!
    Description for Oci network ports to be added to the container image.
    """

    def __init__(self, port):
        """!
        @param port: The port number as available inside the container.
        """
        self.port = int(port)


class EnvironmentVariableDescription:
    """!
    Description for Oci environment variables to be added to the container image.
    """

    def __init__(self, name, value):
        """!
        @param name: The variable name.
        @param value: The value.
        """
        self.name = name
        self.value = value


def ocidebianpackage(debianpackage, fullname=None, ports=None, environment_variables=None, execcommand=None,
                      run=None, run_late=None):
    """!
    Builds an oci container image based on Debian.
    
    You need an existing Debian package source.
    @param debianpackage: An anise.framework.files.Filestructure containing the debian package.
    @param fullname: The full container name.
    @param ports: A list of PortDescription port descriptions to add to the container image.
    @param environment_variables: A list of EnvironmentVariableDescription descriptions to add to the container image.
    @param execcommand: A list of strings to specify the command to execute inside the container after boot.
    @param run: A list of command-line lists to execute for preparation.
    @param run_late: A list of command-line lists to execute for preparation after installing the package.
    @return: An anise.framework.files.Filestructure.
    """
    with universe.executionscope.newsubscope("Creating oci container image"):
        if anise.utils.basic.call(["buildah", "--version"])[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'buildah' seems to be unavailable")
        if anise.utils.basic.call(["podman", "--version"])[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'podman' seems to be unavailable")
        sauthors = ", ".join(universe.base.authors)
        projectversion = universe.packages.getversion()
        containerdir = anise.framework.files.TempDir().path()
        imgname = (fullname or universe.packages.oci.fullnamepattern).format(name=universe.name)
        ports = ports or []
        environment_variables = environment_variables or []
        run = run or []
        run_late = run_late or []
        with anise.utils.basic.ChDir(containerdir):
            cn = internals.call(["buildah", "from", "docker.io/debian:latest"]).strip()
            ociname = "{name}-{version}.oci.tar".format(name=imgname, version=projectversion)
            internals.call(["buildah", "copy", cn, debianpackage.path(), "/root/pkg.deb"])
            internals.call(["buildah", "run", cn, "--", "apt-get", "update"])
            for l in run:
                internals.call(["buildah", "run", cn, "--", *l], raise_on_errors="unable to run 'run' stuff")
            for e in environment_variables:
                internals.call(["buildah", "config", "--env", e.name + "=" + e.value, cn])
            internals.call(["buildah", "run", cn, "--", "/usr/bin/dpkg", "-i", "/root/pkg.deb"],
                           raise_on_errors=None)
            internals.call(["buildah", "run", cn, "--", "apt-get", "-f", "-y", "install"])
            internals.call(["buildah", "config", "--author", sauthors, cn])
            for p in ports:
                internals.call(["buildah", "config", "--port", str(p.port), cn])
            for l in run_late:
                internals.call(["buildah", "run", cn, "--", *l], raise_on_errors="unable to run 'run_late' stuff")
            if execcommand:
                sexecmd = " ".join([shlex.quote(x) for x in execcommand])
                internals.call(["buildah", "config", "--entrypoint", sexecmd, cn])
            internals.call(["buildah", "commit", cn, imgname])
            internals.call(["podman", "save", "-o", ociname, imgname])
            internals.call(["buildah", "rmi", "--force", imgname])
            return anise.framework.files.Filestructure(containerdir + "/" + ociname)
