# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for creation of Python distribution packages.
"""

import os

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

packageswin32 = anise.framework.features.loadfeature("packages.win32").featuremodule


class internals:

    @staticmethod
    def _isnotpythonfile(fullpath):
        """!
        Returns `True` iff the filename does not end with '.py'.
        """
        return not fullpath.endswith(".py")

    @staticmethod
    def buildpythonwheelpackage(*, name, version, homepage, description, long_description, licensename, maturity,
                                authors, email, keywords, fsroot, applicationlinks, requirements, pythonminversion,
                                datafileacceptor=None):
        """!
        Builds a Python wheel package.
        @param name: The package name.
        @param version: The package version string.
        @param homepage: The homepage url.
        @param description: A short project description.
        @param long_description: A longer project description.
        @param licensename: The name of the project license.
        @param maturity: The project maturity string (may be `None`).
        @param authors: The project authors as list of string.
        @param email: The contact email address.
        @param keywords: Project keywords for search services.
        @param fsroot: The root directory for the package content (may become damaged during the process).
        @param applicationlinks: Application symlinks to register (list of PythonWheelPackageApplicationLink).
        @param requirements: List of packages which are required for installation on the target system (in wheel 
                             notation).
        @param pythonminversion: Minimal Python version.
        @param datafileacceptor: An optional function which helps deciding if a file in the source folder is a data
                                 file.
        """
        datafileacceptor = datafileacceptor or internals._isnotpythonfile
        try:
            import setuptools as _foo_test_setuptools
        except ImportError as exc:
            raise anise.framework.exceptions.RequirementsMissingError("'setuptools' seem to be unavailable", exc)
        if anise.utils.basic.call("wheel version", shell=True)[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'wheel' seems to be unavailable")
        setuppy = fsroot + "/setup.py"
        if os.path.exists(setuppy):
            raise anise.framework.exceptions.BadFormatError("There must be no setup.py in the project root.")
        sauthors = ", ".join(authors[:-1]) + (" & " if (len(authors) > 1) else "") + authors[-1]
        smaturity = "'{0}',".format(maturity) if maturity else ""
        rb = "{"
        lb = "}"
        pkgdirs = [fsroot.replace("\\", "/")]
        ppkgdirs = []
        datafilestring = ""
        while len(pkgdirs) > 0:
            pkgdir = pkgdirs.pop()
            for pkgfile in os.listdir(pkgdir):
                fpkgfile = pkgdir + "/" + pkgfile
                if os.path.isdir(fpkgfile):
                    pkgdirs.append(fpkgfile)
                    if os.path.exists(fpkgfile + "/__init__.py"):
                        ppkgdirs.append(fpkgfile)
        for ppkgdir in ppkgdirs:
            pkgname = ppkgdir[len(fsroot) + 1:].replace("/", ".")
            ddfiles = []
            for dfile in os.listdir(ppkgdir):
                fdfile = ppkgdir + "/" + dfile
                if os.path.isfile(fdfile) and datafileacceptor(fdfile):
                    ddfiles.append(dfile)
            if len(ddfiles) > 0:
                gffiles = ["'" + x.replace("'", "\\'") + "'" for x in ddfiles]
                datafilestring += "'{pkgname}': [{filesstring}],\n".format(pkgname=pkgname,
                                                                           filesstring=",".join(gffiles))
        consolescriptsstring = ""
        guiscriptsstring = ""
        if applicationlinks:
            for applicationlink in applicationlinks:
                newscr = "'{0}={1}:{2}',\n".format(applicationlink.linkname, applicationlink.module,
                                                   applicationlink.method)
                if applicationlink.gui:
                    guiscriptsstring += newscr
                else:
                    consolescriptsstring += newscr
        sinstallrequires = repr(requirements or [])
        slong_description = repr(long_description)
        with open(setuppy, "w") as f:
            f.write("""
from setuptools import setup, find_packages
setup(
name='{name}',
version='{version}',
description='{description}',
long_description={slong_description},
description_content_type='text/plain',
long_description_content_type='text/plain',
url='{homepage}',
author='{sauthors}',
author_email='{email}',
license='{licensename}',
classifiers=[
{smaturity}
'Programming Language :: Python :: {pythonminversion}',
],
keywords='{keywords}',
packages=find_packages(exclude=['contrib', 'docs', 'tests*', '_meta']),
install_requires={sinstallrequires},
package_data={rb}
{datafilestring}
{lb},
entry_points={rb}
'console_scripts': [
{consolescriptsstring}
],
'gui_scripts': [
{guiscriptsstring}
],
{lb},
)
    """.format(**locals()))
        with anise.utils.basic.ChDir(fsroot):
            r, o = anise.utils.basic.call("python3", "setup.py", "bdist_wheel", "--python-tag", "py3",
                                          raise_on_errors="unable to create python wheel")
            anise.utils.logging.logdebug(o)
        for f in os.listdir(fsroot + "/dist/"):
            return anise.framework.files.Filestructure(fsroot + "/dist/" + f)

    @staticmethod
    @anise.framework.features.hookhandler(packageswin32.HOOK_WINDOWS_PACKAGE_NSIS_GLOBAL)
    def _nsis_get_global():
        if hasattr(universe, "name"):
            welcometxt = "Welcome to the {} installation!".format(universe.name)
        else:
            welcometxt = "Welcome to the installation!"
        return """
            !macro RequirePython Version Url
                Var /Global AURL
                StrCpy $AURL "${Url}"
                begin:
                Var /Global ANISERV
                Var /Global PYPATH
                StrCpy $PYPATH ""
                StrCpy $ANISERV 1
                StrCpy $0 0
                aniseloop:
                    EnumRegKey $1 HKLM "Software\\Python\\PythonCore" $0
                    StrCmp $1 "" anisedone
                    IntOp $0 $0 + 1
                    ${StrLoc} $3 "$1." "${Version}." ">"
                    ${If} $3 == '0'
                        ReadRegStr $2 HKLM "Software\\Python\\PythonCore\\$1\\InstallPath" ""
                        StrCpy $PYPATH $2
                    ${EndIf}
                    Goto aniseloop
                anisedone:
                ExecWait '$PYPATH\\python -V' $ANISERV
                ${If} $ANISERV != '0'
                    ${If} "$AURL" != ''
                        MessageBox MB_OKCANCEL \"""" + welcometxt + """$\\r$\\n$\\r$\\nThis package needs an 'x86' variant of 'Python${Version}' for running. Do you want to automatically download and install it now?$\\r$\\n$\\r$\\nIf that fails, or if you cancel here, please install it from $AURL (with all extras) and try again." IDOK autoinst
                        Quit
                        autoinst:
                            !insertmacro DownloadHttp "$AURL" "$PLUGINSDIR\\apypkg.msi"
                            IfFileExists "$PLUGINSDIR\\apypkg.msi" dlsucc 0
                            MessageBox MB_OK "Download failed. Please install Python manually."
                            Quit
                            dlsucc:
                            ExecWait "cmd /c echo Please wait... This may take a while!&& msiexec /i $\\"$PLUGINSDIR\\apypkg.msi$\\" /qn ALLUSERS=1" $0
                            ${If} $0 != '0'
                                MessageBox MB_OK "The Python installation failed. Please try to install it manually and try again."
                                Quit
                            ${EndIf}
                            StrCpy $AURL ""
                            Goto begin
                    ${Else}
                        MessageBox MB_OK "This package needs an 'x86' variant of 'Python${Version}' for running.$\\r$\\n$\\r$\\nPlease install it from https://www.python.org/downloads/windows/ (with all extras) and try again."
                        Quit
                    ${EndIf}
                ${EndIf}

            !macroend
        """


class WheelApplicationLink:
    """!
    Specification for one program symlink added by a Python wheel package.
    """

    def __init__(self, *, linkname, module, method, gui):
        """!
        @param linkname: The name of the link as the user will be able to call it.
        @param module: The name of the module which contains the entry method.
        @param method: The name of the entry method.
        @param gui: If the command opens a gui (instead of a terminal application).
        """
        self.linkname = linkname
        self.module = module
        self.method = method
        self.gui = gui


def wheelpackage(source, applicationlinks=None, requirements=None, pythonminversion="3.7"):
    """!
    Builds a python wheel for distribution of Python applications.
    @param source: An anise.framework.files.Filestructure.
    @param applicationlinks: List of WheelApplicationLink.
    @param requirements: List of wheel dependencies to add (list of strings).
    @param pythonminversion: Minimal Python version.
    @return: An anise.framework.files.Filestructure.
    """
    with universe.executionscope.newsubscope("Creating python wheel"):
        readmefs = anise.framework.files.TempDir()
        readmefile = readmefs.path() + "/x"
        with anise.framework.files.TempDir() as tsourcedir:
            universe.documentation.internals.generate_userdocumentation_asplaintext(
                projectname=universe.name, head="Manual", sourcedir=tsourcedir.path(), targetfile=readmefile,
                documentations=universe.documentation.pool.gettexts(), targetdocumentation="FULLREADME",
                aliases=universe.documentation.snippets.toaliases())
        readmetext = anise.utils.basic.readfromfile(readmefile)
        return internals.buildpythonwheelpackage(
            name=universe.name, version=universe.packages.getversion(), homepage=universe.base.websiteurl,
            description=universe.base.summary, long_description=readmetext,
            licensename=universe.base.license.namepythonsetuputils,
            maturity=getattr(universe.base, "maturity", None), authors=universe.base.authors,
            email=universe.base.email, keywords=universe.base.keywords, fsroot=source.path(),
            applicationlinks=applicationlinks, requirements=requirements, pythonminversion=pythonminversion)
