# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for creation of Win32 distribution packages.
"""

import os
import math
import shutil

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


HOOK_WINDOWS_PACKAGE_NSIS_ONINIT = anise.framework.features.Hook(
    doc="This hook is called for collecting additional nsis win32 installer oninit scripts.",
    wizard_functionbody="return ''")

HOOK_WINDOWS_PACKAGE_NSIS_ONINSTALL = anise.framework.features.Hook(
    doc="This hook is called for collecting additional nsis win32 installer oninstall scripts.",
    wizard_functionbody="return ''")

HOOK_WINDOWS_PACKAGE_NSIS_GLOBAL = anise.framework.features.Hook(
    doc="This hook is called for collecting additional nsis win32 installer global scripts.",
    wizard_functionbody="return ''")


class internals:

    @staticmethod
    def buildwindowsinstallerpackage(
            *, name, version, licensefile, authors, email, summary,  architecture, websiteurl, menuentries, rawpkgpath,
            projectfsroot, projecticon=None, globalscript="", oninitscript="", oninstallscript="", usemodernui2=False):
        """!
        Builds a Windows installer executable (.exe).
        @param name: The package name.
        @param version: The package version string.
        @param licensefile: The path to the full text of the project's license.
        @param authors: Project authors as list of strings.
        @param email: Project email adress.
        @param summary: Short project description.
        @param architecture: The Debian architecture name.
        @param websiteurl: The project homepage url.
        @param menuentries: Menu entries to register (list of MenuEntry).
        @param rawpkgpath: The raw source directory; can be damaged.
        @param projectfsroot: The project source directory path.
        @param projecticon: Path to the project icon.
        @param globalscript: NSIS script global part.
        @param oninitscript: NSIS script which runs at initialization of the installer.
        @param oninstallscript: NSIS script which runs after file copying of the installer.
        @param usemodernui2: If Modern UI 2 shall be used (brings a more fresh look).
        """
        if anise.utils.basic.call("makensis -VERSION", shell=True)[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'makensis' seems to be unavailable")
        workdir = anise.framework.files.TempDir().path()
        nsifile = workdir + "/_anise_installer.nsi"
        exefilename = "{name}-{version}_{architecture}.exe".format(**locals())
        exefile = workdir + "/" + exefilename
        installsection = ""
        uninstallsection = ""
        uninstallsection2 = ""
        if not rawpkgpath.endswith("/"):
            rawpkgpath += "/"
        _files = [rawpkgpath]
        sizeb = 0.0
        while len(_files) > 0:
            f = _files.pop()
            ff = f[len(rawpkgpath) + 1:]
            ff2 = ff.replace("/", "\\")
            if os.path.isdir(f):
                for x in os.listdir(f):
                    _files.append(f + "/" + x)
                uninstallsection2 = "RmDir \"$0\\{ff2}\"\n".format(**locals()) + uninstallsection2
            elif not os.path.islink(f):
                sizeb += os.path.getsize(f)
                reldir = os.path.dirname(ff).replace("/", "\\")
                fsl = f.replace("/", "\\")
                installsection += "SetOutPath \"$INSTDIR\\{reldir}\"\nFile \"{fsl}\"\n".format(**locals())
                uninstallsection += "Delete \"$0\\{ff2}\"\n".format(**locals())
        tmpicons = anise.framework.files.TempDir()
        uninstallericon = ""
        if projecticon:
            iconsrc = projectfsroot + "/" + projecticon
            iconfname = "_{name}.ico".format(**locals())
            sizeb += os.path.getsize(iconsrc)
            fsl = tmpicons.path() + "/" + iconfname
            shutil.copyfile(iconsrc, fsl)
            fsl = fsl.replace("/", "\\")
            installsection += 'SetOutPath "$INSTDIR"\nFile "{fsl}"\n'.format(**locals())
            uninstallsection += 'Delete "$0\\{iconfname}"\n'.format(**locals())
            uninstallericon = "$INSTDIR\\" + iconfname
        for menuentry in menuentries:
            icon = ""
            if menuentry.icon:
                ename = menuentry.title
                iconsrc = projectfsroot + "/" + menuentry.icon
                iconfname = "_{name}.{ename}.ico".format(**locals())
                sizeb += os.path.getsize(iconsrc)
                fsl = tmpicons.path() + "/" + iconfname
                shutil.copyfile(iconsrc, fsl)
                fsl = fsl.replace("/", "\\")
                installsection += 'SetOutPath "$INSTDIR"\nFile "' + fsl + '"\n'
                uninstallsection += 'Delete "$0\\{iconfname}"\n'.format(**locals())
                icon = "$INSTDIR\\" + iconfname
            command = menuentry.command.replace("$$", "$INSTDIR")
            params = menuentry.params.replace("$$", "$INSTDIR")
            title = menuentry.title
            for x in '/?<>\:*|"':
                title = title.replace(x, "_")
            if menuentry.workdir:
                installsection += 'SetOutPath "' + menuentry.workdir + '"\n'
            installsection += "createShortCut \"$SMPROGRAMS\\{name}\\{title}.lnk\" \"{command}\" \"{params}\" \"{icon}\"\n" \
                .format(**locals())
            uninstallsection += "delete \"$SMPROGRAMS\\{name}\\{title}.lnk\"\n".format(**locals())
        sizekb = int(math.ceil(sizeb / 1024.))
        lb = "{"
        rb = "}"
        displayversion = version
        versionmajor = version.split(".")[0]
        try:
            versionminor = version.split(".")[1]
        except Exception:
            versionminor = "0"
        smodernui2 = ""
        if usemodernui2:
            smodernui2 = """
            !include "MUI2.nsh"
            !define MUI_ABORTWARNING
            !insertmacro MUI_PAGE_WELCOME
            !insertmacro MUI_LANGUAGE "English"
            """
        anise.utils.basic.writetofile(nsifile, """
            {smodernui2}
            RequestExecutionLevel admin
            InstallDir "$PROGRAMFILES\\{name}"
            LicenseData "{licensefile}"
            Name "{name}"
            Outfile "{exefilename}"
            !include "LogicLib.nsh"
            !include "StrFunc.nsh"
            !macro DownloadHttp Url File
                FileOpen $9 "$PLUGINSDIR\\dlpy.vbs" w
                FileWrite $9 "WScript.StdOut.WriteLine $\\"Please wait... This may take a while!$\\"$\\r$\\n"
                FileWrite $9 "dim xHttp: Set xHttp = createobject($\\"Microsoft.XMLHTTP$\\")$\\r$\\n"
                FileWrite $9 "dim bStrm: Set bStrm = createobject($\\"Adodb.Stream$\\")$\\r$\\n"
                FileWrite $9 "xHttp.Open $\\"GET$\\", $\\"${lb}Url{rb}$\\", False$\\r$\\n"
                FileWrite $9 "xHttp.Send$\\r$\\n"
                FileWrite $9 "with bStrm$\\r$\\n"
                FileWrite $9 "    .type = 1$\\r$\\n"
                FileWrite $9 "    .open$\\r$\\n"
                FileWrite $9 "    .write xHttp.responseBody$\\r$\\n"
                FileWrite $9 "    .savetofile $\\"${lb}File{rb}$\\", 2$\\r$\\n"
                FileWrite $9 "end with$\\r$\\n"
                FileClose $9
                SetOutPath "C:\\"
                ExecWait 'cmd /c cscript /NoLogo "$PLUGINSDIR\\dlpy.vbs"'
            !macroend
            ${lb}StrLoc{rb} ; needed, unless something is missing for the compiler
            page license
            page directory
            Page instfiles
            !macro VerifyUserIsAdmin
                UserInfo::GetAccountType
                pop $0
                ${lb}If{rb} $0 != "admin" ;Require admin rights on NT4+
                    messageBox mb_iconstop "Administrator rights required!"
                    setErrorLevel 740 ;ERROR_ELEVATION_REQUIRED
                    quit
                ${lb}EndIf{rb}
            !macroend
            Var A_PATH_DESKTOP
            !define CSIDL_DESKTOP '0x0'
            Var A_PATH_PROGRAMS
            !define CSIDL_PROGRAMS '0x2'
            Var A_PATH_PERSONAL
            !define CSIDL_PERSONAL '0x5'
            Var A_PATH_FAVORITES
            !define CSIDL_FAVORITES '0x6'
            Var A_PATH_STARTUP
            !define CSIDL_STARTUP '0x7'
            Var A_PATH_RECENT
            !define CSIDL_RECENT '0x8'
            Var A_PATH_SENDTO
            !define CSIDL_SENDTO '0x9'
            Var A_PATH_STARTMENU
            !define CSIDL_STARTMENU '0xB'
            Var A_PATH_MUSIC
            !define CSIDL_MUSIC '0xD'
            Var A_PATH_DESKTOPDIR
            !define CSIDL_DESKTOPDIR '0x10'
            Var A_PATH_COMPUTER
            !define CSIDL_COMPUTER '0x11'
            Var A_PATH_FONTS
            !define CSIDL_FONTS '0x14'
            Var A_PATH_TEMPLATES
            !define CSIDL_TEMPLATES '0x15'
            Var A_PATH_APPDATA
            !define CSIDL_APPDATA '0x1A'
            Var A_PATH_LOCALAPPDATA
            !define CSIDL_LOCALAPPDATA '0x1C'
            Var A_PATH_INTERNETCACHE
            !define CSIDL_INTERNETCACHE '0x20'
            Var A_PATH_COOKIES
            !define CSIDL_COOKIES '0x21'
            Var A_PATH_HISTORY
            !define CSIDL_HISTORY '0x22'
            Var A_PATH_COMMONAPPDATA
            !define CSIDL_COMMONAPPDATA '0x23'
            Var A_PATH_SYSTEM
            !define CSIDL_SYSTEM '0x25'
            Var A_PATH_PROGRAMFILES
            !define CSIDL_PROGRAMFILES '0x26'
            Var A_PATH_MYPICTURES
            !define CSIDL_MYPICTURES '0x27'
            Var A_PATH_COMMONPROGRAMFILES
            !define CSIDL_COMMONPROGRAMFILES '0x2B'
            !macro GetSpecialPaths
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_DESKTOP{rb}, i0)i.r0'
                StrCpy $A_PATH_DESKTOP "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_PROGRAMS{rb}, i0)i.r0'
                StrCpy $A_PATH_PROGRAMS "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_PERSONAL{rb}, i0)i.r0'
                StrCpy $A_PATH_PERSONAL "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_FAVORITES{rb}, i0)i.r0'
                StrCpy $A_PATH_FAVORITES "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_STARTUP{rb}, i0)i.r0'
                StrCpy $A_PATH_STARTUP "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_RECENT{rb}, i0)i.r0'
                StrCpy $A_PATH_RECENT "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_SENDTO{rb}, i0)i.r0'
                StrCpy $A_PATH_SENDTO "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_STARTMENU{rb}, i0)i.r0'
                StrCpy $A_PATH_STARTMENU "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_MUSIC{rb}, i0)i.r0'
                StrCpy $A_PATH_MUSIC "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_DESKTOPDIR{rb}, i0)i.r0'
                StrCpy $A_PATH_DESKTOPDIR "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_COMPUTER{rb}, i0)i.r0'
                StrCpy $A_PATH_COMPUTER "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_FONTS{rb}, i0)i.r0'
                StrCpy $A_PATH_FONTS "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_TEMPLATES{rb}, i0)i.r0'
                StrCpy $A_PATH_TEMPLATES "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_APPDATA{rb}, i0)i.r0'
                StrCpy $A_PATH_APPDATA "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_LOCALAPPDATA{rb}, i0)i.r0'
                StrCpy $A_PATH_LOCALAPPDATA "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_INTERNETCACHE{rb}, i0)i.r0'
                StrCpy $A_PATH_INTERNETCACHE "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_COOKIES{rb}, i0)i.r0'
                StrCpy $A_PATH_COOKIES "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_HISTORY{rb}, i0)i.r0'
                StrCpy $A_PATH_HISTORY "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_COMMONAPPDATA{rb}, i0)i.r0'
                StrCpy $A_PATH_COMMONAPPDATA "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_SYSTEM{rb}, i0)i.r0'
                StrCpy $A_PATH_SYSTEM "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_PROGRAMFILES{rb}, i0)i.r0'
                StrCpy $A_PATH_PROGRAMFILES "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_MYPICTURES{rb}, i0)i.r0'
                StrCpy $A_PATH_MYPICTURES "$1"
                System::Call 'shell32::SHGetSpecialFolderPath(i $HWNDPARENT, t .r1, i ${lb}CSIDL_COMMONPROGRAMFILES{rb}, i0)i.r0'
                StrCpy $A_PATH_COMMONPROGRAMFILES "$1"
            !macroend
            {globalscript}
            function .onInit
                setShellVarContext all
                !insertmacro VerifyUserIsAdmin
                !insertmacro GetSpecialPaths
                {oninitscript}
            functionEnd
            section "install"
                setOutPath "$INSTDIR"
                createDirectory "$SMPROGRAMS\\{name}"
                {installsection}
                {oninstallscript}
                writeUninstaller "$INSTDIR\\uninstall.exe"
                # Registry information for add/remove programs
                WriteRegStr HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "DisplayName" "{name} - {summary}"
                WriteRegStr HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "UninstallString" "$INSTDIR\\uninstall.exe"
                WriteRegStr HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "QuietUninstallString" "$INSTDIR\\uninstall.exe /S"
                WriteRegStr HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "InstallLocation" "$INSTDIR"
                WriteRegStr HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "DisplayIcon" "{uninstallericon}"
                WriteRegStr HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "Publisher" "The {name} Team"
                WriteRegStr HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "HelpLink" "{websiteurl}"
                WriteRegStr HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "URLUpdateInfo" "{websiteurl}"
                WriteRegStr HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "URLInfoAbout" "{websiteurl}"
                WriteRegStr HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "DisplayVersion" "{displayversion}"
                WriteRegDWORD HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "VersionMajor" {versionmajor}
                WriteRegDWORD HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "VersionMinor" {versionminor}
                WriteRegDWORD HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "NoModify" 1
                WriteRegDWORD HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "NoRepair" 1
                WriteRegDWORD HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" "EstimatedSize" {sizekb}
            sectionEnd
            function un.onInit
                SetShellVarContext all
                MessageBox MB_OKCANCEL "Permanantly remove {name}?" IDOK next
                    Abort
                next:
                !insertmacro VerifyUserIsAdmin
                !insertmacro GetSpecialPaths
            functionEnd
            section "uninstall"
                ReadRegStr $0 HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}" InstallLocation
                {uninstallsection}
                {uninstallsection2}
                rmDir "$SMPROGRAMS\\{name}"
                delete "$0\\uninstall.exe"
                rmDir "$0"
                DeleteRegKey HKLM "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{name}"
            sectionEnd
        """.format(**locals()))

        r, o = anise.utils.basic.call(["makensis", nsifile],
                                      raise_on_errors="unable to create windows installer package")
        anise.utils.logging.logdebug(o)
        tmpicons.delete()
        return anise.framework.files.Filestructure(exefile)


class MenuEntry:
    """!
    Specification for one start menu entry added by a Windows package installer.
    """

    def __init__(self, *, title, command, icon=None, params="", workdir="$INSTDIR"):
        """!
        @param title: Menu entry title.
        @param command: Menu entry command.
        @param icon: The path to a menu entry icon.
        @param params: Additional command parameters (as string).
        @param workdir: The working directory for this menu entry.
        """
        self.title = title
        self.command = command
        self.icon = icon
        self.params = params
        self.workdir = workdir


def win32exepackage(source, architecture="all", menuentries=None, projecticon=None, oninitscript="",
                    oninstallscript="", usemodernui2=True):
    """!
    Builds a Win32 installation executable (as .exe file).
    @param source: An anise.framework.files.Filestructure.
    @param architecture: The target architecture name.
    @param menuentries: A list of menu entries for creation on target machines. See packages.win32.MenuEntry.
    @param projecticon: Path to a project icon. Relative to the project root directory.
    @param oninitscript: Additional NSIS on-init script.
    @param oninstallscript: Additional NSIS on-install script.
    @param usemodernui2: If Modern UI 2 shall be used (brings a more fresh look).
    @return: An anise.framework.files.Filestructure.
    """
    menuentries = menuentries or []
    with universe.executionscope.newsubscope("Creating win32 installation package"):
        _oninitscript = "\n".join([f() for f in universe.gethooks(HOOK_WINDOWS_PACKAGE_NSIS_ONINIT)]) \
                        + "\n" + oninitscript
        _oninstallscript = "\n".join([f() for f in universe.gethooks(HOOK_WINDOWS_PACKAGE_NSIS_ONINSTALL)]) \
                           + "\n" + oninstallscript
        _globalscript = "\n".join([f() for f in universe.gethooks(HOOK_WINDOWS_PACKAGE_NSIS_GLOBAL)])
        return internals.buildwindowsinstallerpackage(
            name=universe.name, version=universe.packages.getversion(),
            licensefile=universe.base.license.getfilename(), menuentries=menuentries,
            authors=universe.base.authors[0], email=universe.base.email, summary=universe.base.summary,
            projecticon=projecticon, architecture=architecture, websiteurl=universe.base.websiteurl,
            rawpkgpath=source.path(), projectfsroot=universe.specialpaths.projectdir, globalscript=_globalscript,
            oninitscript=_oninitscript, oninstallscript=_oninstallscript, usemodernui2=usemodernui2)


def deployqt5dlls(milieu, binsource=None, pluginssource=None):
    """!
    Returns the DLLs needed for Qt deployment on Windows.
    @param milieu: The milieu to use. See milieus.Milieu for details.
    @param binsource: The directory where the Qt library binaries are stored.
    @param pluginssource: The directory where the Qt plugins are stored.
    @return: A anise.framework.files.Filestructure.
    """
    milieu = universe.milieus.get(milieu, defaultmilieu={"class": "qt", "qtenv": "qt5"})
    filesystem = anise.framework.features.loadfeature("filesystem").featuremodule
    if binsource is None:
        binsource = anise.framework.files.Filestructure(milieu.bindir)
    if pluginssource is None:
        pluginssource = anise.framework.files.Filestructure(milieu.pluginsdir)
    imageformatssource = anise.framework.files.Filestructure(pluginssource.path() + "/imageformats")
    platformssource = anise.framework.files.Filestructure(pluginssource.path() + "/platforms")
    result = {"/": anise.framework.files.TaskExecution(universe.filesystem.filterbynamepattern,
                                       include=["qt5core.dll", "qt5gui.dll", "qt5concurrent.dll", "qt5webkit.dll",
                                                "qt5webkitwidgets.dll", "qt5widgets.dll", "qt5xml.dll",
                                                "qt5opengl.dll", "qt5printsupport.dll", "qt5positioning.dll",
                                                "qt5sql.dll", "qt5qml.dll", "qt5quick.dll", "qt5webchannel.dll",
                                                "qt5network.dll", "qt5multimedia.dll", "qt5multimediawidgets.dll",
                                                "qt5sensors.dll", "icu.*.dll", "libwinpthread.*.dll",
                                                "libstdc.*.dll", "libgcc.*.dll"],
                                       casesensitive=False, source=binsource),
              "/imageformats": anise.framework.files.TaskExecution(filesystem.filterbynamepattern, include=[
                  "qgif.dll", "qico.dll", "qjp2.dll", "qjpeg.dll", "qsvg.dll", "qtiff.dll", "qwebp.dll"],
                                                   casesensitive=False, source=imageformatssource),
              "/platforms": anise.framework.files.TaskExecution(filesystem.filterbynamepattern,
                                                                include=["qwindows.dll", ],
                                                casesensitive=False, source=platformssource)}
    return anise.framework.files.MergedFilestructure(frommap=result)
