# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for storing packages on disk, for the ability to publish not only the latest release but also older ones.
"""

import os
import pickle
import re
import shutil
import sys
import urllib.parse

import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

distributables = anise.framework.features.loadfeature("distributables").featuremodule


class internals:

    @staticmethod
    def sliceversion(versionstring):
        """!
        Slices a version string to a tuple.
        @param versionstring: Version as string.
        @return: Version tuple.
        """
        m = re.match("^([0-9]+)\\.([0-9]+)([^0-9].*)?$", versionstring)
        if m:
            return (int(m.group(1)), int(m.group(2))), (m.group(3) or "")
        else:
            return 0, versionstring

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION, provides=["PRE"])
    def _initproject():
        universe.packagestore.isenabled = False

    @staticmethod
    def quotename(s):
        """!
        Quotes a name for using as filename.
        @param s: A string to quote.
        @return: A quoted string which can be used as a filename.
        """
        return urllib.parse.quote(s, safe=" ")

    @staticmethod
    def unquotename(s):
        """!
        Unquotes a name quoted by internals::quotename() back.
        @param s: The quoted string.
        @return: The original name.
        """
        return urllib.parse.unquote(s)


def enable():
    """!
    Enables using the package store.
    """
    if universe.packagestore.isenabled:
        return
    universe.packagestore.isenabled = True
    if not hasattr(universe.packagestore, "path"):
        universe.packagestore.path = os.path.expanduser("~/.anise/packagestore")
    if not hasattr(universe.packagestore, "historylength"):
        universe.packagestore.historylength = 3
    if not hasattr(universe.packagestore, "sliceversion"):
        universe.packagestore.sliceversion = internals.sliceversion

    def _injectstored():
        dlgroupspath = universe.packagestore.path + "/" + universe.name
        anise.utils.basic.mkdirs(dlgroupspath)
        if not getattr(universe, "skippackagecreation", False):
            # store the original downloads to our store
            for dlgroup in universe.distributables.pool.list:
                universe.distributables.currentgroup = dlgroup
                dlgrouppath = dlgroupspath + "/" + internals.quotename(dlgroup.name)
                version = ".".join([str(x) for x in universe.packagestore.sliceversion(universe.base.version)[0]])
                patternpathfinal = dlgrouppath + "/" + version
                patternpath = patternpathfinal + "__anisetemp"
                dlgrouppathexisted = os.path.exists(dlgrouppath)
                anise.utils.basic.mkdirs(dlgrouppath)
                if dlgroup.description is not None:
                    with open(dlgrouppath + "/_description", "w") as f:
                        f.write(dlgroup.description)
                if dlgroup.linkto is not None:
                    with open(dlgrouppath + "/_linkto", "w") as f:
                        f.write(dlgroup.linkto)
                if os.path.exists(patternpath):
                    shutil.rmtree(patternpath)
                anise.utils.basic.mkdirs(patternpath)
                try:
                    for ifn, fn in enumerate(dlgroup.sources):
                        basename = os.path.basename(fn.path())
                        dlpath = patternpath + "/" + basename
                        fn.setdata("ctime", str(os.path.getctime(fn.path())))
                        fn.setdata("packagestore.order", ifn)
                        for dkey in fn.datakeys():
                            with open(patternpath + "/_anise_" + basename + "_" + dkey, "wb") as f:
                                v = fn.getdata(dkey, None)
                                if isinstance(v, str):
                                    v = v.encode()
                                else:
                                    v = pickle.dumps([v])
                                f.write(v)
                        fn.dl(to=dlpath)
                except Exception:
                    try:
                        shutil.rmtree(patternpath)
                    except:
                        pass
                    if dlgrouppathexisted:
                        anise.utils.logging.logwarning(
                            "There were problems building a package. The package store contains "
                            "such a package from a former run. Please try adding "
                            "'--b:skippackagecreation=1' to your command-line in order to skip "
                            "creating new packages.")
                    raise
                if os.path.exists(patternpathfinal):
                    shutil.rmtree(patternpathfinal)
                os.rename(patternpath, patternpathfinal)
            universe.distributables.currentgroup = None

        # replace the download list by everything we have stored
        origlist = universe.distributables.pool.list
        universe.distributables.pool.list = []

        for origdlgroup in origlist:
            dlgrouppath = dlgroupspath + "/" + internals.quotename(origdlgroup.name)
            if os.path.isdir(dlgrouppath):
                history = [x for x in os.listdir(dlgrouppath) if os.path.isdir(dlgrouppath + "/" + x)]
                history = [x for x in history if not x.endswith("__anisetemp")]
                history = [(universe.packagestore.sliceversion(x)[0], x) for x in history]
                history.sort(reverse=True)
                dlgrouplist = []
                fetchmore = True
                for _ctime, version in history[:universe.packagestore.historylength]:
                    patternpath = dlgrouppath + "/" + version
                    for fname in os.listdir(patternpath):
                        if fname.startswith("_anise_"):
                            continue
                        fpath = patternpath + "/" + fname
                        dl = anise.framework.files.Filestructure(fpath)
                        for fp in os.listdir(patternpath):
                            p = "_anise_" + fname + "_"
                            if fp.startswith(p):
                                keyname = fp[len(p):]
                                val = anise.utils.basic.readfromfile(patternpath + "/" + fp, binary=True)
                                class BadFormatError(Exception): pass
                                try:
                                    val = pickle.loads(val)
                                    if (not isinstance(val, list)) or len(val) != 1:
                                        raise BadFormatError()
                                    val = val[0]
                                except:
                                    val = val.decode()
                                dl.setdata(keyname, val)
                        dlgrouplist.append(dl)
                        if dl.getdata("packagestore_only_latest", False):
                            fetchmore = False
                    dlgrouplist.sort(key=lambda x: x.getdata("packagestore.order", sys.maxsize))
                    if not fetchmore:
                        break
                dlgdescription = anise.utils.basic.readfromfile(dlgrouppath + "/_description", notexist_default=None)
                dlglinkto = anise.utils.basic.readfromfile(dlgrouppath + "/_linkto", notexist_default=None)
                dlgroup = distributables.FileGroup(dlgrouplist, name=origdlgroup.name,
                                                   description=dlgdescription, linkto=dlglinkto)
                universe.distributables.pool.list.append(dlgroup)
    universe.addhook(distributables.HOOK_PREPARE_DISTRIBUTABLES_POOL, _injectstored,
                     provides=["packagestore", "POST"])
