# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

import anise.framework.features
import anise.framework.projects

from anise.framework.projects import universe

projectdescriptioneditor = anise.framework.features.loadfeature("projectdescriptioneditor").featuremodule


HOOK_GET_TESTING_CONFGUIDES = anise.framework.features.Hook(
    anise.framework.features.HookKind.CLASS, hide=True,
    doc="This hook is called for collecting all configuration guide helpers for testing implementations.")


class TestingImplementation:

    def __init__(self, label):
        self.label = label

    def execute(self, params):
        pass


@anise.framework.features.hookhandler(HOOK_GET_TESTING_CONFGUIDES)
class PyUnitTestingImplementation(TestingImplementation):

    def __init__(self):
        TestingImplementation.__init__(self, "PyUnit Test Class")

    def execute(self, params):
        testclass = universe.ui.userfeedback.input("Please specify the test class name.", "my.tests.TestClass")
        if testclass is None:
            return
        params["testclass"] = anise.framework.projects.IntermediateStructure.Value.fromstring(testclass)
        return "testing.pyunit.pyunittest"


@anise.framework.features.hookhandler(HOOK_GET_TESTING_CONFGUIDES)
class ExternalProgramTestingImplementation(TestingImplementation):

    def __init__(self):
        TestingImplementation.__init__(self, "External Program")

    def execute(self, params):
        command = universe.ui.userfeedback.input("Please specify the external program.", "_meta/tests/test1.sh")
        if command is None:
            return
        params["command"] = cmdlist = anise.framework.projects.IntermediateStructure.Value("list")
        cmdlist.appendarg(anise.framework.projects.IntermediateStructure.Value.fromstring(command))
        return "testing.generic.generictest"


@anise.framework.features.hookhandler(HOOK_GET_TESTING_CONFGUIDES)
class AutodiscoveredPyUnitTestingImplementation(TestingImplementation):

    def __init__(self):
        TestingImplementation.__init__(self, "PyUnit Autodiscover")

    def execute(self, params):
        testdir = universe.ui.userfeedback.input("Please specify the subdirectory to search for tests.", "/")
        if testdir is None:
            return
        while testdir.startswith("/"):
            testdir = testdir[1:]
        params["testdir"] = anise.framework.projects.IntermediateStructure.Value.fromstring(testdir)
        return "testing.pyunit.pyunitdiscoveredtest"


@anise.framework.features.hookhandler(projectdescriptioneditor.HOOK_GET_FEATURE_ACTIONS)
class ProjectDescriptionEditorAssistant(projectdescriptioneditor.ProjectDescriptionEditorAssistant):

    def __init__(self):
        projectdescriptioneditor.ProjectDescriptionEditorAssistant.__init__(
            self, "Manage tests", "Configure how to test if your current project implementation is working fine. Those "
                                  "tests can be executed for quality checks.", "testing")

    def getcustomactions(self, ais):
        return [projectdescriptioneditor.CustomAction("Add a new test package", self._add_testpackage,
                                                      icon="userfeedback/add.png")]

    def _add_testpackage(self, ais):
        packagename = universe.ui.userfeedback.input("Please enter the test package name.", "sometestpackage")
        if packagename is None:
            return
        name = universe.ui.userfeedback.input("Please enter a name for the new test.", "sometest")
        if name is None:
            return
        oprms = {}
        lstchoices = [x() for x in universe.gethooks(HOOK_GET_TESTING_CONFGUIDES)]
        choices = [x.label for x in lstchoices]
        ichoice = universe.ui.userfeedback.choice("Please choose the kind of test you want to add.", choices)
        if ichoice is None:
            return
        tcg = lstchoices[ichoice]
        implname = tcg.execute(oprms)
        if not implname:
            return
        testonrelease = universe.ui.userfeedback.messageyesno("Is the specified test a release-critical one?")
        if testonrelease:
            oprms["testonrelease"] = anise.framework.projects.IntermediateStructure.Value("bool", v="True")
        newlet = anise.framework.projects.IntermediateStructure.Let(
            k="testing.tests.add",
            packagename=anise.framework.projects.IntermediateStructure.Value.fromstring(packagename),
            name=anise.framework.projects.IntermediateStructure.Value.fromstring(name),
            runner=anise.framework.projects.IntermediateStructure.Value("objectref", oref=implname), **oprms)
        ais.addentry(newlet)

    def getentries(self, ais):
        r = []
        for e in ais.entries():
            if isinstance(e, anise.framework.projects.IntermediateStructure.Let) and e.key() == "testing.tests.add":
                r.append(e)
        return r

    def getentrylabel(self, ais, entry):
        return entry.getarg("packagename").value() + ":" + entry.getarg("name").value()


@anise.framework.features.hookhandler(
    projectdescriptioneditor.HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS)
class SetReleaseCriticalCustomAction(projectdescriptioneditor.CustomAction):

    def __init__(self):
        projectdescriptioneditor.CustomAction.__init__(self, "Make release-critical (currently set as non-critical)")

    def visible(self, ais, node):
        return isinstance(node, anise.framework.projects.IntermediateStructure.Let) \
               and node.key() == "testing.tests.add" and not (
            ("testonrelease" in node.args())
            and (anise.framework.projects.IntermediateStructure.Value.istrue(node.getarg("testonrelease").value())))

    def execute(self, ais, node):
        node.setarg("testonrelease", anise.framework.projects.IntermediateStructure.Value("bool", v="1"))


@anise.framework.features.hookhandler(
    projectdescriptioneditor.HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS)
class UnsetReleaseCriticalCustomAction(projectdescriptioneditor.CustomAction):

    def __init__(self):
        projectdescriptioneditor.CustomAction.__init__(self, "Make not release-critical (currently set as critical)")

    def visible(self, ais, node):
        return isinstance(node, anise.framework.projects.IntermediateStructure.Let) \
               and node.key() == "testing.tests.add" and (("testonrelease" in node.args())
            and (anise.framework.projects.IntermediateStructure.Value.istrue(node.getarg("testonrelease").value())))

    def execute(self, ais, node):
        node.removearg("testonrelease")
