# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

import anise.framework.features
import anise.framework.projects
import anise.utils

from anise.framework.projects import universe

projectdescriptioneditor = anise.framework.features.loadfeature("projectdescriptioneditor").featuremodule


## Called in order to get version control system implementations.
HOOK_GET_VERSIONCONTROLSYSTEM_IMPLEMENTATIONS = anise.framework.features.Hook(
    hide=True, doc="This hook is called for collecting version control system implementations.")


def getversioncontrolimplementations():
    return [Vcs(*x()) for x in universe.gethooks(HOOK_GET_VERSIONCONTROLSYSTEM_IMPLEMENTATIONS)]


class Vcs:

    def __init__(self, implclass, featurename, label):
        self.impl = implclass()
        self.implclassname = featurename + "." + implclass.__name__
        self.label = label

    def ask(self, ais):
        result = anise.framework.projects.IntermediateStructure.Let("versioncontrol.setvcs")
        result.setarg("vcs", anise.framework.projects.IntermediateStructure.Value(self.implclassname))
        return result


@anise.framework.features.hookhandler(projectdescriptioneditor.HOOK_GET_FEATURE_ACTIONS)
class ProjectDescriptionEditorAssistant(projectdescriptioneditor.ProjectDescriptionEditorAssistant):

    def __init__(self):
        projectdescriptioneditor.ProjectDescriptionEditorAssistant.__init__(
            self, "Manage version control systems",
            "Bind your project to a version control system for some additional services. When a version control "
            "system is bound to an Anise project, some additional functionality becomes available.", "versioncontrol")

    def getcustomactions(self, ais):
        return [projectdescriptioneditor.CustomAction("Add version control system", self._add_versioncontrolsystem,
                                                      icon="userfeedback/add.png")]

    def _add_versioncontrolsystem(self, ais):
        vcss = getversioncontrolimplementations()
        vcsimplnames = [x.implclassname for x in vcss]
        vcsdict = {x.implclassname: x for x in vcss}
        vcsimpl = universe.ui.userfeedback.choice(
            "Please choose a version control system.", [vcsdict[x].label for x in vcsimplnames])
        if vcsimpl is None:
            return
        selvcs = vcss[vcsimpl]
        newlet = selvcs.ask(ais)
        if not selvcs.impl.findvcs():
            if not selvcs.impl.bindvcs(newlet):
                anise.utils.logging.logerror("Unable to bind version control system.")
        for x in ais.entries():
            if isinstance(x, anise.framework.projects.IntermediateStructure.Let) and x.key() == "versioncontrol.setvcs":
                ais.removeentry(x)
        ais.addentry(newlet)

    def getentries(self, ais):
        r = []
        for e in ais.entries():
            if isinstance(e, anise.framework.projects.IntermediateStructure.Let) and e.key() == "versioncontrol.setvcs":
                r.append(e)
        return r

    def getentrylabel(self, ais, entry):
        vcsdict = {x.implclassname: x for x in getversioncontrolimplementations()}
        ientry = entry.getarg("vcs")
        vctype = vcsdict.get(ientry.vtype())
        if vctype:
            return vctype.label
        else:
            return ientry.vtype()
