# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for some stuff around Python.
"""

import os
import re
import subprocess

import anise.utils

from anise.framework.projects import universe


class internals:

    @staticmethod
    def django_quickstart(relpath):
        """!
        Starts the application in a Django development server for fast testing.
        It updates the database schema before, if required.
        @param relpath: Path to the Django application. Relative to the project root directory.
        """
        with universe.executionscope.newsubscope("Quickstarting the django project"):
            fdjango = universe.specialpaths.projectdir + "/" + relpath
            # detect python
            pycmd = "python3"
            try:
                subprocess.call([pycmd, "--version"])
            except OSError:
                pycmd = "python"
            with anise.utils.basic.ChDir(fdjango):
                for cmdline in [[pycmd, "./manage.py", "makemigrations"],
                                [pycmd, "./manage.py", "migrate"],
                                [pycmd, "./manage.py", "runserver", "--no-color"]]:
                    import pty  # we import it on demand since it's not supported everywhere
                    mypty = pty.openpty()[0]
                    try:
                        import fcntl
                        fcntl.fcntl(mypty, fcntl.F_SETFL, fcntl.fcntl(mypty, fcntl.F_GETFL) | os.O_NONBLOCK)
                    except ImportError:
                        pass
                    prc = subprocess.Popen(cmdline, stdout=mypty, stderr=mypty)
                    line = ""

                    def _procline(l):
                        sev = anise.utils.logging.Severity.Info
                        if l.startswith("Quit the server with"):
                            return
                        if l.startswith("Starting development server at"):
                            sev = anise.utils.logging.Severity.Warning
                            hsaddr = re.search("http://[^\\s]*", l)
                            if hsaddr:
                                haddr = hsaddr.group(0)
                                anise.utils.basic.openfile(haddr)
                        anise.utils.logging.logmessage(l, printcaller=False, severity=sev)
                    while True:
                        while True:
                            try:
                                chunk = os.read(mypty, 1).decode()
                            except IOError as e:
                                if e.errno == 11:
                                    break
                            if chunk in "\r\n":
                                if line:
                                    _procline(line)
                                line = ""
                            else:
                                line += chunk
                        try:
                            prc.wait(1)
                            break
                        except subprocess.TimeoutExpired:
                            pass
                    if line:
                        _procline(line)


def is_django_project(relpath=""):
    """!
    Configures a project to be a Django project.
    @param relpath: Path to the Django application. Relative to the project root directory.
    """
    t = lambda: internals.django_quickstart(relpath)
    t.__doc__ = internals.django_quickstart.__doc__
    universe.addtask(t, "python.django_quickstart")
