# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for signing binaries with cryptographic certificates.
"""

import os

import anise.framework.exceptions
import anise.framework.files
import anise.utils

from anise.framework.projects import universe


class internals:

    @staticmethod
    def win32_osslsigncode(
            *, binpath, pempath, outpath, projectname, projecthomepage, osslsigncodepath, password):
        """!
        Signs a win32 binary with osslsigncode.
        @param binpath: Path to the binary to sign.
        @param pempath: Path to the pem certificate file to use.
        @param outpath: Destination path for the signed binary.
        @param projectname: The project name.
        @param projecthomepage: The project homepage url.
        @param osslsigncodepath: Path to the osslsigncode tool.
        @param password: Certificate password.
        """
        if anise.utils.basic.call(osslsigncodepath + " -v", shell=True)[0] not in [0, 255]:
            raise anise.framework.exceptions.RequirementsMissingError("'osslsigncode' seems to be unavailable")
        callparms = [osslsigncodepath, "sign", "-spc", pempath, "-key", pempath, "-n", projectname, "-in", binpath,
                     "-out", outpath, "-pass", password]
        if projecthomepage is not None:
            callparms += ["-i", projecthomepage]
        r, o = anise.utils.basic.call(*callparms)
        if r != 0:
            raise anise.framework.exceptions.ProcessExecutionFailedError("unable to run osslsigncode", o)
        anise.utils.logging.logdebug(o)

    @staticmethod
    def p12cert2pem(*, opensslpath, p12path, pemoutpath, password):
        """!
        Converts a .p12 certificate to a .pem file.
        @param opensslpath: Path to the OpenSSL binary.
        @param p12path: Path to the p12 certificate file.
        @param pemoutpath: Destination path for the pem file.
        @param password: Certificate password of the p12 file.
        """
        if anise.utils.basic.call(opensslpath + " version", shell=True)[0] != 0:
            raise anise.framework.exceptions.RequirementsMissingError("'openssl' seems to be unavailable")
        r, o = anise.utils.basic.call(opensslpath, "pkcs12", "-in", p12path, "-out", pemoutpath, "-passin",
                                      "pass:"+password, "-passout", "pass:"+password)
        if r != 0:
            raise anise.framework.exceptions.ProcessExecutionFailedError("unable to run openssl", o)
        anise.utils.logging.logdebug(o)


class OsslsigncodeConfiguration:
    """!
    Signing configuration for signing a win32 binary with the 'osslsigncode' tool.
    """

    def __init__(self, p12path, intermediatepath=None, osslsigncodepath="osslsigncode", opensslpath="openssl",
                 password=""):
        """!
        @param p12path: Path to the .p12 file.
        @param intermediatepath: Path to intermediate certificates.
        @param osslsigncodepath: Path to the osslsigncode tool.
        @param opensslpath: Path to the openssl tool.
        @param password: The certificate passphrase.
        """
        self.osslsigncodepath = osslsigncodepath
        self.opensslpath = opensslpath
        self.p12path = p12path
        self.intermediatepath = intermediatepath
        self.password = password
        

def signwin32_osslsigncode(signconfig, source, required=True):
    """!
    Signs a win32 binary with a cryptographic code-signing certificate.
    @param signconfig: An instance of OsslsigncodeConfiguration specifying the signing details.
    @param source: An anise.framework.files.Filestructure.
    @param required: If a failure is critical or can be ignored.
    """
    with universe.executionscope.newsubscope("Creating a signed win32 binary"):
        targetdir = anise.framework.files.TempDir()
        outpath = targetdir.path() + "/" + os.path.basename(source.path())
        projectname = universe.name
        pempath = targetdir.path() + "/crt.der"
        try:
            internals.p12cert2pem(
                opensslpath=signconfig.opensslpath, p12path=signconfig.p12path, pemoutpath=pempath,
                password=signconfig.password)
            if signconfig.intermediatepath:
                with open(signconfig.intermediatepath, "r") as f:
                    intermediatepem = f.read()
                with open(pempath, "a") as f:
                    f.write("\n" + intermediatepem)
            internals.win32_osslsigncode(
                binpath=source.path(), pempath=pempath, outpath=outpath, projectname=projectname,
                projecthomepage=universe.base.websiteurl, osslsigncodepath=signconfig.osslsigncodepath,
                password=signconfig.password)
            return anise.framework.files.Filestructure(outpath)
        except Exception as e:
            if required:
                raise
            else:
                anise.utils.logging.logwarning("Unable to sign win32 binary (leaving it unsigned): " + str(e))
                return source
