# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for user interfaces and interaction with the user.
"""

import os
import shlex
import sys

import anise.framework.exceptions
import anise.framework.features
import anise.utils

from anise.framework.projects import universe


HOOK_UIMODE_SELECTED = anise.framework.features.Hook(
    doc="This hook is called as soon as the user interface mode is determined.",
    wizard_functionadditionallog=["universe.ui.mode"])

HOOK_GET_USERFEEDBACK_IMPLEMENTATION = anise.framework.features.Hook(
    doc="This hook is called for collecting userfeedback implementations.", hide=True)

HOOK_GET_ALLOWIMPLICITCLOSE_IMPLEMENTATION = anise.framework.features.Hook(
    doc="This hook is called for collecting implementations used for specifying if a user may close the user interface"
        "without explicit confirmation.", hide=True)


class internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def initproject1():
        universe.addhook(HOOK_GET_ALLOWIMPLICITCLOSE_IMPLEMENTATION, internals.SetImplicitStopAllowed, provides=["ui"])

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def initproject2():
        if not hasattr(universe, "DEFAULT"):
            universe.addtask(internals._default if universe.specialpaths.projectfile else internals._default_baddir,
                             "DEFAULT", invisible=True)
        mode = getattr(universe.ui, "force_mode", None)
        if not mode:
            if sys.platform == "win32":
                mode = Mode.WEB
            else:
                if os.getenv("DISPLAY"):
                    mode = Mode.WEB
                else:
                    mode = Mode.TERMINAL
        universe.ui.mode = mode
        for hook in universe.gethooks(HOOK_UIMODE_SELECTED):
            hook()
        universe.ui.userfeedback = internals.UserFeedbackProxy()
        universe.ui.internals.taskchooser_welcomemsg = "The following tasks are available for execution."
        universe.ui.internals.taskchooser_emptymsg = "There are currently no tasks defined in this Anise project."
        universe.ui.internals.taskchooser_showallidx = 100.0

    @staticmethod
    def _default():
        """!
        Opens the graphical task chooser and executes the selected task afterwards.
        """
        showall = getattr(universe.ui, "uichooser_showall", False)
        showshowall = (not showall) and (not getattr(universe.ui, "uichooser_hide_showall", False))
        tasks = []
        for task in universe.registeredtasks():
            if (not task.invisible) and (showall or task.important):
                defaultlabel = "{} (in {})".format(task.taskname, task.parentname) if task.parentname else task.taskname
                tasks.append([(task.sortindex, task.parentname or ""), task.name,
                              task.label if task.label else defaultlabel])
        origtasksempty = len(list(tasks)) == 0
        if showshowall:
            tasks.append([(universe.ui.internals.taskchooser_showallidx, ""), None, "show all"])
        if origtasksempty and len(tasks) > 0:
            universe.ui.uichooser_showall = True
            return internals._default()
        elif len(tasks) > 0:
            tasks.sort()
            with universe.gethooks(HOOK_GET_ALLOWIMPLICITCLOSE_IMPLEMENTATION)[0](True):
                itask = universe.ui.userfeedback.choice(universe.ui.internals.taskchooser_welcomemsg,
                                                        [x[2] for x in tasks])
            if itask is not None:
                task = tasks[itask][1]
                if task is None:
                    universe.ui.uichooser_showall = True
                    return internals._default()
                else:
                    universe.taskname = task
                    with universe.executionscope.newsubscope("Executing '{}'".format(task)):
                        return eval(task, universe().__dict__)()
        else:
            universe.ui.userfeedback.messageok(universe.ui.internals.taskchooser_emptymsg)

    @staticmethod
    def _default_baddir():
        """!
        Opens the graphical guide for starts in non-Anise directories.
        """
        while True:
            choices = []
            choiceimages = []
            _ichoice2isubdirs = {}
            _ichoicestarthere = None
            _ichoicecreatefresh = None
            _ichoicecreatedir = None
            _ichoicechangedir = None
            cwd = os.getcwd()
            if os.path.exists(cwd+"/_meta/_projectdesc") \
                    or (os.path.exists(cwd+"/_projectdesc") and (os.path.basename(cwd) != "_meta")):
                _ichoicestarthere = len(choices)
                choices.append("Open this project")
                choiceimages.append("userfeedback/system.png")
            else:
                _ichoicecreatefresh = len(choices)
                choices.append("Create a fresh project here")
                choiceimages.append("userfeedback/magic.png")
                _ichoicecreatedir = len(choices)
                choices.append("Create a new directory here")
                choiceimages.append("userfeedback/folder.png")
            subdirs = sorted([x for x in os.listdir(cwd) if os.path.isdir(cwd+"/"+x)])
            if os.path.abspath(cwd) != os.path.abspath(cwd + "/.."):
                subdirs = [".."] + subdirs
            elif sys.platform == "win32":
                _ichoicechangedir = len(choices)
                choices.append("Go to another drive")
                choiceimages.append("userfeedback/mount.png")
            for i, s in enumerate(subdirs):
                _ichoice2isubdirs[len(choices)] = i
                choices.append(s)
                choiceimages.append("userfeedback/folder.png")
            r = universe.ui.userfeedback.choice("""
Welcome to Anise!

Please choose the root directory of an Anise project or create a fresh one.

Current directory: {cwd}
""".format(**locals()), choices=choices, choiceimages=choiceimages)
            anisecmd = shlex.quote(sys.argv[0])
            try:
                webexecparentcmdopts = ["--s:ui.web.webexec.parentid=" + universe.ui.web.webexec.instance.id,
                                        "--b:ui.web.webexec.returntoparent=0"]
            except AttributeError:
                webexecparentcmdopts = []
            if r is None:
                break
            elif r in _ichoice2isubdirs:
                sd = subdirs[_ichoice2isubdirs[r]]
                os.chdir(sd)
            elif r == _ichoicechangedir:
                drives = anise.utils.basic.getdriveletters()
                r = universe.ui.userfeedback.choice("Please choose the drive to go to.",
                                                    choices=[x.upper()+":\\" for x in drives])
                if r is not None:
                    os.chdir(drives[r]+":/")
            elif r == _ichoicestarthere:
                anise.utils.basic.callanise(*webexecparentcmdopts)
                break
            elif r == _ichoicecreatefresh:
                anise.utils.basic.callanise("--createfresh")
                anise.utils.basic.callanise(*(["console", "--b:firstrun=1"] + webexecparentcmdopts))
                break
            elif r == _ichoicecreatedir:
                r = universe.ui.userfeedback.input("Please choose the new directory name.", "new_directory")
                if r:
                    os.makedirs(r, exist_ok=True)

    class SetImplicitStopAllowed:
        """!
        Controls if implicit application stop is enabled (e.g. not asking user for confirmation).
        @a_infrastructure
        """

        def __init__(self, val):
            pass

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

    class UserFeedback:
        """!
        This interface provides mechanisms for asking the user for some kinds of interaction.
        """

        def __init__(self):
            pass

        def message(self, msg, btns, image):
            """!
            Shows a message and asks the user to trigger one of the buttons.
            @param msg: Message text.
            @param btns: Buttons to show (list of strings).
            @param image: Icon name.
            @return: The index of the selected button.
            """
            raise anise.framework.exceptions.NotImplementedError()

        def input(self, msg, default, image):
            """!
            Asks the user to enter text (single line).
            @param msg: The message to show.
            @param default: The default text.
            @param image: Icon name.
            @return: The entered text, or None if user cancelled.
            """
            raise anise.framework.exceptions.NotImplementedError()

        def multilineinput(self, msg, default, image):
            """!
            Asks the user to enter text (multi line).
            @param msg: The message to show.
            @param default: The default text.
            @param image: Icon name.
            @return: The entered text, or None if user cancelled.
            """
            raise anise.framework.exceptions.NotImplementedError()

        def choice(self, msg, choices, canceltext, image, choiceimages):
            """!
            Asks the user to choose from a list of alternatives.
            @param msg: The message to show.
            @param choices: List of choices.
            @param canceltext: Label of the Cancel button.
            @param image: Icon name.
            @param choiceimages: Icon names for the choices.
            @return: The index of the choice the user made, or None if user cancelled.
            """
            raise anise.framework.exceptions.NotImplementedError()

        def treechoice(self, msg, choicetree, canceltext, actions, image):
            """!
            Asks the user to choose from a tree of alternatives.
            @param msg: The message to show.
            @param choicetree: The tree of available choices. Instance of ChoiceTree.
            @param canceltext: Label of the Cancel button.
            @param actions: Additional actions.
            @param image: Icon name.
            @return: The chosen tree node (ChoiceTree.Node), or None if user cancelled.
            """
            raise anise.framework.exceptions.NotImplementedError()

        def messageyesno(self, msg, image=None):
            """!
            Shows a message and asks the user to trigger either Yes or No.
            @param msg: The message to show.
            @param image: Icon name.
            @return: True, if the user selected Yes.
            """
            return self.message(msg, ["Yes", "No"], image=image) == 0

        def messageok(self, msg, image=None):
            """!
            Shows a message and asks the user to confirm it.
            @param msg: The message to show.
            @param image: Icon name.
            """
            self.message(msg, ["OK"], image=image)

    class UserFeedbackProxy(UserFeedback):
        """!
        This user feedback implementation forwards requests to another
        internal UserFeedback, depending on which mode (web, terminal)
        is selected.
        """

        def __init__(self):
            internals.UserFeedback.__init__(self)

        def _getimpl(self):
            result = None
            for h in universe.gethooks(HOOK_GET_USERFEEDBACK_IMPLEMENTATION):
                result = h()
                if result:
                    break
            return result

        def message(self, msg, btns, image=None):
            return self._getimpl().message(msg, btns, image)

        def input(self, msg, default="", image=None):
            return self._getimpl().input(msg, default, image)

        def multilineinput(self, msg, default="", image=None):
            return self._getimpl().multilineinput(msg, default, image)

        def choice(self, msg, choices, canceltext="Cancel", image=None, choiceimages=None):
            choiceimages = choiceimages or []
            if len(choices) == 0:
                return None
            while len(choiceimages) < len(choices):
                choiceimages.append(None)
            return self._getimpl().choice(msg, choices, canceltext, image, choiceimages)

        def treechoice(self, msg, choicetree, canceltext="Cancel", actions=None, image=None):
            actions = actions or []
            return self._getimpl().treechoice(msg, choicetree, canceltext, actions, image)


class Mode:
    """!
    Enumeration of user interface modes.
    """

    TERMINAL = "TERMINAL"
    WEB = "WEB"


class ChoiceTree:
    """!
    A tree of possible choices. Typically used in UserFeedback.treechoice.

    For adding nodes to the tree, use the ChoiceTree.Node.createchild method, which is available on the root node
    of this tree as well as on all deeper nodes.
    """

    class Node:
        """!
        One node in a tree of possible choices.
        """

        def __init__(self, key, val, parent, childs, image):
            self.key = key
            if isinstance(key, str):
                self.key = self.key.replace("\r", "").replace("\n", "\\n")
            self.val = val
            self.parent = parent
            self.childs = childs
            self.image = image

        def createchild(self, key, val, image=None):
            """!
            Creates a new node in tree as as child of this node.
            @param key: The key string. This is displayed in tree.
            @param val: The value description string. It is displayed in the tree as well.
            @param image: Icon name.
            @return: The new ChoiceTree.Node.
            """
            r = ChoiceTree.Node(key, val, self, [], image)
            self.childs.append(r)
            return r

    def __init__(self):
        self.root = ChoiceTree.Node(None, None, None, [], None)


def chooser_hide_showall(hide=True):
    """!
    Sets the availability of the 'show all' button in the task chooser, which also shows the unimportant tasks.
    @param hide: If the 'show all' button is to be hidden.
    """
    universe.ui.uichooser_hide_showall = hide


def enforce_terminal_mode():
    """!
    Force usage of terminal mode.
    """
    universe.ui.force_mode = Mode.TERMINAL
