# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for web-based user interfaces.
"""

import time

import anise.framework.features

from anise.framework.projects import universe

ui = anise.framework.features.loadfeature("ui").featuremodule


HOOK_SEND_USERFEEDBACK_WEB_REQUEST = anise.framework.features.Hook(
    hide=True, doc="This hook is called by the web infrastructure for requesting userfeedback.")

HOOK_GET_USERFEEDBACK_WEB_ANSWERS = anise.framework.features.Hook(
    hide=True, doc="This hook is called by the web infrastructure for getting userfeedback answers.")


class internals:

    class UserFeedback(ui.internals.UserFeedback):
        """!
        Uses available web browser sessions for user interaction.
        """

        class FeedbackRequest:

            def __init__(self, index, request, data, requestor):
                self.index = index
                self.request = request
                self.data = data
                self.requestor = requestor

            def isactive(self):
                return self.requestor.currentrequestindex == self.index

            def _to_simple_repr_(self):
                return 'anise.FeedbackRequest', {k: v for k, v in self.__dict__.items() if k != "requestor"}

        def __init__(self):
            ui.internals.UserFeedback.__init__(self)
            self.nextrequestindex = 1
            self.currentrequestindex = None

        def _webrequest(self, request, requestdata):
            requestdata = {k: requestdata[k] for k in requestdata if (k != "self" and (not k.startswith("_")))}
            requestindex = self.currentrequestindex = self.nextrequestindex
            self.nextrequestindex += 1
            for h in universe.gethooks(HOOK_SEND_USERFEEDBACK_WEB_REQUEST):
                h(internals.UserFeedback.FeedbackRequest(requestindex, request, requestdata, self))
            result = None
            while result is None:
                for h in universe.gethooks(HOOK_GET_USERFEEDBACK_WEB_ANSWERS):
                    result = h(requestindex)
                    if result is not None:
                        break
                if result is None:
                    time.sleep(0.1)
            self.currentrequestindex = None
            return result[0]

        def message(self, msg, btns, image):
            res = self._webrequest("message", locals())
            res = int(res)
            return res

        def input(self, msg, default, image):
            res = self._webrequest("input", locals())
            return res

        def multilineinput(self, msg, default, image):
            res = self._webrequest("multilineinput", locals())
            return res

        def treechoice(self, msg, choicetree, canceltext, actions, image):
            _i = 0
            _idx2choice = {}

            def _tojson(tc):
                nonlocal _i
                myi = _i
                _i += 1
                _idx2choice[myi] = tc
                return {'key': str(tc.key), 'val': myi, 'icon': tc.image, 'childs': [_tojson(x) for x in tc.childs]}
            choicetree = _tojson(choicetree.root)
            _tojson = None
            res = self._webrequest("treechoice", locals())
            if res is None:
                rres = None
            else:
                res = int(res)
                if res > 0:
                    rres = _idx2choice[res]
                else:
                    rres = actions[-res]
            return rres

        def choice(self, msg, choices, canceltext, image, choiceimages):
            choices = [str(x) for x in choices]
            res = self._webrequest("choice", locals())
            res = int(res)
            if res == -1:
                res = None
            return res

        def isavailable(self):
            return len(universe.gethooks(HOOK_SEND_USERFEEDBACK_WEB_REQUEST)) > 0

    @staticmethod
    @anise.framework.features.hookhandler(ui.HOOK_UIMODE_SELECTED)
    def uimodeselected():
        userfeedback = internals.UserFeedback()
        universe.addhook(
            ui.HOOK_GET_USERFEEDBACK_IMPLEMENTATION, lambda: userfeedback if userfeedback.isavailable() else None,
            provides=[ui.Mode.WEB], prepares=[ui.Mode.TERMINAL])
