# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for helpers related to web-based user interfaces.
"""

import json
import time

import anise.data.Yaji.yaji as yaji
import anise.framework.engine
import anise.framework.features
import anise.framework.globalvars
import anise.framework.report
import anise.utils

from anise.framework.projects import universe

uiweb = anise.framework.features.loadfeature("ui.web").featuremodule


class AniseWebApplication(yaji.Application):

    def __init__(self, **b):
        super().__init__(icon="icons/pde/anise.png", **b)
        self.startmessageonlyterminal = False
        self._wh_hooks = []
        self.userfeedbackanswers = {}
        self.try_addclientresources_from_default_location(__file__)
        self.add_staticfile_location(universe.specialpaths.anisedatadir)

    def _execution_scopepart_added(self, x):
        self.triggerevent("anise_ExecPager_newcontentavailable")

    def _send_userfeedback_web_request(self, request):
        if request:
            adata = dict(data=request.data, idx=request.index)
            if request.request == 'message':
                v = yaji.View('anise_internalUI.UserFeedbackWidget_message', **adata)
            elif request.request == 'input':
                v = yaji.View('anise_internalUI.UserFeedbackWidget_input', **adata)
            elif request.request == 'multilineinput':
                v = yaji.View('anise_internalUI.UserFeedbackWidget_multilineinput', **adata)
            elif request.request == 'choice':
                v = yaji.View('anise_internalUI.UserFeedbackWidget_choice', **adata)
            elif request.request == 'treechoice':
                v = yaji.View('anise_internalUI.UserFeedbackWidget_treechoice', **adata)
            else:
                raise ValueError(f"Bad userfeedback request: {request.request}")
            self.setbodyleft(yaji.VerticalStack(
                yaji.VerticalStack(v, styleClass='clovestyle_vmargin_medium', vstretch=1, hstretch=1),
            hstretch=1, styleClass='anise_webexec_userfeedback_host') if v else None)
            self.triggerevent('scrollDownPager')
            self.appconfig.setconfig('mainviewStyleClass', 'anise_userfeedbackenabled')
        else:
            self.setbodyleft(None)
            self.appconfig.setconfig('mainviewStyleClass', '')

    def _get_userfeedback_web_answers(self, requestindex):
        result = self.userfeedbackanswers.get(requestindex)
        if result is not None:
            self.userfeedbackanswers.pop(requestindex)
        return result

    def _shutdown(self):
        time.sleep(3)
        self.stop()

    def oninitialize(self):
        self.appconfig.setconfig("aniseProjectName", universe.name)

    def start(self):
        self._wh_hooks = [universe.addhook(anise.framework.report.HOOK_EXECUTION_SCOPEPART_ADDED,
                                           self._execution_scopepart_added),
                          universe.addhook(uiweb.HOOK_SEND_USERFEEDBACK_WEB_REQUEST,
                                           self._send_userfeedback_web_request),
                          universe.addhook(uiweb.HOOK_GET_USERFEEDBACK_WEB_ANSWERS,
                                           self._get_userfeedback_web_answers),
                          universe.addhook(anise.framework.engine.HOOK_SHUTDOWN, self._shutdown, provides=["ui.web"])]
        super().start()

    def stop(self):
        super().stop()
        for h in self._wh_hooks:
            universe.removehook(h)
        self._wh_hooks = []

    def onprocessrequesterror(self, request, error):
        anise.utils.logging.logexception(error, prolog=f"Error in Anise web application ({request}): ")

    def onopenbrowsererror(self):
        anise.utils.logging.logerror("There were problems while starting a web browser for controlling Anise\n"
                                     "execution.\n\n"
                                     "Please try to open it manually or start Anise again\n"
                                     "with the '--no-gui' parameter added or read the manual.", printcaller=False)

    def onbrowserreopened(self):
        anise.utils.logging.logwarning("Your Anise browser session was closed prematurely.\nA new one is starting.",
                                       printcaller=False)

    def onopenbrowserinformationoutput(self, kind, message):
        anise.utils.logging.loginfo(message, printcaller=False, terminalonly=True)

    @yaji.RequestHandler.for_urls(auto=True)
    def _do_getprogresses(self, since: int):
        items = []
        progresses = []
        for sp in anise.framework.report._rootscope.getscopepartsbyidrange():
            if isinstance(sp, anise.framework.report.ExecutionScope) \
                    and not isinstance(sp.progress(), anise.framework.report.ExecutionScope.ProgressFinished):
                progresses.append((sp.address(), sp.progress().tojson()))
            if sp.id() > since:
                if isinstance(sp, anise.framework.report.ExecutionScope):
                    item = {"title": sp.title()}
                else:
                    item = {"value": sp.value(), "hints": sp.hints()}
                item.update({"id": sp.id(), "address": sp.address()})
                items.append(item)
        return {'items': items, 'progresses': progresses,
                'currentscope': anise.framework.report.topscope().address()}

    @yaji.RequestHandler.for_urls(auto=True)
    def _do_answeruserfeedback(self, index: int, answer):
        self.userfeedbackanswers[index] = json.loads(answer)
        self._send_userfeedback_web_request(None)
