# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Features for interaction with version control systems.
"""

import datetime
import os
# noinspection PyPep8Naming
import xml.etree.cElementTree as ET

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

ui = anise.framework.features.loadfeature("ui").featuremodule


class CommitMessage:
    """!
    A commit message.
    """

    def __init__(self, message, revision, committime=None, username=None, **kwargs):
        """!
        @param message: The message text.
        @param revision: The revision.
        @param committime: The time of committing.
        @param username: The user who committed.
        @param kwargs: Additional commit data.
        """
        self.message = message
        self.revision = revision
        self.committime = committime
        self.username = username
        self.customdata = kwargs


class ChangeLogRevision:
    """!
    A revision section in a change log.
    """

    def __init__(self, versionlabel, commitmessage):
        """!
        @param versionlabel: The version label.
        @param commitmessage: The versioncontrol.CommitMessage which this version label is defined by.
        """
        self.versionlabel = versionlabel
        self.commitmessage = commitmessage
        self.changes = []


class ChangeLogChange:
    """!
    A change entry in a change log.
    """

    def __init__(self, changetext, commitmessage):
        """!
        @param changetext: The change text.
        @param commitmessage: The versioncontrol.CommitMessage which this version label is defined by.
        """
        self.changetext = changetext
        self.commitmessage = commitmessage


def setvcs(vcs):
    """!
    Sets the version control system for this project.
    @param vcs: Instance of a internals.VersionControlSystem subclass, which implements the interaction with the version
                control system.
    """
    universe.versioncontrol.current = vcs


class internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _prepare_fallbacks():
        if not hasattr(universe.versioncontrol, "current"):
            universe.versioncontrol.current = internals.PseudoVersionControlSystem()

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _prepare_vcstasks():
        if not isinstance(universe.versioncontrol.current, internals.PseudoVersionControlSystem):
            universe.addtask(universe.versioncontrol.current.syncversioncontrolsystem, "versioncontrol.sync")
            universe.addtask(universe.versioncontrol.current.getchangelog, "versioncontrol.getchangelog",
                             important=False)
            universe.addtask(universe.versioncontrol.current.addchange, "versioncontrol.addchangelogentry",
                             important=False)
            universe.addtask(universe.versioncontrol.current.addlabel, "versioncontrol.addversionlabel",
                             important=False)
            universe.addtask(universe.versioncontrol.current.branchfromhere, "versioncontrol.branchfromhere",
                             important=False)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _add_homepage_section():
        def my_homepage_section(homepage, sourcedir, backgroundimage, *a):
            if hasattr(universe.versioncontrol, "getchangelog"):
                changelog = universe.versioncontrol.getchangelog(asstring=False)
                cltxt = ""
                for rev in changelog:
                    if len(rev.changes) > 0:
                        if (not cltxt) or (datetime.datetime.now() - rev.commitmessage.committime
                                           < datetime.timedelta(days=90)):
                            cltxt += f"\n\n**{rev.commitmessage.committime.date()} ({rev.versionlabel})**\n\n"
                            for change in rev.changes:
                                cltxt += f"- {change.changetext}\n"
                if cltxt:
                    return f"The last changes brought the following improvements.\n\n{cltxt}"
        universe.homepage.sections.add(7000, "changelog", "Change Log", my_homepage_section)

    class VersionControlSystem:
        """!
        A version control system.

        Override this for providing an own version control system.
        """

        def __init__(self):
            self.label = self.__class__.__name__

        def getfullversion(self):
            """!
            Returns a complete version string.
            @a_dooverride
            """
            return universe.base.version

        def createrawpackage(self):
            """!
            Generates a raw package. This package is the basis for most (or all) of the other package types.
            @a_dooverride
            """
            with universe.executionscope.newsubscope("Creating the raw source package"):
                return anise.framework.files.Filestructure(
                    universe.specialpaths.projectdir).with_modified_rootname("rawpackage")

        def syncversioncontrolsystem(self, commitmessage="", forceskipped=True, forceunchanged=False):
            """!
            Synchronizes the local working copy with the version control system.
            @a_dooverride
            @param commitmessage: Commit message text.
            @param forceskipped: Force synchronizing even if the user decided to skip that.
            @param forceunchanged: Force synchronizing even if there are no changes to commit.
            """
            raise anise.framework.exceptions.NotImplementedError()

        def findvcs(self):
            """!
            Checks if the current project is bound to a version control system of this kind.
            @a_dooverride_optional
            """
            return False

        def bindvcs(self, l):
            """!
            Binds the current project to this version control system. Uses user interaction for getting details
            like urls.
            @a_dooverride_optional
            @param l: The let configuration node for storing additional parameters.
            """
            raise anise.framework.exceptions.NotImplementedError()

        def fetchrawcommitmessage(self, revision):
            """!
            Fetches a raw commit message.
            Do not use it directly, but getcommitmessage instead.
            @a_dooverride_optional
            @param revision: The revision.
            @return: The raw commit message as string in a custom format.
            """
            raise anise.framework.exceptions.NotImplementedError()

        def getcommitmessagefromrawcommitmessage(self, s):
            """!
            Gets the actual commit message from the raw 'fetchrawcommitmessage' commit message output.
            Do not use it directly, but getcommitmessage instead.
            @a_dooverride_optional
            @param s: The raw commit message string.
            @return: An instance of CommitMessage.
            """
            raise anise.framework.exceptions.NotImplementedError()

        def getpreviousrevision(self, revision):
            """!
            Gets the predecessor revision for a given revision.
            @a_dooverride_optional
            @param revision: A revision.
            @return: The predecessing revision.
            """
            irev = int(revision)
            if irev <= 1:
                return None
            return irev - 1

        def getheadrevision(self):
            """!
            Gets the head revision.
            @a_dooverride_optional
            """
            raise anise.framework.exceptions.NotImplementedError()

        def getcommitmessage(self, revision):
            """!
            Returns the raw commit message (as formatted by the vcs tool) for a revision.
            @param revision: The revision.
            """
            if not hasattr(universe, "commitmessagecachepath"):
                universe.commitmessagecachepath = os.path.expanduser("~/.anise/commitmessagecache")
            commitmessagecachepath = universe.commitmessagecachepath + "/" + universe.name + "/" + \
                str(int(revision / 1000)) + "nnn"
            if os.path.exists(commitmessagecachepath):
                xdoc = ET.ElementTree(ET.fromstring(anise.utils.basic.readfromfile(commitmessagecachepath)))
                xroot = xdoc.getroot()
                for xentry in xroot.getchildren():
                    if xentry.attrib["rev"] == str(revision):
                        return self.getcommitmessagefromrawcommitmessage(xentry.text)  # answer from cache
            res = anise.utils.basic.execwithretry(self.fetchrawcommitmessage, revision, delaysecs=1)
            commitmessagecachepathdir = os.path.dirname(commitmessagecachepath)
            if not os.path.isdir(commitmessagecachepathdir):
                os.makedirs(commitmessagecachepathdir)
            if os.path.exists(commitmessagecachepath):
                xdoc = ET.ElementTree(ET.fromstring(anise.utils.basic.readfromfile(commitmessagecachepath)))
                xroot = xdoc.getroot()
            else:
                xroot = ET.Element("commitmessagecache")
                xdoc = ET.ElementTree(xroot)
            xnewrev = ET.Element("msg")
            xnewrev.attrib["rev"] = str(revision)
            xnewrev.text = res
            xroot.append(xnewrev)
            with open(commitmessagecachepath, "wb") as fxml:
                xdoc.write(fxml)
            return self.getcommitmessagefromrawcommitmessage(res)

        def getchangelog(self, asstring=True):
            """!
            Computes the project change log from information in specially formatted vcs commit messages.
            @param asstring: If to output the changelog as string (instead of an internal data structure).
            """
            result = []
            _currentrev = None
            rev = self.getheadrevision()
            while rev is not None:
                rcommitmsg = self.getcommitmessage(rev)
                msglines = rcommitmsg.message.split("\n")
                for msgline in msglines:
                    if msgline.startswith("##LABEL:"):
                        _currentrev = ChangeLogRevision(msgline[len("##LABEL:"):].strip(), rcommitmsg)
                        result.append(_currentrev)
                    elif msgline.startswith("##CHANGE:"):
                        chg = msgline[len("##CHANGE:"):]
                        if _currentrev:
                            _currentrev.changes.append(ChangeLogChange(chg.strip(), rcommitmsg))
                rev = self.getpreviousrevision(rev)
            if asstring:
                result = [x for x in result if len(x.changes) > 0]
                sresult = "# {name} Changelog\n\n".format(name=universe.name)
                if len(result) == 0:
                    sresult += "No changelog available.\n"
                else:
                    for clrev in result:
                        sresult += "## {label}\n\n".format(label=clrev.versionlabel)
                        for clchange in clrev.changes:
                            change = clchange.changetext
                            first = True
                            while len(change) > 0:
                                if first:
                                    sresult += "- "
                                else:
                                    sresult += "  "
                                first = False
                                if len(change.rstrip()) > 78:
                                    i = change[:79].rfind(" ")
                                    if i == -1:
                                        i = 80
                                    j = change[:].find("\n")
                                    if j > -1:
                                        i = j
                                    sresult += change[:i] + "\n"
                                    change = change[i+1:]
                                else:
                                    sresult += change.rstrip() + "\n"
                                    change = ""
                        sresult += "\n"
                return sresult
            else:
                return result

        def addchange(self, changetext=None):
            """!
            Adds a change to the changelog.
            @param changetext: The changelog entry text.
            """
            if changetext is None:
                changetext = universe.ui.userfeedback.input("Please enter the change description.", "")
            if changetext:
                changetext = changetext.replace("\n", " ")
                self.storechange(changetext)

        def addlabel(self, label=None):
            """!
            Adds a version label for the current state.
            @param label: The label string.
            """
            if label is None:
                universe.versioncontrol.current.syncversioncontrolsystem(forceskipped=True)
                label = universe.ui.userfeedback.input(
                    "Please enter the label name.", universe.packages.getversion())
            if label:
                label = label.replace("\n", " ")
                self.storelabel(label)

        def storechange(self, changetext):
            """!
            Stores a new entry to the changelog.

            Do not use this directly, but addchange instead.
            @param changetext: The changelog entry text.
            """
            self.syncversioncontrolsystem(commitmessage="##CHANGE: " + changetext, forceskipped=True,
                                          forceunchanged=True)

        def storelabel(self, label):
            """!
            Stores a new version label.

            Do not use this directly, but addlabel instead.
            @param label: The label string.
            """
            self.syncversioncontrolsystem(commitmessage="##LABEL: " + label, forceskipped=True, forceunchanged=True)

        def branchfromhere(self, branchname=None):
            """!
            Creates a new branch from here and switches to it. Unpushed changes will become part of that new branch.
            @a_dooverride
            @param branchname: The branch name (None: input dialog).
            """
            raise anise.framework.exceptions.NotImplementedError()

    class PseudoVersionControlSystem(VersionControlSystem):
        """!
        A pseudo version control system implementation which does nothing. This is the default as long as no real
        version control system is configured.
        """

        def __init__(self):
            internals.VersionControlSystem.__init__(self)

        def syncversioncontrolsystem(self, commitmessage="", forceskipped=True, forceunchanged=False):
            pass
