# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

## @package anise.framework.engine
##The anise engine initializes all the anise plattform stuff and executes a task from a project file.

import anise.framework.report
anise.framework.report.install_proxies()
import importlib
import os
import sys
import traceback
if sys.platform != "win32":
    import pwd

import anise.framework.features
import anise.framework.files
import anise.framework.projects
import anise.framework.globalvars
import anise.utils


HOOK_AFTER_PREPARATION = anise.framework.features.Hook(
    doc="This hook is called after the project description file is applied to the "
        "universe object. It is used for user defined preparation work. The "
        "features do some more in framework.features.HOOK_BEFORE_EXECUTION "
        "afterwards.", interesting=-50)

HOOK_TASK_RESULT_AVAILABLE = anise.framework.features.Hook(
    doc="This hook is called when the task result is available, just before "
        "the execution ends.")

HOOK_SHUTDOWN = anise.framework.features.Hook(
    doc="This hook is called in the end of the execution.")


class SpecialPaths:
    """!
    A data structure of some particular special paths.
    """

    def __init__(self, projectfile, projectdir):
        if projectfile:
            self.projectfile = os.path.abspath(projectfile)
            self.projectdir = os.path.abspath(projectdir)
            if self.projectfile.replace("\\", "/").startswith(self.projectdir.replace("\\", "/")+"/"):
                self.relativeprojectfile = self.projectfile[len(self.projectdir)+2:]
            else:
                self.relativeprojectfile = self.projectfile
        else:
            self.projectfile = None
            self.projectdir = anise.framework.files.TempDir().path()
        self.anisedir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        self.anisedatadir = self.anisedir + "/data"
        self.startdir = os.getcwd()


def execute(taskname, projectfile, projectdir, values, loadfeaturesfrom):
    """!
    Executes a task.
    @param taskname: The name of the task to be executed.
    @param projectfile: The path to the anise project description file. Their name is usually `_projectdesc`.
    @param projectdir: The path to the project root directory.
    @param values: Additional values to be set to the universe object.
    @param loadfeaturesfrom: List of paths where anise features shall be loaded from.
    """
    res = None
    # determine some constants like the project dir

    class _MyLocals:
        pass
    mylocals = _MyLocals()
    mylocals.specialpaths = SpecialPaths(projectfile, projectdir)
    # prepare the universe object
    _project = None
    try:
        with anise.framework.report._rootscope.newsubscope(
                "Executing '{}'".format(taskname) if taskname != "DEFAULT" else 'Executing') as executionscope:
            _project = anise.framework.projects.Universe({}, executionscope=executionscope)
            anise.framework.projects.setcurrentuniverse(_project)
            g = dict(importlib.import_module("anise.framework.imports").__dict__)
            _project.append(mylocals.__dict__)
            # load the features and put their stuff into the universe object
            anise.framework.features.internals.resetfeatures()
            for featurepath in loadfeaturesfrom:
                if featurepath == "builtin":
                    featurepath = os.path.dirname(os.path.dirname(__file__)) + "/features"
                anise.framework.features.internals.loadfeatures(featurepath)
            anise.framework.features.internals.injectdefaultstoproject(_project)
            anise.framework.features.internals.injecthookstoproject(_project)
            _project.append(values)
            # execute the HOOK_BEFORE_DEFINITION hookhandlers (for (early) preparation stuff)
            for hook in _project.gethooks(anise.framework.features.HOOK_BEFORE_DEFINITION):
                hook()
            # execute the project description file
            l = _project.__dict__
            if mylocals.specialpaths.projectfile:
                iprojectdesc = anise.framework.projects.IntermediateStructure.fromfile(
                    mylocals.specialpaths.projectfile)
                iprojectdesc.injecttoprojectobject(_project, g, l)
            # load the local anise config
            _project.append(l)
            # execute the HOOK_AFTER_PREPARATION hookhandlers (for own late-preparation stuff)
            for hook in _project.gethooks(HOOK_AFTER_PREPARATION):
                hook()
            # execute the HOOK_BEFORE_EXECUTION hookhandlers (for features late-preparation stuff)
            for hook in _project.gethooks(anise.framework.features.HOOK_BEFORE_EXECUTION):
                hook()
            # apply the 'values' again
            _project.append(values)
            # execute the task
            _project.taskname = taskname
            taskfct = eval(taskname, _project.__dict__)
            gl = g
            gl.update(l)
            res = eval("taskfct()", gl, {'taskfct': taskfct})
            try:
                for hook in _project.gethooks(HOOK_TASK_RESULT_AVAILABLE):
                    hook(res)
            except Exception as e:
                anise.utils.logging.logexception(e)
    except BaseException as e:
        e_full = traceback.format_exc()
        anise.framework.report.topscope().appendchild(
            anise.framework.report.ExecutionScopePart(e_full, hints=["logdebug"]))
        anise.framework.report.topscope().appendchild(
            anise.framework.report.ExecutionScopePart("Error in execution: "+str(e), hints=["logerror"]))
        raise
    finally:
        if _project:
            try:
                for hook in _project.gethooks(HOOK_SHUTDOWN):
                    hook()
            except Exception as e:
                anise.utils.logging.logexception(e)
            anise.framework.projects.unsetcurrentuniverse()
    # return the task result
    return res


def _parsevalue(l):
    """!
    Parses a commandline universe object value assignment like `s:name=foo`.
    """
    i = l.find("=")
    if i == -1:
        return None, None, None
    else:
        return l[0], l[2:i], l[i+1:]


def _cmdline(args):
    """!
@page commandline

<div this_is_a_workaround_for_doxygen_bug></div>

The Anise command-line tool understands this syntax:

@code
anise [taskname] [options]*
@endcode

If a taskname is given, Anise will execute this task, otherwise the `DEFAULT`
task, which typically is a task chooser.

Options can be some of the following:

- `--projectfile [file]` : Use this project file (instead of the typical
   location)

- `--loadfeaturesfrom [path1:...:pathn]` : Sets the paths for loading modules
   (separated by ":"; builtin features by "builtin")

- `--debug` : Enables debug log output

- `--keeptemp` : Don't remove the temporary stuff (for post-execution
   diagnostic)

- `--no-autoopen` : Never automatically open files after execution

- `--no-gui` : Don't try to use graphic mode for user interaction

- `--createfresh` : Create a fresh project description file

- `--s:key=value` : Sets a value to the universe object before execution; also
    works with `i`, `b` and `f` at the beginning for integers, bools and floats
   instead of strings

- `--help` : Shows this help text
   """
    c = 0

    def nextparam():
        nonlocal c
        nonlocal args
        v = args[c] if c < len(args) else None
        c += 1
        return v
    values = {}
    projectfile = None
    taskname = None
    loadfeaturesfrom = "builtin:~/.anise/features:/var/lib/anise/features"
    createfresh = False
    showhelp = False
    if len(args) > 0 and args[0][0:2] != "--":
        taskname = args[0]
        c += 1
    _p = nextparam()
    while _p is not None:
        if _p[0:2] == "--":
            p = _p[2:]
            if p == "projectfile":
                projectfile = nextparam()
            elif p == "loadfeaturesfrom":
                loadfeaturesfrom = nextparam()
            elif p == "debug":
                anise.framework.globalvars.logdebug = True
            elif p == "keeptemp":
                anise.framework.globalvars.removetempdirs = False
            elif p == "no-gui":
                values["ui.force_mode"] = "TERMINAL"
            elif p == "no-autoopen":
                values["base.autoopenresult"] = False
            elif p == "createfresh":
                createfresh = True
            elif p == "help":
                showhelp = True
            elif len(p) > 2 and p[1] == ":":
                stype, skey, svalue = _parsevalue(p)
                if stype is not None:
                    if stype == "s":
                        values[skey] = svalue
                    elif stype == "i":
                        values[skey] = int(svalue)
                    elif stype == "b":
                        values[skey] = anise.framework.projects.IntermediateStructure.Value.istrue(svalue)
                    elif stype == "f":
                        values[skey] = float(svalue)
            else:
                raise Exception("unknown parameter: " + p)
        else:
            raise Exception("parse error: " + _p)
        _p = nextparam()
    loadfeaturesfrom = [x for x in loadfeaturesfrom.split(":") if x.strip() != ""]
    loadfeaturesfrom = [(os.path.abspath(os.path.expanduser(x)) if x != "builtin" else "builtin")
                        for x in loadfeaturesfrom]
    return anise.utils.data.constructobject(taskname=taskname, values=values, projectfile=projectfile,
                                            loadfeaturesfrom=loadfeaturesfrom, createfresh=createfresh,
                                            showhelp=showhelp)


def main():
    """!
    The main method when called from command line. Parses command line arguments and executes the chosen task.
    """
    cmdline = _cmdline(sys.argv[1:])
    if cmdline.showhelp:
        print(_cmdline.__doc__[_cmdline.__doc__.find("@code")+5:].replace("@endcode", ""))
        sys.exit()
    elif cmdline.createfresh:
        if os.path.exists("./_meta/_projectdesc"):
            anise.utils.logging.logerror("There already exists a _meta/_projectdesc.", printcaller=False)
            sys.exit(1)
        anise.utils.basic.mkdirs("./_meta")
        username = pwd.getpwuid(os.getuid())[0] if sys.platform != "win32" else "johndoe"
        anise.utils.basic.writetofile("./_meta/_projectdesc", """<?xml version="1.0" ?>
<aniseprojectdesc>
    <val k="name" v="{0}"/>
    <val k="base.version" v="0.1"/>
    <val k="base.summary" v="something spectacular {0}ish"/>
    <val k="base.email" v="mail@localhost"/>
    <val k="base.authors"><val v="{1}"/></val>
    <val k="base.license" oref="licensing.PublicDomain"/>
    <val k="base.longdescription">
Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam
nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam
erat, sed diam voluptua.
    </val>
</aniseprojectdesc>
""".format(os.path.basename(os.getcwd()), username))
        anise.utils.logging.loginfo(("A new Anise project description file was created as:\n  {0}/_meta/_projectdesc\n"
                                     "Please call Anise from now on from directory:\n  {0}\n\n"
                                     "Call 'anise console' now in order to configure your project.\n"
                                     "Advanced users may directly change the project description file instead.\n\n"
                                     "Please also consider to put this file under version control.\n")
                                    .format(os.getcwd()), printcaller=False)
        sys.exit()
    taskname = cmdline.taskname
    projectfile = cmdline.projectfile
    if projectfile is None and os.path.isfile("./_projectdesc"):
        projectfile = "./_projectdesc"
    elif projectfile is None and os.path.isfile("./_meta/_projectdesc"):
        projectfile = "./_meta/_projectdesc"
    if projectfile is not None:
        projectfile = os.path.abspath(projectfile).replace("\\", "/")
        projectdir = os.path.dirname(projectfile).replace("\\", "/")
        if projectdir.endswith("/_meta"):
            projectdir = projectdir[0:-6]
    else:
        projectdir = None
        anise.utils.logging.logerror(
            "No project description file found. Call Anise from your project root directory.\n\n"
            "For creating a fresh one, go to the project root directory and call\n"
            "anise --createfresh\n", printcaller=False, terminalonly=True)
    if taskname is None:
        taskname = "DEFAULT"
    retcode = 0
    try:
        execute(taskname, projectfile, projectdir, cmdline.values, cmdline.loadfeaturesfrom)
    except BaseException as e:
        retcode = 1
        anise.utils.logging.logexception(e)
    sys.exit(retcode)
