# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

## @package anise.framework.features
## The anise plug-in system.
##
## This package does not include the anise high-level features itself. You would find them in `anise.features`.
## This package is the part of the anise infrastructure which loads and manages those features.
##
## For each anise run, it automatically loads the built-in `anise.features` as well as custom features from
## `~/.anise/features` and `/var/lib/anise/features`. Due to their locations, the latter ones are useful for
## per-machine and per-user customization.
##
## In an arbitrary anise execution context, you may call `anise.framework.features.loadfeature` in order to access
## functionality from other features, like here:
##
##\verbatim
##from anise.framework import features
##_somefeature = features.loadfeature("somefeature").featuremodule
##_somefeature.some_method()
##x = _somefeature.some_variable
##\endverbatim

import importlib
import importlib.machinery
import importlib.util
import os

import anise.framework.exceptions
import anise.utils


class internals:
    """!
    Internal stuff.
    """

    @staticmethod
    def injectdefaultstoproject(project):
        """!
        Adds feature data structures to the universe object.
        """
        for feature in _features:
            fname = feature.name if ("." not in feature.name) else feature.name[feature.name.rindex(".")+1:]
            setattr(anise.utils.data.createobjectstructure(project, feature.name, PseudoFeature)[1], fname,
                    feature.featuremodule)

    @staticmethod
    def injecthookstoproject(project):
        """!
        Adds hooks to the universe object, which are loaded in feature loading.
        """
        for feature in _features:
            for h in feature._hookhandlers:
                project.addhook(h.hookobj, h.function, None, h.provides, h.requires, h.prepares)

    @staticmethod
    def loadfeatures(featurespath, name=""):
        """!
        Loads features from a path.
        """
        l = [(featurespath, name)]
        ll = []
        while len(l) > 0:
            xpath, xname = l.pop()
            _featuresources.append((xpath, xname))
            if os.path.isdir(xpath):
                for f in sorted(os.listdir(xpath)):
                    nname = xname
                    if nname != "":
                        nname += "."
                    nname += f
                    if f.startswith("__"):
                        continue
                    if nname.endswith(".py"):
                        nname = nname[:-3]
                    l.append((xpath+"/"+f, nname))
            elif os.path.isfile(xpath):
                ll.append(xname)
        for xxname in ll:
            loadfeature(xxname)

    @staticmethod
    def resetfeatures():
        """!
        Resets the features.
        """
        global _features
        global _featuresources
        anise.utils.logging.logdebug("Resetting features.")
        _features = []
        _featuresources = []


class HookKind:
    """!
    Enumeration of hook kinds.
    """

    FUNCTION = "FUNCTION"
    CLASS = "CLASS"


class Hook:
    """!
    A hook. See documentation for details.
    """

    def __init__(self, hookkind=HookKind.FUNCTION, *, hide=False, interesting=5, doc=None, wizard_functionparams=None,
                 wizard_functionadditionallog=None, wizard_functionbody=None, wizard_classbase=None,
                 wizard_classbody=None, wizard_loadfeatures=None):
        """!
        @param hookkind: A hook kind (one of HookKind).
        @param hide: If this hook is to be hidden in hook choosers.
        @param interesting: A number which controls how interesting this hook is in a chooser.
        @param doc: Hook documentation text.
        @param wizard_functionparams: Info for hook handler wizards. List of function parameters.
        @param wizard_functionadditionallog: Info for hook handler wizards. List of additional variable to log out in a 
                                             function.
        @param wizard_functionbody: Info for hook handler wizards. Additional function body part.
        @param wizard_classbase: Info for hook handler wizards. Base class.
        @param wizard_classbody: Info for hook handler wizards. Additional class body part.
        @param wizard_loadfeatures: Info for hook handler wizards. Additional features to load.
        """
        self.hookkind = hookkind
        self.hide = hide
        self.interesting = interesting
        self.__doc__ = doc
        self.wizard_functionparams = wizard_functionparams or []
        self.wizard_functionadditionallog = wizard_functionadditionallog or []
        self.wizard_functionbody = wizard_functionbody
        self.wizard_classbase = wizard_classbase
        self.wizard_classbody = wizard_classbody
        self.wizard_loadfeatures = wizard_loadfeatures or []

    def gethandlers(self):
        """!
        Returns a list of all registered handlers for this hook.
        """
        from anise.framework.projects import universe
        return universe.gethooks(self)


class HookHandler:
    """!
    A hook handler bound to some hook. See documentation for details.
    """

    def __init__(self, hookobj, function, provides, requires, prepares):
        """!
        @param hookobj: The hook.
        @param function: The handler function.
        @param provides: A list of symbol names this hook participates to provide (dependency provider).
        @param requires: A list of symbol names this hook requires to be executed before.
        @param prepares: A list of symbol names this hook must prepend.
        """
        self.hookobj = hookobj
        self.function = function
        self.provides = tuple(provides.split(" ") if isinstance(provides, str) else provides)
        self.requires = tuple(requires.split(" ") if isinstance(requires, str) else requires)
        self.prepares = tuple(prepares.split(" ") if isinstance(prepares, str) else prepares)


class PseudoFeature:
    """!
    This is an auxiliary class, which fixes the gap which appears, e.g. if a feature 'foo.bar' exists, but no 'foo' one.
    """
    pass


class Feature:
    """!
    A feature is some encapsuled part of anise high-level functionality. This class represents a loaded feature
    at runtime. The framework as well as other features access this object for interacting with that feature.
    Direct usage of high-level features from within `_projectdesc` is intended to not use this Feature objects
    but to use the embedded properties of the universe object.
    """

    def __init__(self, name, hookhandlers, featuremodule, hooks):
        """!
        @param name: The optional name of the feature.
        @param hookhandlers: All hook handlers this feature registers globally (it may do that later on in a function
                            as well).
        @param featuremodule: The feature module.
        @param hooks: List of Hook instances.
        """
        self.name = name
        self._hookhandlers = list(hookhandlers)
        self.featuremodule = featuremodule
        self.hooks = hooks

_features = []

_featuresources = []


def loadfeature(name):
    """!
    Load an anise.framework.features.Feature by name from the predefined sources (as optional command-line argument).
    If if is already loaded, it returns the loaded feature again.
    """
    global _hooks
    for f in _features:
        if f.name == name:
            return f
    frompath = None
    for srcfile, srcname in _featuresources:
        if srcname == name:
            frompath = srcfile
            break
    if not frompath:
        raise anise.framework.exceptions.UnexpectedSituationError("unable to load feature '" + name + "'.")
    anise.utils.logging.logdebug("Load feature from path '{0}' to name '{1}'.".format(frompath, name))
    _hooks.append([])
    try:
        loader = importlib.machinery.SourceFileLoader(name, frompath)
        spec = importlib.util.spec_from_loader(loader.name, loader)
        fmodule = importlib.util.module_from_spec(spec)
        loader.exec_module(fmodule)
        for hook in _hooks[-1]:
            if name not in hook.provides:
                hook.provides += (name,)
            oname = getattr(hook.function, "__name__", None)
            if oname:
                oname = name + "." + oname
                if oname not in hook.provides:
                    hook.provides += (oname,)
        hooks = {}
        for ds in dir(fmodule):
            dos = getattr(fmodule, ds)
            if isinstance(dos, Hook):
                hooks[ds] = dos
        _feature = Feature(name, _hooks[-1], fmodule, hooks)
    finally:
        _hooks.pop(len(_hooks)-1)
    _features.append(_feature)
    _features.sort(key=lambda x: x.name)
    return _feature

_hooks = []


def hookhandler(name, *, provides=None, requires=None, prepares=None):
    """!
    Used by a feature module to register a handler for a hook globally. The result should be used as Python function
    decorator.
    @param name: The hook which you want to register a handler for.
    @param provides: A list of symbol names this hook participates to provide (dependency provider).
    @param requires: A list of symbol names this hook requires to be executed before.
    @param prepares: A list of symbol names this hook must prepend.
    """
    def _realdecorator(func):
        _hooks[-1].append(HookHandler(name, func, provides or [], requires or [], prepares or []))
        return func
    return _realdecorator


def getfeatures():
    """!
    Returns a list of all features (as Feature).
    """
    return list(_features)


## This hook gets executed by the anise engine right before the `_projectdesc` file is loaded.
HOOK_BEFORE_DEFINITION = Hook(doc="This hook is called before the project description file is loaded.")

## This hook gets executed by the anise engine after the `_projectdesc` file is loaded and right before the
## chosen task itself is called. At this time all the assignments and global statements in the `_projectdesc`
## are already applied.
HOOK_BEFORE_EXECUTION = Hook(doc="This hook is called after the project description file is loaded and custom "
                                 "initialization in HOOK_AFTER_PREPARATION took place.")
