# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

## @package anise.framework.files
## File structures, used for data exchange between methods or as result.

import atexit
import os
import shutil
import sys
import tempfile

import anise.framework.exceptions
import anise.framework.globalvars
import anise.utils


class Filestructure:
    """!
    This class represents a file or a directory structure with many subdirectories and files.
    There are methods for writing those items to disk and/or to modify the original structure
    in the filesystem.

    It is more powerful for file exchange than working just with file paths, because a Filestructure can also
    carry arbitrary metadata information and allows dynamic content creation (see e.g. TaskExecution subclass).

    Instances of this class (and its subclasses) are used by many features for file exchange.
    """

    def __init__(self, path, relative=False):
        """!
        @param path: Path to the file or directory.
        @param relative: If path is relative to the project directory.
        """
        if relative and path:
            from anise.framework.projects import universe
            path = universe.specialpaths.projectdir + "/" + path
        self._path = os.path.abspath(path) if (path is not None) else None
        self._data = {}
        self._data["autoopen"] = None
        self._data["autoopenterminal"] = None
        self._initialized = False

    def datakeys(self):
        """!
        Returns a list of keys for all stored metadata properties.
        """
        return list(self._data.keys())

    def setdata(self, k, v):
        """!
        Sets a metadata property.
        @param k: The key name.
        @param v: The value.
        """
        self.initialize()
        self._data[k] = v

    def getdata(self, k, deflt):
        """!
        Gets a metadata property.
        @param k: The key name.
        @param deflt: The default value (returned if the key doesn't exist).
        """
        self.initialize()
        return self._data.get(k, deflt)

    def initialize(self):
        """!
        Initializes this Filestructure. When this is called, the dynamic content (if any) is created and written
        to disc. It is automatically called for typical usage and is not required to be called manually.
        """
        if self._initialized:
            return
        self._initialize()
        self._initialized = True

    def _initialize(self):
        """!
        The internal initialization routine, which fetches dynamic content (if any). Called from outside before
        actual accesses occur.
        @a_dooverride_optional
        """
        pass

    def path(self):
        """!
        The path to this structure in the filesystem.
        """
        self.initialize()
        return self._path

    def dl(self, *, subpath="", to=None, progresscallback=None):
        """!
        Copies the complete filestructure or a subdirectory to a new destination.
        @param subpath: If given, only this subdirectory will be copied.
        @param to: If given, it specifies the destination path; if not given, a new directory in the current working
                   directory will be used.
        @param progresscallback: Callback function for retrieving progress information.
        """
        self.initialize()
        # prepare destination
        if not to:
            to = os.getcwd() + "/" + os.path.basename(self.path())
            if os.path.exists(to):  # we don't do that if we have an explicit destination
                anise.utils.logging.logwarning("removing old "+to, printcaller=False)
                if os.path.isdir(to):
                    shutil.rmtree(to)
                else:
                    os.unlink(to)
        # copy
        fullpath = os.path.abspath(self.path() + "/" + subpath)
        if os.path.isfile(fullpath):
            dto = os.path.dirname(to)
            anise.utils.basic.mkdirs(dto)
            files = [""]
        elif os.path.isdir(fullpath):
            anise.utils.basic.mkdirs(to)
            files = os.listdir(fullpath)
        else:
            raise anise.framework.exceptions.UnexpectedSituationError("does not exist: " + fullpath)
        ffiles = []
        while len(files) > 0:
            fn = files.pop(0)
            ff = os.path.abspath(fullpath + "/" + fn)
            d = os.path.abspath(to + "/" + fn)
            if os.path.islink(ff):
                os.symlink(os.readlink(ff), d)
            elif os.path.isdir(ff):
                anise.utils.basic.mkdirs(d)
                files += [fn+"/"+x for x in os.listdir(ff)]
            elif os.path.isfile(ff):
                ffiles.append((ff, d, os.path.getsize(ff)))
            else:
                raise anise.framework.exceptions.UnexpectedSituationError("unknown thing: " + ff)
        ffullsize = 0
        for ff, d, fsize in ffiles:
            ffullsize += fsize
        fdonesize = 0
        freportedperc = None
        for ff, d, fsize in ffiles:
            if os.path.isdir(d):
                d = os.path.join(d, os.path.basename(ff))
            with open(ff, 'rb') as fsrc:
                with open(d, 'wb') as fdst:
                    while True:
                        buf = fsrc.read(16*1024)
                        if not buf:
                            break
                        fdst.write(buf)
                        fdonesize += len(buf)
                        fmyperc = int(100. * fdonesize / ffullsize)
                        if freportedperc != fmyperc:
                            freportedperc = fmyperc
                            if progresscallback:
                                progresscallback(ffullsize, fdonesize)
            shutil.copystat(ff, d)
        return to

    def with_modified_rootname(self, newname):
        """!
        Returns a new (temporary) anise.framework.files.Filestructure clone with a new directory name.
        @param newname: The name which the new root directory should get.
        """
        self.initialize()
        newtmp = TempDir()
        nname = newtmp.path() + "/" + newname
        self.dl(to=nname)
        return Filestructure(nname)

    def mv(self, newname):
        """!
        Returns a new anise.framework.files.Filestructure resulting in a movement from the old one.
        @param newname: The new name (absolute or relative).
        """
        self.initialize()
        newpath = newname
        if sys.platform == "win32":
            newpath = newpath.replace("\\", "/")
        if "/" not in newpath:
            newpath = "../" + newpath
        if not newpath.startswith("/"):
            newpath = self.path() + "/" + newpath
        newpath = os.path.abspath(newpath)
        os.rename(self.path(), newpath)
        return Filestructure(newpath)

    def setautoopen(self, path, interminal=False):
        """!
        Places a hint which requests the anise engine to automatically open a file within this structure when
        it is returned to the user.
        @param path: The relative path to the file which should automatically be opened.
        @param interminal: If this hint is for terminal (non-graphical) usage.
        """
        self.initialize()
        if interminal:
            self.setdata("autoopenterminal", path)
        else:
            self.setdata("autoopen", path)


class TempDir(Filestructure):
    """!
    A special anise.framework.files.Filestructure which automatically creates a temporary directory in background.

    Example:

    \verbatim
    def sometask():
        mytmp = files.TempDir()  # creates a temporary directory somewhere in no-man's-land
        myfile = mytmp.path() + "/somefile"
        with open(myfile, "w") as f:
            f.write("some content for our new file")
        return mytmp
    \endverbatim
    """

    def __init__(self, dirname=None, temprootpath=None, namepattern="anise.{pid}.{i}"):
        """!
        @param dirname: An optional name which the root directory should get; if not given, it gets an random name.
        @param temprootpath: An optional alternate temp root directory path.
        @param namepattern: The directory name pattern.
        """
        i = 0
        path = None
        _pt = (temprootpath or tempfile.gettempdir()) + "/" + namepattern
        pid = os.getpid()
        while path is None:
            _p = _pt.format(**locals())
            i += 1
            try:
                if not os.path.exists(_p):
                    os.mkdir(_p)
                    path = _p
                    atexit.register(self.delete)
            except Exception:
                pass
        self.origpath = path
        if dirname:
            path += "/" + dirname
            os.mkdir(path)
        Filestructure.__init__(self, path)

    def __enter__(self):
        return self

    def delete(self):
        """!
        Deletes the content from the hard drive.
        """
        if anise.framework.globalvars.removetempdirs:
            try:
                shutil.rmtree(self.origpath)
            except Exception:
                pass

    def __exit__(self, etype, evalue, etraceback):
        self.delete()


class TaskExecution(Filestructure):
    """!
    Holds a function and some parameters for calling later on (like a delegate;
    or a more explicit, specialized and 'xml-compatible' version of a lambda) in order to
    get a Filestructure as function result. It acts as a proxy on top of this Filestructure.
    """

    def __init__(self, task, **kwargs):
        """!
        @param task: The task implementation for execution.
        """
        Filestructure.__init__(self, None)
        self.task = task
        self.params = kwargs

    def _initialize(self):
        self.inner = self.task(**self.params)
        self._path = self.inner.path()
        self._data = self.inner._data


class TextFileByContent(Filestructure):
    """!
    Holds a function and some parameters for calling later on (like a delegate;
    or a more explicit, specialized and 'xml-compatible' version of a lambda) in order to
    get a Filestructure as function result. It acts as a proxy on top of this Filestructure.
    """

    def __init__(self, content):
        """!
        @param content: The text content for the file.
        """
        fpath = TempDir().path() + "/cnt"
        with open(fpath, "w") as f:
            f.write(content)
        Filestructure.__init__(self, fpath)


class AbstractRebuildDirectoriesFilestructure(TaskExecution):
    """!
    Abstract base class for merging directory structures together. 
    
    See the subclasses.
    """

    def __init__(self):
        TaskExecution.__init__(self, self.getmerged)

    def getdests(self):
        """!
        Returns a list of transfer destinations.
        @return: A list of tuples, each beginning with the destination path (and other custom ones).
        """
        return []

    def getsource(self, desttuple):
        """!
        Returns the transfer source (as anise.framework.files.Filestructure) for a destination tuple (as it was part of
        getdests' result).
        """
        return None

    def getmerged(self):
        from anise.framework.projects import universe
        projectname = getattr(universe, "name", "aniseproject")
        result = TempDir()
        dests = self.getdests()
        for desttuple in sorted(dests, key=lambda aa: aa[0]):
            part = desttuple[0]
            fd = os.path.abspath(result.path() + "/" + part.format(name=projectname))
            anise.utils.basic.mkdirs(os.path.dirname(fd))
            srcfs = self.getsource(desttuple)
            if os.path.isdir(srcfs.path()):
                srcfs.dl(to=fd)
            elif os.path.isfile(srcfs.path()):
                shutil.copy2(srcfs.path(), fd)
            else:
                raise anise.framework.exceptions.UnexpectedSituationError("unknown thing: " + srcfs.path())
        return result


class RemappedFilestructure(AbstractRebuildDirectoriesFilestructure):
    """!
    Remaps a Filestructure's inner directory hierarchy into a new tree.
    """

    def __init__(self, source=None, frommap=None):
        """!
        @param source: The source anise.framework.files.Filestructure.
        @param frommap: A dict of string/Filestructure. Values are relative source paths with the dict key 
                        specifying a target subpath.
        """
        AbstractRebuildDirectoriesFilestructure.__init__(self)
        self.source = source
        self.frommap = frommap or {}

    def getdests(self):
        res = self.frommap
        if hasattr(res, "items"):
            res = res.items()
        return res

    def getsource(self, desttuple):
        return Filestructure(self.source.path() + "/" + desttuple[1])


class MergedFilestructure(AbstractRebuildDirectoriesFilestructure):
    """!
    Merges many Filestructure instances together in one tree.
    """

    def __init__(self, fromlist=None, frommap=None):
        """!
        @param fromlist: A list of Filestructure. Become root directory entries with their original file names.
        @param frommap: A dict of string/Filestructure. Values become part of the result tree with the dict key 
                        specifying a subpath.
        """
        AbstractRebuildDirectoriesFilestructure.__init__(self)
        self.fromlist = fromlist or []
        self.frommap = frommap or {}

    def getdests(self):
        res = self.frommap
        if hasattr(res, "items"):
            res = res.items()
        res = list(res)
        for fl in self.fromlist:
            res.append(("/", fl))
        return res

    def getsource(self, desttuple):
        return desttuple[1]
