# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

## @package anise.framework.projects
## Implementation for the anise universe object and some services around it.

import importlib
# noinspection PyPep8Naming
import xml.etree.cElementTree as ET

import anise.data.PyDepsEngine.depsengine
import anise.framework.exceptions
import anise.framework.features
import anise.utils


_currentuniverse = []


def setcurrentuniverse(project):
    """!
    Sets the current universe object.
    @a_infrastructure
    """
    _currentuniverse.append(project)


def unsetcurrentuniverse():
    """!
    Unsets the current universe object.
    @a_infrastructure
    """
    _currentuniverse.pop()


def getvalue(key, defaultval=None):
    """!
    A convenience method returning the value of a certain property for the current universe object.
    """
    return currentuniverse().getvalue(key, defaultval)


def setvalue(key, value):
    """!
    A convenience method setting the value of a certain property for the current universe object.
    """
    return currentuniverse().setvalue(key, value)


def currentuniverse():
    """!
    Returns the current universe object.
    """
    return _currentuniverse[-1] if len(_currentuniverse) > 0 else None


class _UniverseProxy:
    """!
    Internally used for the universe object for allowing just to write `universe` instead of `universe()`.
    """

    def __getattr__(self, item):
        return getattr(currentuniverse(), item)

    def __setattr__(self, key, value):
        setattr(currentuniverse(), key, value)

    def __call__(self, *args, **kwargs):  # compatibility with existing code
        return currentuniverse()


universe = _UniverseProxy()


class Universe:
    """!
    Implementation for universe objects.
    """

    def __init__(self, d, executionscope=None):
        self.executionscope = executionscope
        self.append(d)
        self.hooks = {}
        self._registeredtasks = {}

    class RegisteredTask:
        """!
        A task which is registered to the universe object. You typically don't need to work with this structure
        directly. Use Universe.addtask to add new tasks.
        """

        def __init__(self, parentname, taskname, label, important, invisible, sortindex, function):
            self.parentname = parentname
            self.taskname = taskname
            self.name = ((parentname + ".") if parentname else "") + taskname
            self.label = label
            self.important = important
            self.invisible = invisible
            self.sortindex = sortindex
            self.function = function

    def addtask(self, function, name="", *, label=None, important=True, invisible=False, sortindex=0):
        """!
        Registers a task and makes it available for execution.
        @param function: The function which implements the task.
        @param name: The task name.
        @param label: The task label text.
        @param important: If this is an important/interesting task (presented differently in choosers).
        @param invisible: If this task should be hidden in choosers.
        @param sortindex: A number which controls the display order of tasks in choosers.
        """
        if not name or name.endswith("."):
            if not getattr(function, "__name__", None):
                raise anise.framework.exceptions.BadInputError(
                    "'function' must have a '__name__' attribute or 'name' must be a complete function name.")
            name += function.__name__
        ildot = name.rfind(".")
        if ildot > -1:
            parentname = name[:ildot]
            taskname = name[ildot+1:]
            parent = eval(parentname, self.__dict__)
        else:
            parentname = None
            taskname = name
            parent = self
        setattr(parent, taskname, function)
        self._registeredtasks[name] = self.RegisteredTask(parentname, taskname, label, important, invisible, sortindex,
                                                          function)

    def registeredtasks(self):
        """!
        Returns a list of all registered tasks (as Universe.RegisteredTask).
        """
        return self._registeredtasks.values()

    def append(self, d):
        """!
        Stores a dictionary of values as properties of this universe object.
        """
        class DataStructure:
            pass
        for key in d:
            o = self
            parts = key.split(".")
            for kkey in parts[:-1]:
                ino = getattr(o, kkey, None)
                if ino is None:
                    ino = DataStructure()
                    setattr(o, kkey, ino)
                o = ino
            setattr(o, parts[-1], d[key])

    def addhook(self, hookobj, fct, params=None, provides=None, requires=None, prepares=None):
        """!
        Adds a handler for a hook to the universe object.
        @return: An anise.framework.features.HookHandler instance.
        """
        if params is None:
            params = {}
        if callable(fct):
            nfct = (lambda *a, **b: fct(*a, **dict(list(params.items())+list(b.items()))))
        else:
            nfct = fct
        hook = anise.framework.features.HookHandler(hookobj, nfct, provides or [], requires or [], prepares or [])
        hooklst = self.hooks.get(hookobj, None)
        if not hooklst:
            hooklst = self.hooks[hookobj] = anise.data.PyDepsEngine.depsengine.Engine()
        hooklst.add_object(hook, own=provides, afteroptional=requires, beforeoptional=prepares)
        return hook

    def removehook(self, hookhandler):
        """!
        Removes a hook, which was added before.
        @param hookhandler: The HookHandler you got in addhook.
        """
        hooklst = self.hooks.get(hookhandler.hookobj, None)
        if hooklst:
            hooklst.remove_object(hookhandler)

    def gethooks(self, hookobj):
        """!
        Gets the handlers for a hook in the correct order.
        @param hookobj: The hook you want to get handlers for.
        """
        result = []
        hooklst = self.hooks.get(hookobj, None)
        if hooklst:
            result = [x.function for x in hooklst.get_objects()]
        return result

    def getvalue(self, key, defaultvalue=None):
        """!
        Returns the value of a certain property for the current universe object.
        
        If this property does not exist, it returns the `defaultval`.
        `getvalue("foo.bar")` is similar to `universe.foo.bar`.
        @param key: The key name of the property to get.
        @param defaultvalue: This value is returned if the specified property does not exist.
        """
        x = self
        keysegs = key.split(".")
        notfound = object()
        if (len(keysegs) > 0) and (keysegs[0] in ["universe", "universe()"]):
            keysegs.pop(0)
        for k in keysegs:
            if k:
                x = getattr(x, k, notfound)
                if x == notfound:
                    return defaultvalue
        return x

    def setvalue(self, key, value):
        """!
        Sets the value of a certain property for the current universe object.
        
        If a parent property does not exist, it does nothing. Returns True if succeeded. 
        `setvalue("foo.bar", "baz")` is similar to `universe.foo.bar = "baz"`.
        @param key: The key name of the property to set.
        @param value: The new value for the specified property.
        """
        keysegs = [k for k in key.split(".") if k]
        if len(keysegs) > 0:
            x = self.getvalue(".".join(keysegs[:-1]), None)
            if x:
                setattr(x, keysegs[-1], value)
                return True


class IntermediateStructure:
    """!
    A data structure which holds an in-memory representation of the project description.
    It is not the in-memory representation of the live universe object state.
    It is like the file
    structure of the project description file, but not exactly the same. There is a parser and a generator
    for translating between a project description file and an IntermediateStructure.
    """

    class ParseProjectDescriptionError(anise.framework.exceptions.BadFormatError):
        """!
        Parsing error like bad input xml.
        """
        pass

    class ParseProjectDescriptionInternalError(anise.framework.exceptions.InternalError):
        """!
        Parsing error which seems like a bug.
        """
        pass

    class Node:
        """!
        A node in the intermediate structure.
        """

        def __init__(self, k, **args):
            """!
            @param k: The node key.
            """
            self._parent = None
            self._args = {}
            self._key = k
            for arg in args:
                self.setarg(arg, args[arg])
            self.islet = False
            self.isvalue = False

        def key(self):
            """!
            The key name.
            """
            return self._key

        def parent(self):
            """!
            Returns the parent node (either IntermediateStructure or IntermediateStructure.Node).
            """
            return self._parent

        def args(self):
            """!
            Returns the list of stored argument keys.
            """
            return self._args.keys()

        def getarg(self, key):
            """!
            Returns an argument (as IntermediateStructure.Value).
            @param key: The argument key.
            """
            return self._args.get(key, None)

        def setarg(self, key, node):
            """!
            Sets an argument.
            @param key: The argument key.
            @param node: The argument value (as IntermediateStructure.Value).
            """
            if not isinstance(node, IntermediateStructure.Value):
                raise anise.framework.exceptions.BadInputError("Not an IntermediateStructure.Value: "+repr(node))
            node._parent = self
            node._key = key
            oldnode = self.getarg(key)
            if oldnode:
                oldnode.remove()
            self._args[key] = node

        def removearg(self, key):
            """!
            Removes an argument.
            @param key: The argument key.
            """
            return self._args.pop(key)

        def clearargs(self):
            """!
            Clears the list of arguments.
            """
            self._args.clear()

        def remove(self):
            """!
            Removes this node.
            """
            if isinstance(self.parent(), IntermediateStructure):
                self.parent().removeentry(self)
            else:
                self.parent().removearg(self.key())

    class Value(Node):
        """!
        A Value entry in the project description structure. This corresponds to a `val` node in the xml format.
        It represents an arbitrary value (primitive or composed type).
        """

        def __init__(self, t, *, k=None, v="", oref="None", **args):
            """!
            @param t The value type
            @param k The key (i.e. property name in parent object)
            @param v (optional) The value string for a primitive value.
            @param oref (optional) The oref string for a objectref value.
            """
            IntermediateStructure.Node.__init__(self, k, **args)
            self._vtype = t
            self._value = v
            self._oref = oref
            self.isvalue = True

        def appendarg(self, node):
            """!
            Appends a new IntermediateStructure.Value to a list node.
            """
            if self.vtype() != "list":
                raise anise.framework.exceptions.BadInputError("appendarg requires a list, but got: "+repr(self))
            self.setarg("i9999999", node)
            self._realign_listvalues()

        def _realign_listvalues(self):
            """!
            Makes internal list realignment tasks.

            This is required to call after the list structure changes.
            """
            tmp = {}
            for i, x in enumerate(sorted(self.args())):
                tmp["i" + ("00000" + str(i))[:6] + "0"] = self.getarg(x)
            self.clearargs()
            for x in tmp:
                self.setarg(x, tmp[x])

        def value(self):
            """!
            Returns the value (for primitive values).
            """
            return self._value

        def setvalue(self, v):
            """!
            Sets the value (for primitive values).
            """
            self._value = v

        def vtype(self):
            """!
            Returns the value type name.
            """
            return self._vtype

        def setvtype(self, t):
            """!
            Sets the value type name.
            """
            self._vtype = t

        def oref(self):
            """!
            Returns the objectref (for oref values).
            """
            return self._oref

        def setoref(self, v):
            """!
            Sets the objectref (for oref values).
            """
            self._oref = v

        def _setprimitive(self, v):
            """!
            Setting a primitive value with typechecking.
            """
            vv = str(v)
            eval(self._vtype)(vv)  # value checking
            self.setvalue(vv)

        def getprimitiverepr(self):
            """!
            Returns a string representation of the value, if it is a primitive one (None otherwise).
            """
            if self.vtype() == "str":
                return '"' + self.value() + '"'
            elif self.vtype() in ["int", "float"]:
                return str(self.value())
            elif self.vtype() == "bool":
                return str("True" if IntermediateStructure.Value.istrue(self.value()) else "False")
            elif self.vtype() == "objectref":
                return self.oref()

        @staticmethod
        def istrue(s):
            """!
            Returns True iff the given string can represent a boolean 'True' value.
            """
            return s.lower() in ["1", "y", "true"]

        @staticmethod
        def fromstring(val, key=None):
            """!
            Convenience method which creates a string Value.
            """
            return IntermediateStructure.Value("str", k=key, v=str(val))

        @staticmethod
        def frombool(val, key=None):
            """!
            Convenience method which creates a boolean Value.
            """
            return IntermediateStructure.Value("bool", k=key, v="True" if val else "False")

        def __str__(self):
            if self.vtype() in ["str", "int", "bool", "float"]:
                return str(self.value())
            elif self.vtype() == "objectref":
                return str(self.oref())
            elif self.vtype() == "list":
                return str([str(self.getarg(x)) for x in sorted(self.args())])
            elif self.vtype() == "dict":
                return str({x: str(self.getarg(x)) for x in self.args()})
            else:
                return "['{0}' ({1}) {2}]".format(self.key(), self.vtype(), "; ".join(
                                                    [x + "=" + str(self.getarg(x)) for x in sorted(self.args())]))

    class Let(Node):
        """!
        A Let entry in the project description structure. This corresponds to a `let` node in the xml format.
        It represents the execution of a parameterized part of anise feature program logic.
        """

        def __init__(self, k, **args):
            """!
            @param k The full qualified method name.
            """
            IntermediateStructure.Node.__init__(self, k, **args)
            self.islet = True

        def __str__(self):
            return "[LET '{0}' {1}]".format(self.key(), "; ".join([x+"="+str(self._args[x])
                                                                   for x in sorted(self._args)]))

    def __init__(self, entries=None):
        """!
        @param entries: The new list of direct subitems (IntermediateStructure.Node).
        """
        self._entries = []
        self.sourceformat = None
        if entries:
            for ent in entries:
                self.addentry(ent)

    def entries(self):
        """!
        Returns the list of direct subitems (IntermediateStructure.Node).
        """
        return list(self._entries)

    def addentry(self, node):
        """!
        Adds a new entry to the list of direct subentries.
        @param node: The new IntermediateStructure.Node.
        """
        self._entries.append(node)
        node._parent = self

    def removeentryat(self, i):
        """!
        Removes an entry from the list of direct subentries.
        @param i: The entry position to remove.
        """
        return self._entries.pop(i)

    def removeentry(self, node):
        """!
        Removes an entry from the list of direct subentries.
        @param node: The IntermediateStructure.Node to remove.
        """
        return self._entries.remove(node)

    def setentries(self, entries):
        """!
        Sets the list of direct subentries.
        @param entries: The new list of IntermediateStructure.Node.
        """
        self._entries.clear()
        for ent in entries:
            self.addentry(ent)

    def __str__(self):
        return "[IS\n{0}\n]".format("\n  ".join([str(x) for x in self.entries()]))

    @staticmethod
    def fromxml(xmlstring):
        """!
        Generates an anise.framework.projects.IntermediateStructure for an xml document.
        """
        entries = []
        xdoc = ET.ElementTree(ET.fromstring(xmlstring))
        xroot = xdoc.getroot()
        for xentry in xroot:
            if xentry.tag == "val":
                entries.append(IntermediateStructure._fromxml_val(xentry))
            elif xentry.tag == "let":
                entries.append(IntermediateStructure._fromxml_let(xentry))
            else:
                raise IntermediateStructure.ParseProjectDescriptionError("invalid tag: "+xentry.tag)
        result = IntermediateStructure(entries=entries)
        result.sourceformat = "xml"
        return result

    @staticmethod
    def fromfile(filename):
        """!
        Generates an anise.framework.projects.IntermediateStructure for an xml file.
        @param filename: Path to the project description file.
        """
        # we currently assume xml here since it is the only format
        result = IntermediateStructure.fromxml(anise.utils.basic.readfromfile(filename))
        result.sourcefilename = filename
        return result

    def tofile(self, filename):
        """!
        Saves the current state of this intermediate structure to a file.
        @param filename: The destination path.
        """
        if self.sourceformat == "xml":
            content = self.toxml()
        else:
            raise anise.framework.exceptions.BadInputError(
                "Unknown project description file format: " + repr(self.sourceformat))
        anise.utils.basic.writetofile(filename, content)

    def save(self):
        """!
        Saves the current state of this intermediate structure back to the disk.

        This works only for instances which were loaded from disk as well.
        """
        self.tofile(self.sourcefilename)

    def toxml(self):
        """!
        Generates an xml representation for this intermediate structure.
        """
        xroot = ET.Element("aniseprojectdesc")
        for entry in self.entries():
            if isinstance(entry, IntermediateStructure.Value):
                tag = IntermediateStructure._toxml_val(entry, None)
            elif isinstance(entry, IntermediateStructure.Let):
                tag = IntermediateStructure._toxml_let(entry)
            else:
                raise IntermediateStructure.ParseProjectDescriptionInternalError("Unknown thing: "+str(entry))
            xroot.append(tag)
        import xml.dom.minidom  # argh
        return xml.dom.minidom.parseString(ET.tostring(xroot)).toprettyxml(indent="    ")

    def find(self, key, lets=False, vals=False):
        """!
        Find all `let`s and/or `val`s with a certain key.
        """
        r = [x for x in self.entries() if x.key() == key]
        if not lets:
            r = [x for x in r if not isinstance(x, IntermediateStructure.Let)]
        if not vals:
            r = [x for x in r if not isinstance(x, IntermediateStructure.Value)]
        return r

    def injecttoprojectobject(self, project, g, l):
        """!
        Injects the (always static) information from this project description into a living universe object
        (typically for executing a task on top of it).
        @param project: The universe object.
        @param g: The globals dict.
        @param l: The locals dict.
        """
        for entry in self.entries():
            if isinstance(entry, IntermediateStructure.Value):
                x = IntermediateStructure._injecttoprojectobject_value(entry, g, l)
                project.append({entry.key(): x})
            elif isinstance(entry, IntermediateStructure.Let):
                IntermediateStructure._injecttoprojectobject_let(entry, g, l)
            else:
                raise IntermediateStructure.ParseProjectDescriptionInternalError("Unknown thing: "+str(entry))

    @staticmethod
    def _injecttoprojectobject_let(entry, g, l):
        """!
        Executes the let action from an IntermediateStructure.Let.
        """
        fct = IntermediateStructure._injecttoprojectobject_resolve(entry.key(), g, l)
        nparams = {}
        for y in entry.args():
            nparams[y] = IntermediateStructure._injecttoprojectobject_value(entry.getarg(y), g, l)
        fct(**nparams)

    @staticmethod
    def resolveobjectref(oref):
        """!
        Returns the resolved object for an objectref string.
        @param oref: The objectref string (e.g. 'homepage.makehomepage' or 'some.foo.Class().meth()+1').
        """
        g = dict(importlib.import_module("anise.framework.imports").__dict__)
        l = universe().__dict__
        return IntermediateStructure._injecttoprojectobject_resolve(oref, g, l)

    @staticmethod
    def _injecttoprojectobject_resolve(s, g, l):
        """!
        Returns the resolved object for an objectref string.
        """
        from anise.framework import imports
        if hasattr(imports, s):
            return getattr(imports, s)
        return eval(s, g, l)

    @staticmethod
    def _injecttoprojectobject_value(entry, g, l):
        """!
        Returns a plain Python value from an IntermediateStructure.Value.
        """
        if entry.vtype() == "str":
            result = entry.value()
        elif entry.vtype() == "int":
            result = int(entry.value()) if entry.value() else 0
        elif entry.vtype() == "bool":
            result = IntermediateStructure.Value.istrue(entry.value())
        elif entry.vtype() == "float":
            result = float(entry.value()) if entry.value() else 0.0
        elif entry.vtype() == "objectref":
            result = IntermediateStructure._injecttoprojectobject_resolve(entry.oref(), g, l)
        elif entry.vtype() == "list":
            result = []
            for k in sorted(entry.args()):
                result.append(IntermediateStructure._injecttoprojectobject_value(entry.getarg(k), g, l))
        elif entry.vtype() == "dict":
            result = {}
            for k in entry.args():
                result[k] = IntermediateStructure._injecttoprojectobject_value(entry.getarg(k), g, l)
        else:
            ptype = IntermediateStructure._injecttoprojectobject_resolve(entry.vtype(), g, l)
            args = {}
            for k in entry.args():
                args[k] = IntermediateStructure._injecttoprojectobject_value(entry.getarg(k), g, l)
            result = ptype(**args)
        return result

    @staticmethod
    def _fromxml_val(xentry):
        """!
        Creates an IntermediateStructure.Value from an xml node.
        """
        oargs = dict(xentry.attrib)
        if "t" in oargs:
            vtype = oargs["t"]
            del oargs["t"]
        elif IntermediateStructure._fromxml_val_isstring(xentry):
            vtype = "str"
        elif IntermediateStructure._fromxml_val_isdict(xentry):
            vtype = "dict"
        elif IntermediateStructure._fromxml_val_islist(xentry):
            vtype = "list"
        elif "oref" in oargs:
            vtype = "objectref"
        else:
            raise IntermediateStructure.ParseProjectDescriptionError("no 't' in '{0} {1}'".format(
                                    xentry.tag, xentry.attrib))
        result = IntermediateStructure.Value(vtype, k=xentry.attrib.get("k", None))
        if "k" in oargs:
            del oargs["k"]
        if (vtype == "str") and ("v" not in xentry.attrib):
            oargs["v"] = xentry.text
        if vtype in ["str", "int", "bool", "float"]:
            args = oargs
            result.setvalue(args.pop("v"))
        elif vtype == "objectref":
            args = oargs
            result.setoref(args.pop("oref"))
        else:
            args = {}
            for y in oargs:
                args[y] = IntermediateStructure.Value.fromstring(oargs[y], key=y)
        for ichild, xchild in enumerate(xentry):
            if vtype == "list":
                key = "i" + ("00000" + str(ichild))[:6] + "0"
            else:
                if "k" not in xchild.attrib:
                    raise IntermediateStructure.ParseProjectDescriptionError("no 'k' in '{0} {1}'".format(
                                            xchild.tag, xchild.attrib))
                key = xchild.attrib["k"]
            args[key] = IntermediateStructure._fromxml_val(xchild)
        for y in args:
            result.setarg(y, args[y])
        return result

    @staticmethod
    def _fromxml_val_isstring(xentry):
        """!
        Checks if a particular xml node represents a string.
        """
        if len([y for y in xentry.attrib.keys() if y not in ["k", "t", "v"]]) > 0:
            return False
        return len(xentry) == 0

    @staticmethod
    def _fromxml_val_isdict(xentry, withoutkeys=False):
        """!
        Checks if a particular xml node represents a value dict.
        """
        if len([y for y in xentry.attrib.keys() if y not in ["k", "t"]]) > 0:
            return False
        cl = list(xentry)
        if len(cl) == 0:
            return False
        return len([c for c in cl if ("k" in c.attrib) == withoutkeys]) == 0

    @staticmethod
    def _fromxml_val_islist(xentry):
        """!
        Checks if a particular xml node represents a value list.
        """
        return IntermediateStructure._fromxml_val_isdict(xentry, withoutkeys=True)

    @staticmethod
    def _fromxml_let(xentry):
        """!
        Creates an IntermediateStructure.Let from an xml node.
        """
        rparams = dict(xentry.attrib)
        del rparams["k"]
        params = {}
        for y in rparams:
            params[y] = IntermediateStructure.Value.fromstring(rparams[y], y)
        for xchild in xentry:
            if "k" not in xchild.attrib:
                raise IntermediateStructure.ParseProjectDescriptionError("no 'k' in '{0} {1}'".format(
                                        xchild.tag, xchild.attrib))
            params[xchild.attrib["k"]] = IntermediateStructure._fromxml_val(xchild)
        return IntermediateStructure.Let(xentry.attrib["k"], **params)

    @staticmethod
    def _toxml_val(entry, xcurr):
        """!
        Converts an IntermediateStructure.Value to an xml node.
        """
        xresult = ET.Element("val")
        if not(isinstance(entry.parent(), IntermediateStructure.Value) and entry.parent().vtype() == "list"):
            xresult.attrib["k"] = str(entry.key())
        if entry.vtype() == "str":
            txt = str(entry.value())
            if (len(txt) < 75) and ("\n" not in txt):
                if (xcurr is not None) and ("k" in xresult.attrib) and (len(txt)+len(xresult.attrib["k"])+4) <= 80:
                    xcurr.attrib[xresult.attrib["k"]] = txt
                    xresult = None
                else:
                    xresult.attrib["v"] = txt
            else:
                xresult.text = txt
        elif entry.vtype() in ["int", "bool", "float"]:
            xresult.attrib["v"] = str(entry.value())
            xresult.attrib["t"] = str(entry.vtype())
        elif entry.vtype() == "objectref":
            xresult.attrib["oref"] = str(entry.oref())
        else:
            if (entry.vtype() in ["dict", "list"]) and len(entry.args()) > 0:
                pass
            else:
                xresult.attrib["t"] = str(entry.vtype())
            for y in sorted(entry.args()):
                xnn = IntermediateStructure._toxml_val(entry.getarg(y), xresult if (entry.vtype()
                                                                                    not in ["list", "dict"]) else None)
                if xnn is not None:
                    xresult.append(xnn)
        return xresult

    @staticmethod
    def _toxml_let(entry):
        """!
        Converts an IntermediateStructure.Let to an xml node.
        """
        xresult = ET.Element("let")
        xresult.attrib["k"] = str(entry.key())
        for param in sorted(entry.args()):
            xparam = IntermediateStructure._toxml_val(entry.getarg(param), xresult)
            if xparam is not None:
                xresult.append(xparam)
        return xresult
