# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

## @package anise.framework.report
## The Anise report facility keeps track of the steps of the execution process, the progresses and all output messages.

import contextvars
import re
import sys
import threading

import anise.framework.exceptions
import anise.framework.features
import anise.framework.globalvars
import anise.utils


HOOK_EXECUTION_SCOPEPART_ADDED = anise.framework.features.Hook(
    doc="This hook is called when a new execution scope part appears.", wizard_functionparams=["newpart"])


class ExecutionScopePart:
    """!
    Base class of ExecutionScope (read there for more). It also includes logging output or output from external
    programs, which are displayed together with real execution scopes during task execution.
    """

    def __init__(self, value, hints=None):
        """!
        @param value: The content value.
        @param hints: A list of hint symbols (typically used for log message severity).
        """
        self._address = []
        self._id = 0
        self._id_counter_ = 1
        self._scopepartsbyid = {}
        self._value = value
        self._hints = hints or []

    def id(self):
        """!
        Returns the ID of this part.
        """
        return self._id

    def value(self):
        """!
        Returns the content value of this execution scope.
        """
        return self._value

    def hints(self):
        """!
        Returns the content hint.

        This is typically used for storing the severity of log messages, so the presentation can differ.
        """
        return self._hints

    def address(self):
        """!
        Returns the address of this part.
        """
        return tuple(self._address)

    def getscopepartsbyidrange(self, minid=None, maxid=None):
        """!
        Gets a list of execution scope parts by a range of IDs.
        """
        _lock.acquire()
        result = []
        if minid is None:
            minid = 0
        if maxid is None:
            maxid = max(_rootscope._scopepartsbyid.keys())
        for i in range(minid, maxid+1):
            x = _rootscope._scopepartsbyid.get(i)
            if x:
                result.append(x)
        _lock.release()
        return result


class ExecutionScope(ExecutionScopePart):
    """!
    An execution scope is one piece of work to do in a complete task executions.

    It is the parent of another execution scope (just the root scope doesn't have a parent) and can have child scopes.
    This leads to a tree structure of scope nodes. Each node can have metadata infos like a title or a progress.
    Those are displayed during the execution of a task (together with other ExecutionScopePart subclasses' instances)
    """

    class Progress:
        """!
        Progress info for an execution scope.
        """

        def __init__(self):
            pass

        def tojson(self):
            """!
            Returns a simple object structure for json transfers.
            """
            return self.__dict__

    class ProgressDeterminate(Progress):
        """!
        Progress info for a determinate progress.
        """

        def __init__(self, value):
            """!
            @param value: The progress between 0.0 and 1.0.
            """
            ExecutionScope.Progress.__init__(self)
            self.value = min(max(value, 0), 1)
            self.kind = "determinate"
            self.stopped = False

    class ProgressIndeterminate(Progress):
        """!
        Progress info for an indeterminate progress.
        """

        def __init__(self):
            ExecutionScope.Progress.__init__(self)
            self.kind = "indeterminate"
            self.stopped = False

    class ProgressFinished(Progress):
        """!
        Progress info for a finished execution.
        """

        def __init__(self):
            ExecutionScope.Progress.__init__(self)
            self.kind = "finished"
            self.stopped = True

    class ProgressDoneWithUnknownSuccess(Progress):
        """!
        Progress info for an execution which is terminated but without explicit success information.
        """

        def __init__(self):
            ExecutionScope.Progress.__init__(self)
            self.kind = "donewithunknownsuccess"
            self.stopped = True

    class ProgressFailed(Progress):
        """!
        Progress info for a failed execution.
        """

        def __init__(self):
            ExecutionScope.Progress.__init__(self)
            self.kind = "failed"
            self.stopped = True

    def __init__(self, title):
        """!
        @param title: The scope title text.
        """
        ExecutionScopePart.__init__(self, self)
        self._subparts = []
        self._progress = ExecutionScope.ProgressIndeterminate()
        self._title = title

    def appendchild(self, child):
        """!
        Appends a child scope to this scope.
        """
        from anise.framework.projects import currentuniverse
        _lock.acquire()
        child._parentscope = self
        child._address = list(self._address+[len(self._subparts)])
        child._id = _rootscope._id_counter_
        self._subparts.append(child)
        _rootscope._scopepartsbyid[child.id()] = child
        _rootscope._id_counter_ += 1
        if currentuniverse():
            for h in currentuniverse().gethooks(HOOK_EXECUTION_SCOPEPART_ADDED):
                h(child)
        _lock.release()

    def progress(self):
        """!
        Returns the current progress state of this scope.
        @return: An instance of ExecutionScope.Progress.
        """
        return self._progress

    def setprogress(self, value):
        """!
        Sets the progress of this scope.
        @param value: An instance of ExecutionScope.Progress.
        """
        self._progress = value

    def title(self):
        """!
        Returns the title string of this scope.
        """
        return self._title

    def rawsubparts(self):
        """!
        Returns a list of all direct subparts of this scope.
        """
        return list(self._subparts)

    def newsubscope(self, title):
        """!
        Adds a new sub scope to the current top scope.
        @param title: The scope title text.
        """
        result = ExecutionScope(title=title)
        topscope().appendchild(result)
        return result

    def __enter__(self):
        if self._title:
            anise.utils.logging.loginfo("Begin: "+self._title, printcaller=False, terminalonly=True)
        _scopestack.set(_scopestack.get() + (self,))
        _globscopestack.append(self)
        return self

    def __exit__(self, etype, evalue, etraceback):
        _scopestack.set(_scopestack.get()[:-1])
        _globscopestack.pop()

        def _finishscope(sc):
            if etype:
                sc.setprogress(ExecutionScope.ProgressFailed())
            elif not isinstance(sc.progress(), ExecutionScope.ProgressFailed):
                for csc in sc.rawsubparts():
                    if isinstance(csc, ExecutionScope):
                        _finishscope(csc)
                sc.setprogress(ExecutionScope.ProgressFinished())
        _finishscope(self)
        if self.title():
            if etype:
                anise.utils.logging.logerror("Failed: "+self.title(), printcaller=False, terminalonly=True)
            else:
                anise.utils.logging.loginfo("Finished: "+self.title(), printcaller=False, terminalonly=True)


_scopestack = contextvars.ContextVar('_scopestack', default=tuple())
_globscopestack = []


class RootExecutionScope(ExecutionScope):
    """!
    The root execution scope.
    """

    def __init__(self):
        ExecutionScope.__init__(self, title=None)


_rootscope = RootExecutionScope()
_rootscope.__enter__()

_lock = threading.Lock()


class AniseOutputStream:
    """!
    Proxy for terminal output/error channel, so we can record output messages.
    """

    def __init__(self, originalstream):
        self.result = ""
        self.originalstream = originalstream

    def write(self, b):
        b = re.sub("\033\\[[0-9;]+m", "", b)
        topscope().appendchild(ExecutionScopePart(b))
        if self.originalstream:
            self.originalstream.write(b)

    def __getattr__(self, item):
        return getattr(self.originalstream, item)

    def __dir__(self):
        result = set(self.__dict__.keys())
        for k in dir(self.originalstream):
            result.add(k)
        return list(result)


def topscope(i=0):
    """!
    Returns a scope from the top of the stack (without removing it).
    @param i: Which scope to return (counting from top).
    """
    return (_scopestack.get() or _globscopestack)[-i-1]


def install_proxies():
    """!
    Installs stdout/stderr proxies, so we can record all output made.
    """
    sys.stdout = AniseOutputStream(sys.stdout)
    sys.stderr = AniseOutputStream(sys.stderr)


def _logmessage(text, severity, terminalonly, tofile, textescaper):
    """!
    A message logger implementation which also records the logging output to the report.
    """
    from anise.utils.logging import Severity
    if (severity != Severity.Debug) or anise.framework.globalvars.logdebug:
        rtext = ((len(_scopestack.get()) - 1) * ". ") + text
        print(textescaper(rtext), file=tofile)
    if not terminalonly:
        if severity == Severity.Debug:
            hints = ["logdebug"]
        elif severity == Severity.Info:
            hints = ["loginfo"]
        elif severity == Severity.Warning:
            hints = ["logwarning"]
        elif severity == Severity.Error:
            hints = ["logerror"]
        else:
            raise anise.framework.exceptions.BadInputError("unknown severity: " + repr(severity))
        topscope().appendchild(ExecutionScopePart(text, hints=hints))


anise.utils.logging.setlowlevellogger(_logmessage)
