# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

## @package anise.utils.data
## Helpers for creation and handling of some data structures.

import hashlib
import random
import socket


def constructobject(**d):
    """!
    Creates an object with properties from the given keyword arguments.
    """
    class Aniseobject:
        pass
    r = Aniseobject()
    for k in d:
        setattr(r, k, d[k])
    return r


def createobjectstructure(root, newnames, objclass=None):
    """!
    Creates a chain of new objects, so you can access property chains like `foo.bar.baz.bam` on the root object.
    @param root: The object to populate with the new structure.
    @param newnames: The complete property chain (as string).
    @param objclass: Which object class to use for the cascaded new members.
    @return: Tuple of the deepest object in the cascade (corresponding to the last part of the chain) and its parent.
    """
    if objclass is None:
        class _AniseObject:
            pass
        objclass = _AniseObject
    parent = root
    parentparent = None
    for newname in newnames.split("."):
        parentparent = parent
        if hasattr(parent, newname):
            parent = getattr(parent, newname)
        else:
            _parent = objclass()
            setattr(parent, newname, _parent)
            parent = _parent
    return parent, parentparent


def getuniqueidentifier():
    """!
    Returns a random and (statistically nearly) unique string.
    """
    r = socket.gethostname() + "".join([str(random.random()) for x in range(4)])
    return hashlib.sha224(r.encode()).hexdigest()


def stripname(name, lowercase=True):
    """!
    Strips a name to something very bare with only alphabetic letters.
    @param name: The input name.
    @param lowercase: If only lowercase characters are allowed.
    """
    res = ""
    for c in name:
        if c.lower() in "abcdefghijklmnopqrstuvwxyz":
            res += c
        elif c == "0":
            res += "Zero"
        elif c == "1":
            res += "One"
        elif c == "2":
            res += "Two"
        elif c == "3":
            res += "Three"
        elif c == "4":
            res += "Four"
        elif c == "5":
            res += "Five"
        elif c == "6":
            res += "Six"
        elif c == "7":
            res += "Seven"
        elif c == "8":
            res += "Eight"
        elif c == "9":
            res += "Nine"
    if lowercase:
        res = res.lower()
    return res
