# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

## @package anise.utils.logging
## Logging.

import inspect
import sys
import traceback

import anise.framework.exceptions
import anise.framework.globalvars


class Severity:
    """!
    Enumeration of log message severities.
    """
    Debug = 0
    Info = 100
    Warning = 200
    Error = 300


def logmessage(text, severity, printcaller=True, skiplevels=1, terminalonly=False):
    """!
    Logs a text message.
    @param text: The log message.
    @param severity: The message severity. One of the values from anise.utils.logging.Severity.
    @param printcaller: Also logs the caller name.
    @param skiplevels: If given, it controls the position on the call stack which is used for logging the caller name.
    @param terminalonly: If this log message shall just go to the terminal and not to the report, the web interface, 
                         etc. Only used by the infrastructure.
    """
    text = str(text)
    if severity == Severity.Error:
        color = "91"
    elif severity == Severity.Warning:
        color = "93"
    elif severity == Severity.Info:
        color = "94"
    elif severity == Severity.Debug:
        color = "37"
    else:
        raise anise.framework.exceptions.BadInputError("Unknown severity: "+repr(severity))
    if printcaller:
        stack = inspect.stack()
        rparentframe = stack[skiplevels]
        parentframe = rparentframe[0]
        name = []
        pmodule = inspect.getmodule(parentframe)
        if pmodule:
            name.append(pmodule.__name__)
        if 'self' in parentframe.f_locals:
            name.append(parentframe.f_locals['self'].__class__.__name__)
        if len(name) == 0:
            name.append(parentframe.f_code.co_name)
        text = ".".join(name) + ": " + text
    tofile = getattr(sys.stdout, "originalstream", sys.stdout)
    def textescaper(t):
        if (sys.platform != "win32") and sys.stdout.isatty():
            return "\033[{color}m{t}\033[0m".format(color=color, t=t)
        else:
            return t
    _logmessage(text, severity, terminalonly, tofile, textescaper)


def _logmessage_default(text, severity, terminalonly, tofile, textescaper):
    """!
    This is the fallback lowlevel logging implementation. Typically it
    gets overridden by a different one in early initialization steps.
    """
    if (severity != Severity.Debug) or anise.framework.globalvars.logdebug:
        print(textescaper(text), file=tofile)


_logmessage = _logmessage_default


def setlowlevellogger(fct):
    """!
    Sets the lowlevel logging implementation.
    @param fct: The logger function.
    """
    global _logmessage
    _logmessage = fct


def logdebug(text, printcaller=True, skiplevels=1, terminalonly=False):
    """!
    Logs a text message with debug severity.
    @param text: The log message.
    @param printcaller: Also logs the caller name.
    @param skiplevels: If given, it controls the position on the call stack which is used for logging the caller name.
    @param terminalonly: If this log message shall just go to the terminal and not to the report, the web interface, 
                         etc. Only used by the infrastructure.
    """
    logmessage(text, Severity.Debug, printcaller, skiplevels+1, terminalonly)


def loginfo(text, printcaller=True, skiplevels=1, terminalonly=False):
    """!
    Logs a text message with info severity.
    @param text: The log message.
    @param printcaller: Also logs the caller name.
    @param skiplevels: If given, it controls the position on the call stack which is used for logging the caller name.
    @param terminalonly: If this log message shall just go to the terminal and not to the report, the web interface, 
                         etc. Only used by the infrastructure.
    """
    logmessage(text, Severity.Info, printcaller, skiplevels+1, terminalonly)


def logwarning(text, printcaller=True, skiplevels=1, terminalonly=False):
    """!
    Logs a text message with warning severity.
    @param text: The log message.
    @param printcaller: Also logs the caller name.
    @param skiplevels: If given, it controls the position on the call stack which is used for logging the caller name.
    @param terminalonly: If this log message shall just go to the terminal and not to the report, the web interface, 
                         etc. Only used by the infrastructure.
    """
    logmessage(text, Severity.Warning, printcaller, skiplevels+1, terminalonly)


def logerror(text, printcaller=True, skiplevels=1, terminalonly=False):
    """!
    Logs a text message with error severity.
    @param text: The log message.
    @param printcaller: Also logs the caller name.
    @param skiplevels: If given, it controls the position on the call stack which is used for logging the caller name.
    @param terminalonly: If this log message shall just go to the terminal and not to the report, the web interface, 
                         etc. Only used by the infrastructure.
    """
    logmessage(text, Severity.Error, printcaller, skiplevels+1, terminalonly)


def logexception(exception, prolog="Error: "):
    """!
    Logs an exception.
    @param exception: The occurred exception. This only works from within the `except`-handler of this exception.
    @param prolog: This is prepended to the exception text in the output.
    """
    e_full = traceback.format_exc()
    logdebug(e_full, printcaller=False)
    logerror(prolog+str(exception), printcaller=False)
