# Copyright (C) 2014-2020, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

## @package anise.utils.ssh
## Utilities for ssh usage.

import shlex

from anise.utils import basic


def call(server, command, options=None, port=22, identityfile=None):
    """!
    Executes a command via ssh.
    @param server: The ssh server hostname.
    @param command: The command to execute.
    @param options: Additional ssh options (as string list).
    @param port: Port number.
    @param identityfile: Optional identity file for authentication.
    """
    options = options or []
    idfileoptions = ["-i", identityfile] if identityfile else []
    r, o = basic.call("ssh", "-p", str(port), *(options+idfileoptions+[server, command]),
                      raise_on_errors="unable to call ssh")
    return o


class Mount(basic.Mount):
    """!
    Mounts a volume via ssh.
    """

    def __init__(self, src, dst, options=None, port=22, identityfile=None):
        """!
        @param src: The remote mount source (an ssh address like `user\@machine`).
        @param dst: The local mount destination.
        @param options: Optional additional mount options.
        @param port: Optional tcp port.
        @param identityfile: Optional identity file for authentication.
        """
        options = options or []
        idfileoptions = ["-o", "IdentityFile="+shlex.quote(identityfile)] if identityfile else []
        basic.Mount.__init__(self, src, dst, options + idfileoptions + ["-p", str(port)],
                             mounter=["sshfs"], unmounter=["fusermount", "-u"], needsroot=False)


def copy(src, dst, options=None, port=22, identityfile=None):
    """!
    Copies a local directory to a remote place via ssh.
    @param src: The source.
    @param dst: The destination.
    @param options: Optional additional mount options.
    @param port: Optional tcp port.
    @param identityfile: Optional identity file for authentication.
    """
    options = options or []
    idfileoptions = ["-o", "IdentityFile="+shlex.quote(identityfile)] if identityfile else []
    basic.call("scp", "-r", "-P", str(port), *(options+idfileoptions+[src, dst]), raise_on_errors="unable to call ssh")
